
/* need more stack depth? apply here */

#define MACRO_STACK_DEPTH 16
#define REPEAT_STACK_DEPTH 16

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define FAILED 0
#define SUCCEEDED 1

#define OFF 0
#define ON 1

#define STACK_CALCULATE_DELAY 2

#define STACK_NONE 0
#define STACK_INSIDE 1
#define STACK_OUTSIDE 2

#define CASE_MODE_UPPER 0
#define CASE_MODE_INSENSITIVE 1

#define SECTION_STATUS_FREE 0
#define SECTION_STATUS_FORCE 1
#define SECTION_STATUS_OVERWRITE 2
#define SECTION_STATUS_HEADER 3
#define SECTION_STATUS_SEMIFREE 4
#define SECTION_STATUS_ABSOLUTE 5

#define INPUT_NUMBER_EOL 2
#define INPUT_NUMBER_ADDRESS_LABEL 3
#define INPUT_NUMBER_STRING 4
#define INPUT_NUMBER_STACK 5
#define INPUT_NUMBER_FLOAT 6

#define GET_NEXT_TOKEN_STRING 2
#define EVALUATE_TOKEN_NOT_IDENTIFIED 2
#define EVALUATE_TOKEN_EOP 6

#define OUTPUT_OBJECT 0
#define OUTPUT_ROM 1
#define OUTPUT_LIBRARY 2
#define OUTPUT_NONE 3
#define OUTPUT_PROGRAM 4

#define REFERENCE_TYPE_DIRECT_16BIT 0
#define REFERENCE_TYPE_RELATIVE 1
#define REFERENCE_TYPE_DIRECT_8BIT 2
#define REFERENCE_TYPE_DIRECT_24BIT 3

#define MAX_NAME_LENGTH 128

#define ERROR_LEVEL_MIN 0
#define ERROR_LEVEL_MAX 1

#define ERROR_NONE 0
#define ERROR_LOG 1
#define ERROR_DIR 2
#define ERROR_NUM 3
#define ERROR_INC 4
#define ERROR_INB 5
#define ERROR_UNF 6
#define ERROR_INP 7
#define ERROR_STC 8
#define ERROR_WRN 9

#define SKIP_NO 0
#define SKIP_8BIT 1
#define SKIP_16BIT 2

#define WIDTH_8BIT 0
#define WIDTH_16BIT 1

/* A = absolute section */
/* c = stack (1 byte)   */
/* C = stack (2 bytes)  */
/* T = stack (3 bytes)  */
/* d = data (1 byte)    */
/* y = data (2 bytes)   */
/* z = data (3 bytes)   */
/* D = data block       */
/* O = origin           */
/* B = ROM bank         */
/* b = BASE (65816)     */
/* L = label            */
/* r = 16bit reference  */
/* R = pc ref           */
/* Q = 8bit reference   */
/* q = 24bit reference  */
/* S = section          */
/* s = end of section   */
/* x = dsb              */
/* X = dsw              */
/* E = end of file      */
/* f = file name id     */
/* a = align (byte)     */
/* k = line number      */

#ifdef GB

/* opcode types */

/* 0 - plain text  8b */
/* 1 - x              */
/* 2 - ?              */
/* 3 - plain text 16b */
/* 8 - *           8b */
/* 9 - *          16b */

#define OP_SIZE_MAX 16

#ifdef AMIGA
struct optcode {
  char *op;
  int  hex;
  short int type;
  short int value;
};
#else
struct optcode {
  char *op;
  int  hex;
  int  type;
  int  value;
};
#endif

#endif

#ifdef MCS6502

/* opcode types */

/* 0 - plain text  8b */
/* 1 - x              */
/* 2 - ?              */

#define OP_SIZE_MAX 16

#ifdef AMIGA
struct optcode {
  char *op;
  int  hex;
  short int  type;
  short int  skip_8bit;
};
#else
struct optcode {
  char *op;
  int  hex;
  int  type;
  int  skip_8bit;
};
#endif

#endif

#ifdef MCS6510

/* opcode types */

/* 0 - plain text  8b */
/* 1 - x              */
/* 2 - ?              */

#define OP_SIZE_MAX 16

#ifdef AMIGA
struct optcode {
  char *op;
  int  hex;
  short int  type;
  short int  skip_8bit;
};
#else
struct optcode {
  char *op;
  int  hex;
  int  type;
  int  skip_8bit;
};
#endif

#endif

#ifdef Z80

/* opcode types */

/* 0 - plain text  8b */
/* 1 - x              */
/* 2 - ?              */
/* 3 - plain text 16b */
/* 4 - x              */
/* 5 - x          24b */
/* 6 - ?          16b */
/* 7 - x x        16b */
/* 8 - *           8b */
/* 9 - *          16b */
/* a - * x        24b */

#define OP_SIZE_MAX 32

#ifdef AMIGA
struct optcode {
  char *op;
  int  hex;
  short int type;
  short int hex_x;
  short int value;
};
#else
struct optcode {
  char *op;
  int  hex;
  int  type;
  int  hex_x;
  int  value;
};
#endif

#endif

#ifdef W65816

/* opcode types */

/* 0 - plain text  8b */
/* 1 - x              */
/* 2 - ?              */
/* 3 - &              */
/* 4 - x/? (mem/acc)  */
/* 5 - x x            */
/* 6 - REP/SEP        */
/* 7 - x/? (index)    */

#define OP_SIZE_MAX 16

#ifdef AMIGA
struct optcode {
  char *op;
  int  hex;
  short int  type;
  short int  skip_xbit;
};
#else
struct optcode {
  char *op;
  int  hex;
  int  type;
  int  skip_xbit;
};
#endif

#endif

#ifdef C8051

/* opcode types */

/* 0 - plain text  8b */
/* 1 - ?              */
/* 2 - x              */
/* 3 - x x            */
/* 4 - x x (reverse)  */
/* 5 - ! (11b)        */

#define OP_SIZE_MAX 16

#ifdef AMIGA
struct optcode {
  char *op;
  int  hex;
  short int  type;
};
#else
struct optcode {
  char *op;
  int  hex;
  int  type;
};
#endif

#endif


#define DEFINITION_TYPE_VALUE 0
#define DEFINITION_TYPE_STRING 1

struct gba_def {
  char   alias[MAX_NAME_LENGTH];
  char   string[MAX_NAME_LENGTH];
  int    value;
  int    type;
  int    size;
  struct gba_def *next;
};

struct gba_macro {
  char name[MAX_NAME_LENGTH];
  int  arguments_start[256];
  int  arguments;
  int  supplied_arguments;
  int  start;
  int  calls;
  int  filename_id;
  int  start_line;
  struct gba_macro *next;
};

struct label_def {
  char label[MAX_NAME_LENGTH];
  int  section_status;
  int  section_id;
  int  address; /* in bank */
  int  bank;
  int  slot;
#ifdef W65816
  int  base;
#endif
  int  type;
  int  filename_id;
  int  linenumber;
  struct label_def *next;
};

struct section_def {
  char name[MAX_NAME_LENGTH];
  int  address; /* in bank */
  int  bank;
  int  slot;
  int  size;
  int  status; /* 1 = force, 0 = free */
  int  id;
  int  i;
  int  maxsize_status; /* ON/OFF */
  int  maxsize;
  unsigned char *data;
  struct section_def *next;
};

struct incbin_file_data {
  struct incbin_file_data *next;
  char   *data;
  char   name[MAX_NAME_LENGTH];
  int    size;
};

struct export_def {
  char   name[MAX_NAME_LENGTH];
  struct export_def *next;
};

struct active_file_info {
  int    filename_id;
  int    line_current;
  struct active_file_info *next;
  struct active_file_info *prev;
};

struct file_name_info {
  char   name[MAX_NAME_LENGTH];
  int    id;
  struct file_name_info *next;
};

struct slot {
  int address;
  int size;
  struct slot *next;
};

#define STACK_ITEM_TYPE_VALUE 0
#define STACK_ITEM_TYPE_OPERATOR 1
#define STACK_ITEM_TYPE_STRING 2
#define STACK_ITEM_TYPE_DELETED 3

#define SI_OP_PLUS 0
#define SI_OP_MINUS 1
#define SI_OP_MULTIPLY 2
#define SI_OP_LEFT 3
#define SI_OP_RIGHT 4
#define SI_OP_OR 5
#define SI_OP_AND 6
#define SI_OP_DIVIDE 7
#define SI_OP_POWER 8
#define SI_OP_SHIFT_LEFT 9
#define SI_OP_SHIFT_RIGHT 10
#define SI_OP_MODULO 11

#define SI_SIGN_POSITIVE 0
#define SI_SIGN_NEGATIVE 1

#define STACKS_TYPE_8BIT 0
#define STACKS_TYPE_16BIT 1
#define STACKS_TYPE_24BIT 2

struct stackitem {
  int type;
  int value;
  int sign;
  char string[64];
};

struct stack {
  struct stackitem *stack;
  struct stack *next;
  int filename_id;
  int stacksize;
  int linenumber;
  int type;
  int bank;
  int slot;
#ifdef W65816
  int base;
#endif
  int section_status;
  int section_id;
  int address;
};

struct macro_runtime {
  struct gba_macro *macro;
  int macro_end;
  int macro_end_line;
  int macro_end_filename_id;
  int supplied_arguments;
};

struct repeat_runtime {
  int start;
  int start_line;
  int counter;
};

#define TYPE_STRING 0
#define TYPE_VALUE 1
#define TYPE_LABEL 2
#define TYPE_STACK_CALCULATION 3
