
/*
 wla - part of wla dx gb-z80/z80/6502/6510/65816 macro assembler package by/copyright
 ville helin <vhelin@cc.hut.fi>. this is gpl software.
 */

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef SUCCEEDED
#undef FAILED
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef UNIX
#include <unistd.h>
#endif

#include "main.h"
#include "defines.h"

#include "parse.h"
#include "include_file.h"
#include "pass_1.h"
#include "pass_2.h"
#include "pass_3.h"
#include "pass_4.h"


FILE *file_out_ptr = NULL;

/* amiga specific definitions */

#ifdef AMIGA
long __stack = 65536;
#endif

#ifdef AMIGA
#ifdef GB
char version_string[] = "$VER: WLA-GB 7.8 (10.05.2001)";
#endif
#ifdef Z80
char version_string[] = "$VER: WLA-Z80 7.8 (10.05.2001)";
#endif
#ifdef MCS6502
char version_string[] = "$VER: WLA-6502 7.8 (10.05.2001)";
#endif
#ifdef W65816
char version_string[] = "$VER: WLA-65816 7.8 (10.05.2001)";
#endif
#ifdef MCS6510
char version_string[] = "$VER: WLA-6510 7.8 (10.05.2001)";
#endif
#ifdef C8051
char version_string[] = "$VER: WLA-8051 7.8 (10.05.2001)";
#endif
#endif

char gba_tmp_name[32], gba_unfolded_name[32];

extern struct incbin_file_data *incbin_file_data_first, *ifd_tmp;
extern struct label_def *label_tmp, *labels;
extern struct gba_macro *macros_first;
extern struct gba_def *defines, *tmp_def;
extern struct export_def *export_first, *export_last, *export_tmp;
extern struct stack *stacks_first, *stacks_tmp, *stacks_last, *stacks_header_first, *stacks_header_last;
extern char *unfolded_buffer;
extern char *include_in_tmp, *tmp_a;
extern char *rom_banks, *rom_banks_usage_table;
extern char include_dir[MAX_NAME_LENGTH];
extern int include_in_tmp_size, tmp_a_size, *banks, *bankaddress;

int output_format = OUTPUT_NONE, verbose_mode = OFF, compact_mode = OFF, error_level = ERROR_LEVEL_MAX, test_mode = OFF, case_mode = CASE_MODE_INSENSITIVE;
int symbol_mode = OFF, extra_definitions = OFF;

char final_name[MAX_NAME_LENGTH];



int main(int argc, char *argv[]) {

  int i = SUCCEEDED, x;


  atexit(procedures_at_exit);

  if (argc == 4) {
    if (parse_flags(argv[1]) == FAILED)
      i = FAILED;
    else
      strcpy(final_name, argv[3]);
  }
  else if (argc == 3) {
    if (parse_flags(argv[1]) == FAILED)
      i = FAILED;
    else {
      for (x = 0; x < strlen(argv[2]) && *(argv[2] + x) != '.'; x++)
	final_name[x] = *(argv[2] + x);
      final_name[x++] = '.';
      if (output_format == OUTPUT_ROM) {
#ifdef GB
	final_name[x++] = 'g';
	final_name[x++] = 'b';
	final_name[x] = 0;
#else
	final_name[x++] = 'r';
	final_name[x++] = 'o';
	final_name[x++] = 'm';
	final_name[x] = 0;
#endif
      }
      else if (output_format == OUTPUT_OBJECT) {
	final_name[x++] = 'o';
	final_name[x] = 0;
      }
      else if (output_format == OUTPUT_LIBRARY) {
	final_name[x++] = 'l';
	final_name[x++] = 'i';
	final_name[x++] = 'b';
	final_name[x] = 0;
      }
#ifndef GB
      else if (output_format == OUTPUT_PROGRAM) {
	final_name[x++] = 'p';
	final_name[x++] = 'r';
	final_name[x++] = 'g';
	final_name[x] = 0;
      }
#endif
    }
  }
  else
    i = FAILED;

#ifdef GB
  if (symbol_mode == ON && output_format != OUTPUT_ROM) {
    fprintf(stderr, "WARNING: Symbolic information is available only in the ROM output mode.\n");
    symbol_mode = OFF;
  }
#endif

  if (compact_mode == ON && output_format == OUTPUT_ROM) {
    fprintf(stderr, "WARNING: Compact mode is off in ROM output mode.\n");
    compact_mode = OFF;
  }

#ifndef GB
  if (compact_mode == ON && output_format == OUTPUT_PROGRAM) {
    fprintf(stderr, "WARNING: Compact mode is off in PROGRAM output mode.\n");
    compact_mode = OFF;
  }
#endif

  if (i == FAILED || output_format == OUTPUT_NONE) {
#ifdef GB
    printf("\nWLA GB-Z80 Macro Assembler v7.8\n");
    printf("Copyright 1998-2001 Ville Helin\n");
    printf("USAGE: %s -[cfstuvx]{lor} <ASM FILE> [OUTPUT FILE]\n", argv[0]);
    printf("Commands:             Options:\n");
    printf("l  Library file       c  Compact local labels\n");
    printf("o  Object file        f  Crippled error message engine\n");
    printf("r  ROM file           s  NO$GMB symbol file\n");
    printf("                      t  Test compile\n");
    printf("                      u  Mnemonics are in uppercase.\n");
    printf("                      v  Verbose messages\n");
    printf("                      x  Extra compile time definitions.\n\n");
#endif
#ifndef GB
#ifdef Z80
    printf("\nWLA Z80 Macro Assembler v7.8\n");
#endif
#ifdef MCS6502
    printf("\nWLA 6502 Macro Assembler v7.8\n");
#endif
#ifdef W65816
    printf("\nWLA 65816 Macro Assembler v7.8\n");
#endif
#ifdef MCS6510
    printf("\nWLA 6510 Macro Assembler v7.8\n");
#endif
#ifdef C8051
    printf("\nWLA 8051 Macro Assembler v7.8\n");
#endif
    printf("Copyright 1998-2001 Ville Helin\n");
    printf("USAGE: %s -[cfstuvx]{blor} <ASM FILE> [OUTPUT FILE]\n", argv[0]);
    printf("Commands:             Options:\n");
    printf("b  Program file       c  Compact local labels\n");
    printf("l  Library file       f  Crippled error message engine\n");
    printf("o  Object file        s  NO$GMB symbol file\n");
    printf("r  ROM file           t  Test compile\n");
    printf("                      u  Mnemonics are in uppercase\n");
    printf("                      v  Verbose messages\n");
    printf("                      x  Extra compile time definitions.\n\n");
#endif

    return 0;
  }

  if (strcmp(argv[2], final_name) == 0) {
    fprintf(stderr, "MAIN: Input and output files share the same name!\n");
    return 1;
  }

  generate_tmp_names();

  file_out_ptr = fopen(gba_tmp_name, "wb");
  if (file_out_ptr == NULL) {
    fprintf(stderr, "MAIN: Error opening file \"%s\".\n", gba_tmp_name);
    return 1;
  }

  /* small inits */
  include_dir[0] = 0;
  if (extra_definitions == ON)
    generate_extra_definitions();

  /* start the process */
  if (include_file(argv[2]) == FAILED)
    return 1;

  if (pass_1() == FAILED)
    return 1;
  if (pass_2() == FAILED)
    return 1;
  if (pass_3() == FAILED)
    return 1;
  if (pass_4() == FAILED)
    return 1;

  return 0;
}


int parse_flags(char *f) {

  int l;


  if (*f != '-') {
    return FAILED;
  }

  l = strlen(f);
  if (l == 1)
    return FAILED;

  for (f++, l--; l > 0; l--, f++) {
    switch (*f) {

    case 'o':
      if (output_format != OUTPUT_NONE)
	return FAILED;
      output_format = OUTPUT_OBJECT;
      continue;

    case 'l':
      if (output_format != OUTPUT_NONE)
	return FAILED;
      output_format = OUTPUT_LIBRARY;
      continue;

    case 'r':
      if (output_format != OUTPUT_NONE)
	return FAILED;
      output_format = OUTPUT_ROM;
      continue;

#ifndef GB
    case 'b':
      if (output_format != OUTPUT_NONE)
	return FAILED;
      output_format = OUTPUT_PROGRAM;
      continue;
#endif

    case 'c':
      compact_mode = ON;
      continue;

    case 'f':
      error_level = ERROR_LEVEL_MIN;
      continue;

    case 'v':
      verbose_mode = ON;
      continue;

    case 't':
      test_mode = ON;
      continue;

    case 'u':
      case_mode = CASE_MODE_UPPER;
      continue;

    case 's':
      symbol_mode = ON;
      continue;

    case 'x':
      extra_definitions = ON;
      continue;

    default:
      return FAILED;
    }
  }

  return SUCCEEDED;
}


void procedures_at_exit(void) {

  struct gba_macro *m;


  if (file_out_ptr != NULL)
    fclose(file_out_ptr);

  tmp_def = defines;
  while (tmp_def != NULL) {
    defines = tmp_def->next;
    free(tmp_def);
    tmp_def = defines;
  }

  m = macros_first;
  while (m != NULL) {
    macros_first = m->next;
    free(m);
    m = macros_first;
  }

  label_tmp = labels;
  while (label_tmp != NULL) {
    labels = label_tmp->next;
    free(label_tmp);
    label_tmp = labels;
  }

  export_tmp = export_first;
  while (export_tmp != NULL) {
    export_last = export_tmp->next;
    free(export_tmp);
    export_tmp = export_last;
  }

  ifd_tmp = incbin_file_data_first;
  while(ifd_tmp != NULL) {
    incbin_file_data_first = ifd_tmp->next;
    if (ifd_tmp->data != NULL)
      free(ifd_tmp->data);
    free(ifd_tmp);
    ifd_tmp = incbin_file_data_first;
  }

  stacks_tmp = stacks_first;
  while (stacks_tmp != NULL) {
    free(stacks_tmp->stack);
    stacks_first = stacks_tmp->next;
    free(stacks_tmp);
    stacks_tmp = stacks_first;
  }

  stacks_tmp = stacks_header_first;
  while (stacks_tmp != NULL) {
    free(stacks_tmp->stack);
    stacks_first = stacks_tmp->next;
    free(stacks_tmp);
    stacks_tmp = stacks_first;
  }

  if (unfolded_buffer != NULL)
    free(unfolded_buffer);

  if (include_in_tmp != NULL)
    free(include_in_tmp);
  if (tmp_a != NULL)
    free(tmp_a);

  if (rom_banks != NULL)
    free(rom_banks);
  if (rom_banks_usage_table != NULL)
    free(rom_banks_usage_table);
  if (banks != NULL)
    free(banks);
  if (bankaddress != NULL)
    free(bankaddress);

  remove(gba_tmp_name);
  remove(gba_unfolded_name);
}


int generate_tmp_names(void) {

#ifdef UNIX
  sprintf(gba_tmp_name, ".wla%da", (int)getpid());
  sprintf(gba_unfolded_name, ".wla%db", (int)getpid());
#endif

#ifdef AMIGA
  sprintf(gba_tmp_name, "wla_a.tmp");
  sprintf(gba_unfolded_name, "wla_b.tmp");
#endif

#ifdef MSDOS
#ifndef WIN32
  sprintf(gba_tmp_name, "wla_a.tmp");
  sprintf(gba_unfolded_name, "wla_b.tmp");
#else
  sprintf(gba_tmp_name, ".wla%lda", GetCurrentProcessId());
  sprintf(gba_unfolded_name, ".wla%ldb", GetCurrentProcessId());
#endif  
#endif

  return SUCCEEDED;
}


int generate_extra_definitions(void) {

  time_t t;
  char *q, c[256];
  int a, s;


  /* generate WLA_TIME */
  time(&t);
  q = ctime(&t);
  strcpy(c, q);
  /* remove the linefeed */
  s = strlen(c);
  for (a = 0; a < s; a++)
    if (c[a] == 0x0A) {
      c[a] = 0;
    }
  add_a_new_definition("WLA_TIME", 0, c, DEFINITION_TYPE_STRING, strlen(c));

  return SUCCEEDED;
}
