
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "defines.h"

#include "include_file.h"
#include "parse.h"
#include "pass_1.h"



#ifdef GB
char licenseecodenew_c1, licenseecodenew_c2;
int computechecksum_defined = 0, computecomplementcheck_defined = 0;
int romgbc = 0, romdmg = 0, romsgb = 0;
int cartridgetype = 0, cartridgetype_defined = 0;
int licenseecodenew_defined = 0, licenseecodeold = 0, licenseecodeold_defined = 0;
#endif

#ifdef Z80
char sdsctag_name_str[MAX_NAME_LENGTH], sdsctag_notes_str[MAX_NAME_LENGTH];
int sdsctag_name_type, sdsctag_notes_type, sdsc_ma, sdsc_mi;
int sdsctag_name_value, sdsctag_notes_value;
int computesmschecksum_defined = 0, sdsctag_defined = 0, smstag_defined = 0;
#endif

int org_defined = 0, background_defined = 0, background_size = 0;
int bank = 0, bank_defined = 0;
int rombanks = 0, rombanks_defined = 0, romtype = 0, max_address;
int rambanks = 0, rambanks_defined = 0;
int emptyfill, emptyfill_defined = 0;
int section_status = OFF, section_id = 1, line_count_status = ON;
int d, i, ind, inz, ifdef = 0, t, slots_amount = 0;
int memorymap_defined = 0;
int defaultslot_defined = 0, defaultslot;
int banksize_defined = 0, banksize;
int rombankmap_defined = 0, *banks = NULL, *bankaddress = NULL;
int bankheader_status = OFF;
int macro_active = 0;
int repeat_active = 0;

int unfolded_size;
char *unfolded_buffer = NULL;
FILE *fold_out;

#ifndef GB
extern struct stack *stacks_header_first, *stacks_header_last;
extern int latest_stack, stacks_inside, stacks_outside;
int stack_inserted;
int header_defined = 0, header;
int footer_defined = 0, footer;
#endif

#ifdef W65816
int hirom_defined = 0, lorom_defined = 0, slowrom_defined = 0, fastrom_defined = 0, snes_mode = 0;
#endif

#if defined GB || W65816
char name[32];
int name_defined = 0;
#endif

char tmp[4096], emsg[256];
char *tmp_bf;
char cp[256];

unsigned char *rom_banks = NULL, *rom_banks_usage_table = NULL;

struct export_def *export_first = NULL, *export_last = NULL, *export_tmp;
struct optcode *opt_tmp;
struct gba_def *defines = NULL, *tmp_def, *next_def;
struct gba_macro *macros_first = NULL, *macros_last, *macros_current;
struct section_def *sections_first = NULL, *sections_last = NULL, *sec_tmp, *sec_next;
struct macro_runtime macro_stack[MACRO_STACK_DEPTH];
struct repeat_runtime repeat_stack[REPEAT_STACK_DEPTH];
struct slot slots[256];

extern FILE *file_out_ptr;
extern char *buffer, *unfolded_buffer, label[MAX_NAME_LENGTH], include_dir[256], full_name[256];
extern int size, unfolded_size, input_number_error_msg, verbose_mode, output_format, error_level, open_files, case_mode;
extern int header, header_defined, i_e2, latest_stack, ss;
extern int i_e1, extra_definitions, string_size, flo_l, flo_h;
extern double f;
extern struct active_file_info *active_file_info_first, *active_file_info_last, *active_file_info_tmp;
extern struct file_name_info *file_name_info_first, *file_name_info_last, *file_name_info_tmp;
extern struct stack *stacks_first, *stacks_tmp, *stacks_last;

int opcode_n[256], opcode_p[256];

#ifdef MCS6502
int skip_xbit = SKIP_NO;
#endif
#ifdef MCS6510
int skip_xbit = SKIP_NO;
#endif
#ifdef W65816
int skip_xbit = SKIP_NO;
int width_index = WIDTH_8BIT, width_memory = WIDTH_8BIT;
#endif


/* remember to run opcodesgen/gen with the proper flags defined (GB/Z80/MCS6502/MCS6510/W65816) */

#ifdef AMIGA
__far /* put the following big table in the FAR data section */
#endif

#ifdef GB
#include "opcodes_gb.c"
#include "opcodes_gb_tables.c" /* opcode_n[256] and opcode_p[256] */
#endif
#ifdef Z80
#include "opcodes_z80.c"
#include "opcodes_z80_tables.c" /* opcode_n[256] and opcode_p[256] */
#endif
#ifdef MCS6502
#include "opcodes_6502.c"
#include "opcodes_6502_tables.c" /* opcode_n[256] and opcode_p[256] */
#endif
#ifdef MCS6510
#include "opcodes_6510.c"
#include "opcodes_6510_tables.c" /* opcode_n[256] and opcode_p[256] */
#endif
#ifdef W65816
#include "opcodes_65816.c"
#include "opcodes_65816_tables.c" /* opcode_n[256] and opcode_p[256] */
#endif
#ifdef C8051
#include "opcodes_8051.c"
#include "opcodes_8051_tables.c" /* opcode_n[256] and opcode_p[256] */
#endif


int pass_1(void) {

  struct gba_macro *m;
  int q, p;



  if (verbose_mode == ON)
    printf("Pass 1...\n");

  /* mark all slots as empty */
  for (q = 0; q < 256; q++)
    slots[q].size = 0;

  /* start from the very first character */
  i = 0;

  /* output the file id if requested */
  if (error_level == ERROR_LEVEL_MAX)
    fprintf(file_out_ptr, "f %d ", active_file_info_tmp->filename_id);

  while ((t = get_next_token()) == SUCCEEDED) {
    q = evaluate_token();

    if (q == SUCCEEDED)
      continue;
    else if (q == EVALUATE_TOKEN_EOP)
      return SUCCEEDED;
    else if (q == EVALUATE_TOKEN_NOT_IDENTIFIED) {

      struct gba_def *pd;


      /* check if it is of the form "LABEL:XYZ" */
      for (q = 0; q < ss; q++)
	if (tmp[q] == ':')
	  break;

      /* it is a label after all */
      if (q != ss) {

	tmp[q] = 0;

	if (org_defined == 0) {
	  sprintf(emsg, "\"%s\" needs a position in memory.\n", tmp);
	  pass_1_print_error(emsg, ERROR_LOG);
	  return FAILED;
	}
	if (ss > MAX_NAME_LENGTH - 1) {
	  sprintf(emsg, "The string \"%s\" is too long. Max string length is %d bytes.\n", tmp, MAX_NAME_LENGTH - 1);
	  pass_1_print_error(emsg, ERROR_NONE);
	  return FAILED;
	}
	if (bankheader_status == ON) {
	  pass_1_print_error("BANKHEADER sections don't take labels.\n", ERROR_LOG);
	  return FAILED;
	}

	/* check out for \@-symbols */
	if (macro_active != 0) {
	  if (tmp[q - 2] == '\\' && tmp[q - 1] == '@')
	    sprintf(&tmp[q - 2], "%d%c", macros_current->calls - 1, 0);
	}

	if (error_level == ERROR_LEVEL_MIN)
	  fprintf(file_out_ptr, "L %s ", tmp);
	else
	  fprintf(file_out_ptr, "k %d L %s ", active_file_info_last->line_current, tmp);

	/* move to the end of the label */
	i -= ss - q - 1;

	continue;
      }

      /* it must be a macro */
      m = macros_first;
      while (m != NULL) {
	if (strcmp(m->name, tmp) == 0)
	  break;
	m = m->next;
      }

      /* was it nothing we know of? */
      if (m == NULL) {
	sprintf(emsg, "Parse error at \"%s\".\n", tmp);
	pass_1_print_error(emsg, ERROR_INP);
	return FAILED;
      }

      /* run until .ENDM */
      if (macro_active == MACRO_STACK_DEPTH) {
	sprintf(emsg, "ERROR: Macro stack depth (%d macros) reached.\n", MACRO_STACK_DEPTH);
	pass_1_print_error(emsg, ERROR_NONE);
	return FAILED;
      }

      macro_active++;
      macros_current = m;
      m->calls++;

      /* collect agrument starting positions */
      for (p = 0, q = 0; 1; q++) {
	while (buffer[i] == ',' || buffer[i] == ' ')
	  i++;
	if (buffer[i] == 0x0A)
	  break;
	m->arguments_start[q] = i;
	p++;
	if (skip_next_token() == FAILED)
	  break;
      }

      m->supplied_arguments = p;
      if (p > m->arguments) {
	sprintf(emsg, "Too many arguments to macro \"%s\".\n", m->name);
	pass_1_print_error(emsg, ERROR_INP);
	return FAILED;
      }

      macro_stack[macro_active - 1].macro = m;
      macro_stack[macro_active - 1].macro_end = i;
      macro_stack[macro_active - 1].supplied_arguments = p;
      if (error_level == ERROR_LEVEL_MAX) {
	macro_stack[macro_active - 1].macro_end_line = active_file_info_last->line_current;
	macro_stack[macro_active - 1].macro_end_filename_id = active_file_info_last->filename_id;
	active_file_info_last->line_current = m->start_line;

	if ((extra_definitions == ON) && (active_file_info_last->filename_id != m->filename_id))
	  redefine("WLA_FILENAME", 0, get_file_name(m->filename_id), DEFINITION_TYPE_STRING, strlen(get_file_name(m->filename_id)));

	active_file_info_last->filename_id = m->filename_id;
      }

      i = m->start;

      pd = defines;
      while (pd != NULL) {
	if (strcmp(pd->alias, "NARGS") == 0) {
	  pd->value = p;
	  break;
	}
	pd = pd->next;
      }

      if (pd == NULL)
	if (add_a_new_definition("NARGS", p, NULL, DEFINITION_TYPE_VALUE, 0) == FAILED)
	  return FAILED;
      continue;
    }
    else if (q == FAILED)
      return FAILED;
  }

  return FAILED;
}


int evaluate_token(void) {

  int f, z, x, y;
#if defined Z80 || C8051
  int e, v, h;
  char labelx[256];
#endif
#ifdef W65816
  int e, v, h;
  char labelx[256];
#endif


  /* is it a directive? */
  if (tmp[0] == '.')
    return parse_directive();

  /* is it a label? */
  if (tmp[ss - 1] == ':') {

    tmp[ss - 1] = 0;

    if (org_defined == 0) {
      sprintf(emsg, "\"%s\" needs a position in memory.\n", tmp);
      pass_1_print_error(emsg, ERROR_LOG);
      return FAILED;
    }
    if (ss > MAX_NAME_LENGTH - 1) {
      sprintf(emsg, "The string \"%s\" is too long. Max string length is %d bytes.\n", tmp, MAX_NAME_LENGTH - 1);
      pass_1_print_error(emsg, ERROR_NONE);
      return FAILED;
    }
    if (bankheader_status == ON) {
      pass_1_print_error("BANKHEADER sections don't take labels.\n", ERROR_LOG);
      return FAILED;
    }

    /* check out for \@-symbols */
    if (macro_active != 0) {
      if (tmp[ss - 3] == '\\' && tmp[ss - 2] == '@')
	sprintf(&tmp[ss - 3], "%d%c", macros_current->calls - 1, 0);
    }

    if (error_level == ERROR_LEVEL_MIN)
      fprintf(file_out_ptr, "L %s ", tmp);
    else
      fprintf(file_out_ptr, "k %d L %s ", active_file_info_last->line_current, tmp);

    return SUCCEEDED;
  }

  /* OPCODE? */

  ind = opcode_p[(unsigned int)tmp[0]];
  opt_tmp = &opt_table[ind];

  for (f = opcode_n[(unsigned int)tmp[0]]; f > 0; f--) {

    for (inz = 0, d = SUCCEEDED; inz < OP_SIZE_MAX; inz++) {
      if (tmp[inz] == 0 && opt_tmp->op[inz] == 0 && buffer[i] == 0x0A) {
	if (opt_tmp->type == 0)
	  fprintf(file_out_ptr, "d %d ", opt_tmp->hex);
	else
	  fprintf(file_out_ptr, "y %d ", opt_tmp->hex);
	return SUCCEEDED;
      }
      if (tmp[inz] == 0 && opt_tmp->op[inz] == ' ' && buffer[i] == ' ')
	break;
      if (case_mode == CASE_MODE_INSENSITIVE) {
	if (opt_tmp->op[inz] != toupper((int)tmp[inz])) {
	  d = FAILED;
	  break;
	}
	continue;
      }
      else if (case_mode == CASE_MODE_UPPER && tmp[inz] == opt_tmp->op[inz])
	continue;

      d = FAILED;
      break;
    }

    if (d == FAILED) {
      opt_tmp = &opt_table[++ind];
      continue;
    }

    /* beginning matches the input */
    x = inz + 1;
    inz = i + 1;

#ifndef GB
    /* no stack rollback */
    stack_inserted = STACK_NONE;
#endif

    switch (opt_tmp->type) {

#ifdef GB
#include "decode_gb.c"
#endif
#ifdef Z80
#include "decode_z80.c"
#endif
#ifdef MCS6502
#include "decode_6502.c"
#endif
#ifdef MCS6510
#include "decode_6510.c"
#endif
#ifdef W65816
#include "decode_65816.c"
#endif
#ifdef C8051
#include "decode_8051.c"
#endif

    }

#ifndef GB
    /* perform stack rollback? */
    if (stack_inserted != STACK_NONE) {
      if (latest_stack == 0) {
	if (stack_inserted == STACK_OUTSIDE) {
	  delete_stack(stacks_first);
	  stacks_first = NULL;
	  stacks_last = NULL;
	  stacks_outside--;
	}
	else {
	  delete_stack(stacks_header_first);
	  stacks_header_first = NULL;
	  stacks_header_last = NULL;
	  stacks_inside--;
	}
      }
      else {
	struct stack *s;
	int y;


	if (stack_inserted == STACK_OUTSIDE) {
	  s = stacks_first;
	  stacks_outside--;

	  for (y = 0; y < stacks_outside - 1; y++)
	    s = s->next;

	  delete_stack(s->next);
	  s->next = NULL;
	  stacks_last = s;
	  if (stacks_outside == 1)
	    stacks_first = s;
	}
	else {
	  s = stacks_header_first;
	  stacks_inside--;

	  for (y = 0; y < stacks_inside - 1; y++)
	    s = s->next;

	  delete_stack(s->next);
	  s->next = NULL;
	  stacks_header_last = s;
	  if (stacks_inside == 1)
	    stacks_header_first = s;
	}
      }
    }
#endif

    opt_tmp = &opt_table[++ind];
  }

  input_number_error_msg = 0; /* allow error messages from input_numbers() */

  return EVALUATE_TOKEN_NOT_IDENTIFIED;
}


int redefine(char *name, int value, char *string, int type, int size) {

  struct gba_def *p;



  p = defines;
  while (p != NULL) {
    if (strcmp(p->alias, name) == 0)
      break;
    p = p->next;
  }

  /* it wasn't defined previously */
  if (p == NULL) {
    add_a_new_definition(name, value, string, type, size);
    return SUCCEEDED;
  }

  if (type == DEFINITION_TYPE_VALUE) {
    p->type = type;
    p->value = value;
  }
  else {
    p->type = type;
    strncpy(p->string, string, size);
  }

  return SUCCEEDED;
}


int add_a_new_definition(char *name, int value, char *string, int type, int size) {

  struct gba_def *d, *l = NULL;



  d = defines;
  while (d != NULL) {
    l = d;
    if (strcmp(name, d->alias) == 0) {
      sprintf(emsg, "\"%s\" was defined for the second time.\n", name);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }
    d = d->next;
  }

  d = malloc(sizeof(struct gba_def));
  if (d == NULL) {
    sprintf(emsg, "Out of memory while trying to add a new definition (\"%s\").\n", name);
    pass_1_print_error(emsg, ERROR_DIR);
    return FAILED;
  }

  if (defines == NULL)
    defines = d;
  else
    l->next = d;

  strcpy(d->alias, name);
  d->next = NULL;
  d->type = type;

  if (type == DEFINITION_TYPE_VALUE)
    d->value = value;
  else {
    memcpy(d->string, string, size);
    d->string[size] = 0;
    d->size = size;
  }

  return SUCCEEDED;
}


void pass_1_print_error(char *error, int i) {

  char error_dir[] = "DIRECTIVE_ERROR:";
  char error_unf[] = "UNFOLD_ALIASES:";
  char error_num[] = "INPUT_NUMBER:";
  char error_inc[] = "INCLUDE_FILE:";
  char error_inb[] = "INCBIN_FILE:";
  char error_inp[] = "INPUT_ERROR:";
  char error_log[] = "LOGIC_ERROR:";
  char error_stc[] = "STACK_CALCULATE:";
  char error_wrn[] = "WARNING:";
  char *t = NULL;


  switch (i) {
  case ERROR_LOG:
    t = error_log;
    break;
  case ERROR_UNF:
    t = error_unf;
    break;
  case ERROR_INC:
    t = error_inc;
    break;
  case ERROR_INB:
    t = error_inb;
    break;
  case ERROR_DIR:
    t = error_dir;
    break;
  case ERROR_INP:
    t = error_inp;
    break;
  case ERROR_NUM:
    t = error_num;
    break;
  case ERROR_STC:
    t = error_stc;
    break;
  case ERROR_WRN:
    t = error_wrn;
    break;
  case ERROR_NONE:
    if (error_level == ERROR_LEVEL_MIN)
      fprintf(stderr, "%s", error);
    else
      fprintf(stderr, "%s:%d: %s", get_file_name(active_file_info_last->filename_id), active_file_info_last->line_current, error);
    return;
  }

  if (error_level == ERROR_LEVEL_MIN)
    fprintf(stderr, "%s %s", t, error);
  else
    fprintf(stderr, "%s:%d: %s %s", get_file_name(active_file_info_last->filename_id), active_file_info_last->line_current, t, error);
  return;
}


void next_line(void) {

  if (line_count_status == OFF)
    return;

  if (active_file_info_last != NULL) {
    active_file_info_last->line_current++;
  }
}


int parse_directive(void) {

  int o, q;



  /* ORG */

  if (strcmp(cp, "ORG") == 0) {
    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take ORG definitions.\n", ERROR_DIR);
      return FAILED;
    }
    if (bank_defined == 0) {
      pass_1_print_error("No BANK is defined.\n", ERROR_LOG);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;

    if (q != SUCCEEDED) {
      pass_1_print_error("ORG needs a positive integer value.\n", ERROR_DIR);
      return FAILED;
    }

    org_defined = 1;
    if (output_format != OUTPUT_LIBRARY)
      fprintf(file_out_ptr, "O %d ", d);
    return SUCCEEDED;
  }

  /* BANK */

  if (strcmp(cp, "BANK") == 0) {
    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take BANK definitions.\n", ERROR_DIR);
      return FAILED;
    }
    if (section_status == ON) {
      sprintf(emsg, "Section \"%s\" is open. Do not try to change the bank.\n", sections_last->name);
      pass_1_print_error(emsg, ERROR_LOG);
      return FAILED;
    }
    if (rombanks_defined == 0 && output_format != OUTPUT_LIBRARY) {
      pass_1_print_error("ROMBANKS is not yet defined.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || d < 0) {
      pass_1_print_error("BANK number must be zero or positive.\n", ERROR_DIR);
      return FAILED;
    }

    if (rombanks <= d && output_format != OUTPUT_LIBRARY) {
      sprintf(emsg, "ROM banks == %d, selected bank %d.\n", rombanks, d);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    bank = d;
    bank_defined = 1;

    if (compare_next_token("SLOT", 4) == SUCCEEDED) {
      skip_next_token();

      q = input_number();

      if (q == FAILED)
	return FAILED;
      if (q != SUCCEEDED || d > 255 || d < 0) {
	pass_1_print_error("SLOT needs an unsigned 8bit value as an ID.\n", ERROR_DIR);
	return FAILED;
      }

      if (slots[d].size == 0) {
	sprintf(emsg, "There is no SLOT number %d.\n", d);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }

      if (output_format != OUTPUT_LIBRARY)
	fprintf(file_out_ptr, "B %d %d ", bank, d);

      ind = bank;
      inz = d;
    }
    else if (output_format != OUTPUT_LIBRARY) {
      fprintf(file_out_ptr, "B %d %d ", d, defaultslot);
      ind = d;
      inz = defaultslot;
    }

    if (slots[inz].size < banks[ind]) {
      sprintf(emsg, "SLOT %d's size %d < BANK %d's size %d.\n", inz, slots[inz].size, ind, banks[ind]);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }
    if (slots[inz].size > banks[ind]) {
      sprintf(emsg, "SLOT %d's size %d > BANK %d's size %d, but the bank fits inside.\n", inz, slots[inz].size, ind, banks[ind]);
      pass_1_print_error(emsg, ERROR_WRN);
    }

    return SUCCEEDED;
  }

  /* DB? */

  if (strcmp(cp, "DB") == 0) {

    inz = input_number();
    for (ind = 0; inz == SUCCEEDED || inz == INPUT_NUMBER_STRING || inz == INPUT_NUMBER_ADDRESS_LABEL || inz == INPUT_NUMBER_STACK; ind++) {

      if (inz == INPUT_NUMBER_STRING) {
	for (o = 0; o < string_size; o++)
	  fprintf(file_out_ptr, "d %d ", (int)label[o]);
	inz = input_number();
	continue;
      }

      if (inz == SUCCEEDED && (d < -127 || d > 255)) {
	sprintf(emsg, "DB expects 8bit data, %d is out of range!\n", d);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }

      if (inz == SUCCEEDED)
	fprintf(file_out_ptr, "d %d ", d);
      else if (inz == INPUT_NUMBER_ADDRESS_LABEL && error_level == ERROR_LEVEL_MAX)
	fprintf(file_out_ptr, "k %d Q %s ", active_file_info_last->line_current, label);
      else if (inz == INPUT_NUMBER_ADDRESS_LABEL)
	fprintf(file_out_ptr, "Q %s ", label);
      else if (inz == INPUT_NUMBER_STACK)
	fprintf(file_out_ptr, "c %d ", latest_stack);

      inz = input_number();
    }

    if (inz == FAILED)
      return FAILED;

    if (inz == INPUT_NUMBER_EOL && ind == 0) {
      pass_1_print_error("DB needs data.\n", ERROR_INP);
      return FAILED;
    }

    if (error_level == ERROR_LEVEL_MAX && inz == INPUT_NUMBER_EOL)
      next_line();

    return SUCCEEDED;
  }

  /* DW? */

  if (strcmp(cp, "DW") == 0) {

    inz = input_number();
    for (ind = 0; inz == SUCCEEDED || inz == INPUT_NUMBER_ADDRESS_LABEL || inz == INPUT_NUMBER_STACK; ind++) {

      if (inz == SUCCEEDED && (d < -32768 || d > 65535)) {
	sprintf(emsg, "DW expects 16bit data, %d is out of range!\n", d);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }

      if (inz == SUCCEEDED)
	fprintf(file_out_ptr, "y %d", d);
      else if (inz == INPUT_NUMBER_ADDRESS_LABEL && error_level == ERROR_LEVEL_MAX)
	fprintf(file_out_ptr, "k %d r %s ", active_file_info_last->line_current, label);
      else if (inz == INPUT_NUMBER_ADDRESS_LABEL)
	fprintf(file_out_ptr, "r %s ", label);
      else if (inz == INPUT_NUMBER_STACK)
	fprintf(file_out_ptr, "C %d ", latest_stack);

      inz = input_number();
    }

    if (inz == FAILED)
      return FAILED;

    if ((inz == INPUT_NUMBER_EOL || inz == INPUT_NUMBER_STRING) && ind == 0) {
      pass_1_print_error("DW needs data.\n", ERROR_INP);
      return FAILED;
    }

    if (error_level == ERROR_LEVEL_MAX && inz == INPUT_NUMBER_EOL)
      next_line();

    return SUCCEEDED;
  }

  /* DSB? */

  if (strcmp(cp, "DSB") == 0) {

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error("DSB needs size.\n", ERROR_INP);
      return FAILED;
    }

    if (d < 1 || d > 65535) {
      sprintf(emsg, "DSB expects a 16bit positive integer as size, %d is out of range!\n", d);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    inz = d;

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (!(q == SUCCEEDED || q == INPUT_NUMBER_ADDRESS_LABEL || q == INPUT_NUMBER_STACK)) {
      pass_1_print_error("DSB needs data.\n", ERROR_INP);
      return FAILED;
    }

    if (q == SUCCEEDED && (d > 255 || d < -127)) {
      sprintf(emsg, "DSB expects 8bit data, %d is out of range!\n", d);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED)
      fprintf(file_out_ptr, "x %d %d ", inz, d);
    else if (q == INPUT_NUMBER_ADDRESS_LABEL && error_level == ERROR_LEVEL_MIN) {
      for (q = 0; q < inz; q++)
	fprintf(file_out_ptr, "R %s ", label);
    }
    else if (q == INPUT_NUMBER_ADDRESS_LABEL) {
      fprintf(file_out_ptr, "k %d ", active_file_info_last->line_current);
      for (q = 0; q < inz; q++)
	fprintf(file_out_ptr, "R %s ", label);
    }
    else if (q == INPUT_NUMBER_STACK) {
      for (q = 0; q < inz; q++)
	fprintf(file_out_ptr, "c %d ", latest_stack);
    }

    return SUCCEEDED;
  }

  /* DSW? */

  if (strcmp(cp, "DSW") == 0) {

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error("DSW needs size.\n", ERROR_INP);
      return FAILED;
    }

    if (d < 1 || d > 65535) {
      sprintf(emsg, "DSW expects a 16bit positive integer as size, %d is out of range!\n", d);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    inz = d;

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (!(q == SUCCEEDED || q == INPUT_NUMBER_ADDRESS_LABEL || q == INPUT_NUMBER_STACK)) {
      pass_1_print_error("DSW needs data.\n", ERROR_INP);
      return FAILED;
    }

    if (q == SUCCEEDED && (d < -32768 || d > 65535)) {
      sprintf(emsg, "DSW expects 16bit data, %d is out of range!\n", d);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED)
      fprintf(file_out_ptr, "X %d %d ", inz, d);
    else if (q == INPUT_NUMBER_ADDRESS_LABEL && error_level == ERROR_LEVEL_MIN) {
      for (q = 0; q < inz; q++)
	fprintf(file_out_ptr, "r %s ", label);
    }
    else if (q == INPUT_NUMBER_ADDRESS_LABEL) {
      fprintf(file_out_ptr, "k %d ", active_file_info_last->line_current);
      for (q = 0; q < inz; q++)
	fprintf(file_out_ptr, "r %s ", label);
    }
    else if (q == INPUT_NUMBER_STACK) {
      for (q = 0; q < inz; q++)
	fprintf(file_out_ptr, "C %d ", latest_stack);
    }

    return SUCCEEDED;
  }

  /* EVEN */

  if (strcmp(cp, "EVEN") == 0) {
    fprintf(file_out_ptr, "a 2 ");
    return SUCCEEDED;
  }

  /* ALIGN */

  if (strcmp(cp, "ALIGN") == 0) {

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error("ALIGN needs size.\n", ERROR_INP);
      return FAILED;
    }

    fprintf(file_out_ptr, "a %d ", d);

    return SUCCEEDED;
  }

  /* INCDIR */

  if (strcmp(cp, "INCDIR") == 0) {
    if (get_next_token() != GET_NEXT_TOKEN_STRING) {
      pass_1_print_error("INCDIR needs a directory string.\n", ERROR_DIR);
      return FAILED;
    }

    q = ss;
    if (q == 0) {
      include_dir[0] = 0;
      return SUCCEEDED;
    }
    strcpy(include_dir, tmp);

#ifdef MSDOS
    if (include_dir[q - 1] != '\\') {
      include_dir[q] = '\\';
      include_dir[q + 1] = 0;
    }
#else
    if (include_dir[q - 1] != '/') {
      include_dir[q] = '/';
      include_dir[q + 1] = 0;
    }
#endif

    return SUCCEEDED;
  }

  /* INCLUDE */

  if (strcmp(cp, "INCLUDE") == 0) {

    inz = input_number();

    if (inz != INPUT_NUMBER_STRING) {
      pass_1_print_error("INCLUDE needs a file name string.\n", ERROR_DIR);
      return FAILED;
    }
    if (macro_active != 0) {
      pass_1_print_error("You cannot include a file inside a MACRO.\n", ERROR_DIR);
      return FAILED;
    }

    if (include_file(label) == FAILED)
      return FAILED;

    if (error_level == ERROR_LEVEL_MAX)
      fprintf(file_out_ptr, "f %d ", active_file_info_tmp->filename_id);

    return SUCCEEDED;
  }

  /* INCBIN */

  if (strcmp(cp, "INCBIN") == 0) {

    int s, r;


    if (org_defined == 0 && output_format != OUTPUT_LIBRARY) {
      pass_1_print_error("Before you can INCBIN data you'll need to use ORG.\n", ERROR_LOG);
      return FAILED;
    }

    inz = input_number();

    if (inz != INPUT_NUMBER_STRING) {
      pass_1_print_error("INCBIN needs a file name string.\n", ERROR_DIR);
      return FAILED;
    }

    if (incbin_file(label, &ind, &inz, &s, &r) == FAILED)
      return FAILED;

    /* D [id] [swap] [skip] [size] */
    fprintf(file_out_ptr, "D %d %d %d %d ", ind, inz, s, r);

    return SUCCEEDED;
  }

  /* SECTION */

  if (strcmp(cp, "SECTION") == 0) {

    int l;


    if (section_id > 255) {
      pass_1_print_error("Out of section numbers. Please start a new file.\n", ERROR_DIR);
      return FAILED;
    }

    if (section_status == ON) {
      sprintf(emsg, "There is already an open section called \"%s\".", sections_last->name);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }
    else if (output_format != OUTPUT_LIBRARY && bank_defined == 0) {
      pass_1_print_error(".SECTION requires a predefined bank.\n", ERROR_DIR);
      return FAILED;
    }
    else if (output_format != OUTPUT_LIBRARY && org_defined == 0) {
      pass_1_print_error(".SECTION requires a starting address for positioning.\n", ERROR_DIR);
      return FAILED;
    }

    if (get_next_token() == FAILED)
      return FAILED;

    /* every library section starts @ the beginning of the bank */
    if (output_format == OUTPUT_LIBRARY)
      org_defined = 1;

    sec_tmp = malloc(sizeof(struct section_def));
    if (sec_tmp == NULL) {
      sprintf(emsg, "Out of memory while allocating room for a new SECTION \"%s\".\n", tmp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    sec_tmp->maxsize_status = OFF;

    /* check if the section size is supplied inside the name */
    l = strlen(tmp) - 1;

    for (; l >= 0 && tmp[l] != '_'; l--);

    if (tmp[l] == '_') {

      int m, n;


      n = l;
      l++;
      if (tmp[l] == '$') {
	for (l++, m = 0; tmp[l] != 0; l++) {
	  if (tmp[l] >= '0' && tmp[l] <= '9')
	    m = (m << 4) + tmp[l] - '0';
	  else if (tmp[l] >= 'a' && tmp[l] <= 'f')
	    m = (m << 4) + tmp[l] - 'a' + 10;
	  else if (tmp[l] >= 'A' && tmp[l] <= 'F')
	    m = (m << 4) + tmp[l] - 'A' + 10;
	  else
	    break;
	}
	  
	if (tmp[l] == 0) {
	  tmp[n] = 0; /* strip the size from the name */
	  sec_tmp->maxsize_status = ON;
	  sec_tmp->maxsize = m;
	}
      }
      else if (tmp[l] >= '0' && tmp[l] <= '9') {
	for (m = 0; tmp[l] != 0; l++) {
	  if (tmp[l] >= '0' && tmp[l] <= '9')
	    m = (m * 10) + tmp[l] - '0';
	  else
	    break;
	}

	if (tmp[l] == 0) {
	  tmp[n] = 0; /* strip the size from the name */
	  sec_tmp->maxsize_status = ON;
	  sec_tmp->maxsize = m;
	}
      }
    }

    if (strcmp(tmp, "BANKHEADER") == 0) {
      if (output_format == OUTPUT_LIBRARY) {
	pass_1_print_error("Library files don't take bank header sections.\n", ERROR_DIR);
	return FAILED;
      }
      sec_next = sections_first;
      while (sec_next != NULL) {
	if (strcmp(sec_next->name, tmp) == 0 && sec_next->bank == bank) {
	  sprintf(emsg, "BANKHEADER section was defined for the second time for bank %d.\n", bank);
	  pass_1_print_error(emsg, ERROR_DIR);
	  return FAILED;
	}
	sec_next = sec_next->next;
      }
    }
    else {
      sec_next = sections_first;
      while (sec_next != NULL) {
	if (strcmp(sec_next->name, tmp) == 0) {
	  sprintf(emsg, "SECTION \"%s\" was defined for the second time.\n", tmp);
	  pass_1_print_error(emsg, ERROR_DIR);
	  return FAILED;
	}
	sec_next = sec_next->next;
      }
    }

    strcpy(sec_tmp->name, tmp);
    sec_tmp->next = NULL;

    if (sections_first == NULL) {
      sections_first = sec_tmp;
      sections_last = sec_tmp;
    }
    else {
      sections_last->next = sec_tmp;
      sections_last = sec_tmp;
    }

    if (compare_next_token("FORCE", 5) == SUCCEEDED) {
      if (output_format == OUTPUT_LIBRARY) {
	pass_1_print_error("Libraries don't take FORCE sections.\n", ERROR_DIR);
	return FAILED;
      }
      sec_tmp->status = SECTION_STATUS_FORCE;
      if (skip_next_token() == FAILED)
	return FAILED;
    }
    else if (compare_next_token("FREE", 4) == SUCCEEDED) {
      sec_tmp->status = SECTION_STATUS_FREE;
      if (skip_next_token() == FAILED)
	return FAILED;
    }
    else if (compare_next_token("SEMIFREE", 8) == SUCCEEDED) {
      if (output_format == OUTPUT_LIBRARY) {
	pass_1_print_error("Libraries don't take SEMIFREE sections.\n", ERROR_DIR);
	return FAILED;
      }
      sec_tmp->status = SECTION_STATUS_SEMIFREE;
      if (skip_next_token() == FAILED)
	return FAILED;
    }
    else if (compare_next_token("OVERWRITE", 9) == SUCCEEDED) {
      if (output_format == OUTPUT_LIBRARY) {
	pass_1_print_error("Libraries don't take OVERWRITE sections.\n", ERROR_DIR);
	return FAILED;
      }
      sec_tmp->status = SECTION_STATUS_OVERWRITE;
      if (skip_next_token() == FAILED)
	return FAILED;
    }
    else
      sec_tmp->status = SECTION_STATUS_FREE;

    if (strcmp(sec_tmp->name, "BANKHEADER") == 0) {
      sec_tmp->status = SECTION_STATUS_HEADER;
      bankheader_status = ON;
    }

    sec_tmp->id = section_id;
    sec_tmp->bank = bank;
    section_id++;
    section_status = ON;
    fprintf(file_out_ptr, "S %d ", sec_tmp->id);
    return SUCCEEDED;
  }

  /* ELSE */

  if (strcmp(cp, "ELSE") == 0) {

    int m, r;


    if (ifdef == 0) {
      pass_1_print_error("There must be IFxxx before ELSE.\n", ERROR_DIR);
      return FAILED;
    }

    /* find the next compiling point */
    r = 1;
    m = macro_active;
    /* disable macro decoding */
    macro_active = 0;
    while (get_next_token() != FAILED) {
      if (tmp[0] == '.') {
	if (strcmp(cp, "ENDIF") == 0)
	  r--;
	if (strcmp(cp, "E") == 0)
	  break;
	if (strcmp(cp, "IFDEF") == 0 || strcmp(cp, "IFNDEF") == 0 || strcmp(cp, "IFGR") == 0 || strcmp(cp, "IFLE") == 0 || strcmp(cp, "IFEQ") == 0 ||
	    strcmp(cp, "IFNEQ") == 0 || strcmp(cp, "IFDEFM") == 0 || strcmp(cp, "IFNDEFM") == 0 || strcmp(cp, "IF") == 0 || strcmp(cp, "IFEXISTS") == 0 ||
	    strcmp(cp, "IFGREQ") == 0 || strcmp(cp, "IFLEEQ") == 0)
	  r++;
      }
      if (r == 0) {
	ifdef--;
	macro_active = m;
	return SUCCEEDED;
      }
    }

    pass_1_print_error("ELSE must end to ENDIF.\n", ERROR_DIR);
    return FAILED;
  }

  /* ENDIF */

  if (strcmp(cp, "ENDIF") == 0) {

    if (ifdef == 0) {
      pass_1_print_error("ENDIF was given before any IF directive.\n", ERROR_DIR);
      return FAILED;
    }

    ifdef--;
    return SUCCEEDED;
  }

  /* IFDEF */

  if (strcmp(cp, "IFDEF") == 0) {

    struct gba_def *d;


    if (get_next_token() == FAILED)
      return FAILED;

    d = defines;
    while (d != NULL) {
      if (strcmp(tmp, d->alias) == 0) {
	ifdef++;
	return SUCCEEDED;
      }
      d = d->next;
    }

    return find_next_point("IFDEF");
  }

  /* IF */

  if (strcmp(cp, "IF") == 0) {

    int a, o;


    q = input_number();
    if (q != SUCCEEDED) {
      pass_1_print_error("IF needs immediate data.\n", ERROR_INP);
      return FAILED;
    }

    a = d;

    if (get_next_token() == FAILED)
      return FAILED;

    if (strcmp(tmp, "<") == 0)
      o = 0;
    else if (strcmp(tmp, ">") == 0)
      o = 1;
    else if (strcmp(tmp, "==") == 0)
      o = 2;
    else if (strcmp(tmp, "!=") == 0)
      o = 3;
    else if (strcmp(tmp, ">=") == 0)
      o = 4;
    else if (strcmp(tmp, "<=") == 0)
      o = 5;
    else {
      pass_1_print_error("IF needs an operator. Supported operators are '<', '>', '>=', '<=', '!=' and '=='.\n", ERROR_INP);
      return FAILED;
    }

    q = input_number();
    if (q != SUCCEEDED) {
      pass_1_print_error("IF needs immediate data.\n", ERROR_INP);
      return FAILED;
    }

    if ((o == 0 && a < d) || (o == 1 && a > d) || (o == 2 && a == d) || (o == 3 && a != d) || (o == 4 && a >= d) || (o == 5 && a <= d)) {
      ifdef++;
      return SUCCEEDED;
    }
    else
      return find_next_point("IF");
  }

  /* IFGR/IFLE/IFEQ/IFNEQ/IFGREQ/IFLEEQ */

  if (strcmp(cp, "IFGR") == 0 || strcmp(cp, "IFLE") == 0 || strcmp(cp, "IFEQ") == 0 || strcmp(cp, "IFNEQ") == 0 ||
      strcmp(cp, "IFGREQ") == 0 || strcmp(cp, "IFLEEQ") == 0) {

    char emsg[256];
    int y, o;


    if (strcmp(&cp[2], "GR") == 0)
      o = 0;
    else if (strcmp(&cp[2], "LE") == 0)
      o = 1;
    else if (strcmp(&cp[2], "EQ") == 0)
      o = 2;
    else if (strcmp(&cp[2], "NEQ") == 0)
      o = 3;
    else if (strcmp(&cp[2], "GREQ") == 0)
      o = 4;
    else
      o = 5;

    q = input_number();
    if (q != SUCCEEDED) {
      sprintf(emsg, "%s needs immediate data.\n", cp);
      pass_1_print_error(emsg, ERROR_INP);
      return FAILED;
    }

    y = d;

    q = input_number();
    if (q != SUCCEEDED) {
      sprintf(emsg, "%s needs immediate data.\n", cp);
      pass_1_print_error(emsg, ERROR_INP);
      return FAILED;
    }

    if ((o == 0 && y > d) || (o == 1 && y < d) || (o == 2 && y == d) || (o == 3 && y != d) || (o == 4 && y >= d) || (o == 5 && y <= d)) {
      ifdef++;
      return SUCCEEDED;
    }
    else {
      strcpy(emsg, cp);
      return find_next_point(emsg);
    }
  }

  /* IFEXISTS */

  if (strcmp(cp, "IFEXISTS") == 0) {

    FILE *f;


    inz = input_number();

    if (inz != INPUT_NUMBER_STRING) {
      pass_1_print_error("IFEXISTS needs a file name string.\n", ERROR_DIR);
      return FAILED;
    }

    f = fopen(label, "r");
    if (f == NULL)
      return find_next_point("IFEXISTS");

    fclose(f);
    ifdef++;
    return SUCCEEDED;
  }

  /* IFNDEF */

  if (strcmp(cp, "IFNDEF") == 0) {

    struct gba_def *d;
    char emsg[256];


    if (get_next_token() == FAILED)
      return FAILED;

    d = defines;
    while (d != NULL) {
      if (strcmp(tmp, d->alias) == 0) {
	strcpy(emsg, cp);
	return find_next_point(emsg);
      }
      d = d->next;
    }

    ifdef++;
    return SUCCEEDED;
  }

  /* IFDEFM/IFNDEFM */

  if (strcmp(cp, "IFDEFM") == 0 || strcmp(cp, "IFNDEFM") == 0) {

    char e, emsg[256];
    int k, o;


    if (cp[2] == 'N')
      o = 0;
    else
      o = 1;

    if (macro_active == 0) {
      sprintf(emsg, "%s can be only used inside a macro.\n", cp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    for (; i < size; i++) {

      if (buffer[i] == 0x0A)
	break;
      else if (buffer[i] == '\\') {
	e = buffer[++i];
	if (e >= '0' && e <= '9') {
	  d = (e - '0') * 10;
	  for (k = 2; k < 8; k++, d *= 10) {
	    e = buffer[++i];
	    if (e >= '0' && e <= '9')
	      d += e - '0';
	    else
	      break;
	  }

	  d /= 10;
	  if ((o == 0 && macros_current->supplied_arguments < d) ||
	      (o == 1 && macros_current->supplied_arguments >= d)) {
	    ifdef++;
	    return SUCCEEDED;
	  }
	  else {
	    strcpy(emsg, cp);
	    return find_next_point(emsg);
	  }
	}
	break;
      }
    }

    sprintf(emsg, "%s needs an argument.\n", cp);
    pass_1_print_error(emsg, ERROR_DIR);
    return FAILED;
  }

#ifdef GB
  /* NAME */

  if (strcmp(cp, "NAME") == 0) {
    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take NAME.\n", ERROR_DIR);
      return FAILED;
    }

    if ((ind = get_next_token()) == FAILED)
      return FAILED;

    if (ind != GET_NEXT_TOKEN_STRING) {
      pass_1_print_error("NAME requires a string of 1 to 16 letters.\n", ERROR_DIR);
      return FAILED;
    }

    /* no name has been defined so far */
    if (name_defined == 0) {
      for (ind = 0; tmp[ind] != 0 && ind < 16; ind++)
	name[ind] = tmp[ind];

      if (ind == 16 && tmp[ind] != 0) {
	pass_1_print_error("NAME requires a string of 1 to 16 letters.\n", ERROR_DIR);
	return FAILED;
      }

      for ( ; ind < 16; name[ind] = 0, ind++);

      name_defined = 1;
    }
    /* compare the names */
    else {

      for (ind = 0; tmp[ind] != 0 && name[ind] != 0 && ind < 16; ind++)
	if (name[ind] != tmp[ind])
	  break;

      if (ind == 16 && tmp[ind] != 0) {
	pass_1_print_error("NAME requires a string of 1 to 16 letters.\n", ERROR_DIR);
	return FAILED;
      }
      if (ind != 16 && (name[ind] != 0 || tmp[ind] != 0)) {
	pass_1_print_error("NAME was already defined.\n", ERROR_DIR);
	return FAILED;
      }
    }

    return SUCCEEDED;
  }
#endif

#ifdef W65816
  /* NAME */

  if (strcmp(cp, "NAME") == 0) {
    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .NAME.\n", ERROR_DIR);
      return FAILED;
    }

    if ((ind = get_next_token()) == FAILED)
      return FAILED;

    if (ind != GET_NEXT_TOKEN_STRING) {
      pass_1_print_error(".NAME requires a string of 1 to 21 letters.\n", ERROR_DIR);
      return FAILED;
    }

    /* no name has been defined so far */
    if (name_defined == 0) {

      for (ind = 0; tmp[ind] != 0 && ind < 21; ind++)
	name[ind] = tmp[ind];

      if (ind == 21 && tmp[ind] != 0) {
	pass_1_print_error(".NAME requires a string of 1 to 21 letters.\n", ERROR_DIR);
	return FAILED;
      }

      for ( ; ind < 21; name[ind] = 0, ind++);

      name_defined = 1;
    }
    /* compare the names */
    else {

      for (ind = 0; tmp[ind] != 0 && name[ind] != 0 && ind < 21; ind++)
	if (name[ind] != tmp[ind])
	  break;

      if (ind == 21 && tmp[ind] != 0) {
	pass_1_print_error(".NAME requires a string of 1 to 21 letters.\n", ERROR_DIR);
	return FAILED;
      }
      if (ind != 21 && (name[ind] != 0 || tmp[ind] != 0)) {
	pass_1_print_error(".NAME was already defined.\n", ERROR_DIR);
	return FAILED;
      }
    }

    return SUCCEEDED;
  }
#endif

  /* ENDS */

  if (strcmp(cp, "ENDS") == 0) {

    if (section_status == OFF) {
      pass_1_print_error("There is no open section for closing.\n", ERROR_DIR);
      return FAILED;
    }

    section_status = OFF;
    bankheader_status = OFF;
    fprintf(file_out_ptr, "s ");
    return SUCCEEDED;
  }

  /* ROMBANKS */

  if (strcmp(cp, "ROMBANKS") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take ROMBANKS.\n", ERROR_DIR);
      return FAILED;
    }
    if (banksize_defined == 0) {
      pass_1_print_error("No ROMBANKSIZE defined.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || d < 1) {
      pass_1_print_error("ROMBANKS needs a positive integer value.\n", ERROR_DIR);
      return FAILED;
    }

#ifdef GB
    if (d != 2 && d != 4 && d != 8 && d != 16 && d != 32 && d != 64 &&
	d != 128 && d != 72 && d != 80 && d != 96) {
      pass_1_print_error("Unsupported amount of ROM banks.\n", ERROR_DIR);
      return FAILED;
    }

    if (d == 2)
      romtype = 0;
    else if (d == 4)
      romtype = 1;
    else if (d == 8)
      romtype = 2;
    else if (d == 16)
      romtype = 3;
    else if (d == 32)
      romtype = 4;
    else if (d == 64)
      romtype = 5;
    else if (d == 128)
      romtype = 6;
    else if (d == 72)
      romtype = 0x52;
    else if (d == 80)
      romtype = 0x53;
    else if (d == 96)
      romtype = 0x54;
#endif

    /* check that the old bank map (smaller) and the new one equal as much as they can */
    if (rombanks_defined != 0) {
      if (rombanks < d)
	inz = rombanks;
      else
	inz = d;

      for (ind = 0; ind < inz; ind++) {
	if (banks[ind] != banksize) {
	  pass_1_print_error("The old and the new ROMBANKMAPs don't match.\n", ERROR_DIR);
	  return FAILED;
	}
      }

      if (d <= rombanks)
	return SUCCEEDED;

      sprintf(emsg, "Upgrading from %d to %d ROM banks.\n", rombanks, d);
      pass_1_print_error(emsg, ERROR_WRN);
    }

    rombanks = d;
    rombanks_defined = 1;
    max_address = d * banksize;

    if (rom_banks != NULL)
      free(rom_banks);
    if (rom_banks_usage_table != NULL)
      free(rom_banks_usage_table);

    rom_banks = malloc(sizeof(unsigned char) * max_address);
    rom_banks_usage_table = malloc(sizeof(unsigned char) * max_address);
    if (rom_banks == NULL || rom_banks_usage_table == NULL) {
      pass_1_print_error("Out of memory while allocating ROM banks.\n", ERROR_DIR);
      return FAILED;
    }

    memset(rom_banks_usage_table, 0, sizeof(unsigned char) * max_address);

    if (banks != NULL)
      free(banks);
    if (bankaddress != NULL)
      free(bankaddress);

    banks = malloc(sizeof(int) * rombanks);
    bankaddress = malloc(sizeof(int) * rombanks);
    if (banks == NULL || bankaddress == NULL) {
      pass_1_print_error("Out of memory while allocating ROM banks.\n", ERROR_DIR);
      return FAILED;
    }

    for (inz = 0, ind = 0; ind < d; ind++) {
      banks[ind] = banksize;
      bankaddress[ind] = inz;
      inz += banksize;
    }

    return SUCCEEDED;
  }

  /* ROMBANKMAP */

  if (strcmp(cp, "ROMBANKMAP") == 0) {

    int b = 0, a = 0, bt = 0, bt_defined = 0, x, bs = 0, bs_defined = 0;


    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take ROMBANKMAP.\n", ERROR_DIR);
      return FAILED;
    }

    /* ROMBANKMAP has been defined previously */
    if (rombankmap_defined != 0 || rombanks_defined != 0) {
      o = 0;
      while ((ind = get_next_token()) == SUCCEEDED) {

	if (strcmp(tmp, ".ENDRO") == 0) {
	  break;
	}
	else if (strcmp(tmp, "BANKSTOTAL") == 0) {

	  q = input_number();

	  if (q == FAILED)
	    return FAILED;
	  if (q != SUCCEEDED || d <= 0) {
	    pass_1_print_error("BANKSTOTAL needs a positive value.\n", ERROR_DIR);
	    return FAILED;
	  }

	  if (rombanks < d) {

	    banks = realloc(banks, sizeof(int) * d);
	    bankaddress = realloc(bankaddress, sizeof(int) * d);
	    if (banks == NULL || bankaddress == NULL) {
	      pass_1_print_error("Out of memory while allocating ROM banks.\n", ERROR_DIR);
	      return FAILED;
	    }
	  }

	  bt = d;
	  bt_defined = 1;
	}
	else if (strcmp(tmp, "BANKSIZE") == 0) {

	  if (bt_defined == 0) {
	    pass_1_print_error("BANKSTOTAL needs to be defined prior to BANKSIZE.\n", ERROR_DIR);
	    return FAILED;
	  }

	  q = input_number();

	  if (q == FAILED)
	    return FAILED;
	  if (q != SUCCEEDED || d <= 0) {
	    pass_1_print_error("BANKSIZE needs a positive value.\n", ERROR_DIR);
	    return FAILED;
	  }

	  bs = d;
	  bs_defined = 1;
	}
	else if (strcmp(tmp, "BANKS") == 0) {

	  if (bs_defined == 0) {
	    pass_1_print_error("BANKSIZE needs to be defined prior to BANKS.\n", ERROR_DIR);
	    return FAILED;
	  }

	  q = input_number();

	  if (q == FAILED)
	    return FAILED;
	  if (q != SUCCEEDED || d <= 0) {
	    pass_1_print_error("BANKS needs a positive value.\n", ERROR_DIR);
	    return FAILED;
	  }

	  for (x = 0; x < d; x++) {
	    if (b > bt) {
	      pass_1_print_error("More BANKS than BANKSTOTAL suggests.\n", ERROR_DIR);
	      return FAILED;
	    }

	    /* new banks? */
	    if (x >= rombanks) {
	      banks[o] = bs;
	      bankaddress[o] = a;
	    }
	    /* compare old banks */
	    else if (banks[o] != bs) {
	      pass_1_print_error("The old and the new ROMBANKMAPs don't match.\n", ERROR_DIR);
	      return FAILED;
	    }

	    o++;
	    b++;
	    a += bs;
	  }
	}
	else {
	  ind = FAILED;
	  break;
	}
      }
    }
    /* no ROMBANKMAP has been defined */
    else {
      o = 0;
      while ((ind = get_next_token()) == SUCCEEDED) {

	if (strcmp(tmp, ".ENDRO") == 0) {
	  break;
	}
	else if (strcmp(tmp, "BANKSTOTAL") == 0) {

	  q = input_number();

	  if (q == FAILED)
	    return FAILED;
	  if (q != SUCCEEDED || d <= 0) {
	    pass_1_print_error("BANKSTOTAL needs a positive value.\n", ERROR_DIR);
	    return FAILED;
	  }

	  banks = malloc(sizeof(int) * d);
	  bankaddress = malloc(sizeof(int) * d);
	  if (banks == NULL || bankaddress == NULL) {
	    pass_1_print_error("Out of memory while allocating ROM banks.\n", ERROR_DIR);
	    return FAILED;
	  }

	  bt = d;
	  bt_defined = 1;
	}
	else if (strcmp(tmp, "BANKSIZE") == 0) {

	  if (bt_defined == 0) {
	    pass_1_print_error("BANKSTOTAL needs to be defined prior to BANKSIZE.\n", ERROR_DIR);
	    return FAILED;
	  }

	  q = input_number();

	  if (q == FAILED)
	    return FAILED;
	  if (q != SUCCEEDED || d <= 0) {
	    pass_1_print_error("BANKSIZE needs a positive value.\n", ERROR_DIR);
	    return FAILED;
	  }

	  bs = d;
	  bs_defined = 1;
	}
	else if (strcmp(tmp, "BANKS") == 0) {

	  if (bs_defined == 0) {
	    pass_1_print_error("BANKSIZE needs to be defined prior to BANKS.\n", ERROR_DIR);
	    return FAILED;
	  }

	  q = input_number();

	  if (q == FAILED)
	    return FAILED;
	  if (q != SUCCEEDED || d <= 0) {
	    pass_1_print_error("BANKS needs a positive value.\n", ERROR_DIR);
	    return FAILED;
	  }

	  for (x = 0; x < d; x++) {
	    if (b > bt) {
	      pass_1_print_error("More BANKS than BANKSTOTAL suggests.\n", ERROR_DIR);
	      return FAILED;
	    }
	    banks[o] = bs;
	    bankaddress[o] = a;
	    o++;
	    b++;
	    a += bs;
	  }
	}
	else {
	  ind = FAILED;
	  break;
	}
      }
    }

    if (ind != SUCCEEDED) {
      pass_1_print_error("Error in ROMBANKMAP definition.\n", ERROR_DIR);
      return FAILED;
    }

    /* no banks definded? */
    if (bt == 0) {
      pass_1_print_error("No ROM banks were defined inside the ROMBANKMAP.\n", ERROR_DIR);
      return FAILED;	
    }
    if (bt != b) {
      pass_1_print_error("Not all ROM banks were defined inside the ROMBANKMAP.\n", ERROR_DIR);
      return FAILED;	
    }

#ifdef GB
    if (b != 2 && b != 4 && b != 8 && b != 16 && b != 32 && b != 64 &&
	b != 128 && b != 72 && b != 80 && b != 96) {
      pass_1_print_error("Unsupported amount of ROM banks.\n", ERROR_DIR);
      return FAILED;
    }

    if (b == 2)
      romtype = 0;
    else if (b == 4)
      romtype = 1;
    else if (b == 8)
      romtype = 2;
    else if (b == 16)
      romtype = 3;
    else if (b == 32)
      romtype = 4;
    else if (b == 64)
      romtype = 5;
    else if (b == 128)
      romtype = 6;
    else if (b == 72)
      romtype = 0x52;
    else if (b == 80)
      romtype = 0x53;
    else if (b == 96)
      romtype = 0x54;
#endif

    if (rombanks_defined != 0) {
      if (b > rombanks) {
	sprintf(emsg, "Upgrading from %d to %d ROM banks.\n", rombanks, b);
	pass_1_print_error(emsg, ERROR_WRN);
      }
      else
	return SUCCEEDED;
    }

    rombanks = b;
    rombanks_defined = 1;
    for (max_address = 0, q = 0; q < rombanks; q++)
      max_address += banks[q];

    if (rom_banks != NULL)
      free(rom_banks);
    if (rom_banks_usage_table != NULL)
      free(rom_banks_usage_table);

    rom_banks = malloc(sizeof(unsigned char) * max_address);
    rom_banks_usage_table = malloc(sizeof(unsigned char) * max_address);
    if (rom_banks == NULL || rom_banks_usage_table == NULL) {
      pass_1_print_error("Out of memory while allocating ROM banks.\n", ERROR_DIR);
      return FAILED;
    }

    memset(rom_banks_usage_table, 0, sizeof(unsigned char) * max_address);
    rombankmap_defined = 1;

    return SUCCEEDED;
  }

  /* MEMORYMAP */

  if (strcmp(cp, "MEMORYMAP") == 0) {

    int slotsize = 0, slotsize_defined = 0, s = 0;



    if (memorymap_defined == 1) {
      pass_1_print_error("MEMORYMAP can be defined only once.\n", ERROR_DIR);
      return FAILED;
    }
    if (output_format == OUTPUT_LIBRARY)
      pass_1_print_error("Libraries don't need MEMORYMAP.\n", ERROR_WRN);

    while ((ind = get_next_token()) == SUCCEEDED) {

      if (strcmp(tmp, ".ENDME") == 0) {

	if (defaultslot_defined == 0) {
	  pass_1_print_error("No DEFAULTSLOT defined.\n", ERROR_DIR);
	  return FAILED;
	}

	if (slots[defaultslot].size == 0) {
	  sprintf(emsg, "Unknown DEFAULTSLOT %d.\n", defaultslot);
	  pass_1_print_error(emsg, ERROR_DIR);
	  return FAILED;
	}

	break;
      }
      else if (strcmp(tmp, "SLOTSIZE") == 0) {

	q = input_number();

	if (q == FAILED)
	  return FAILED;
	if (q != SUCCEEDED) {
	  pass_1_print_error("SLOTSIZE needs an immediate value.\n", ERROR_DIR);
	  return FAILED;
	}

	slotsize = d;
	slotsize_defined = 1;
      }
      else if (strcmp(tmp, "DEFAULTSLOT") == 0) {

	if (defaultslot_defined != 0) {
	  pass_1_print_error("DEFAULTSLOT can be defined only once.\n", ERROR_DIR);
	  return FAILED;
	}

	q = input_number();

	if (q == FAILED)
	  return FAILED;
	if (q != SUCCEEDED || d > 255 || d < 0) {
	  pass_1_print_error("DEFAULTSLOT needs an immediate value (0-255) as an ID.\n", ERROR_DIR);
	  return FAILED;
	}

	defaultslot_defined = 1;
	defaultslot = d;
      }
      else if (strcmp(tmp, "SLOT") == 0) {

	if (slotsize_defined == 0) {
	  pass_1_print_error("SLOTSIZE must be defined prior to SLOT.\n", ERROR_DIR);
	  return FAILED;
	}

	q = input_number();

	if (q == FAILED)
	  return FAILED;
	if (q != SUCCEEDED || d > 255 || d < 0) {
	  pass_1_print_error("SLOT needs a positive value (0-255) as an ID.\n", ERROR_DIR);
	  return FAILED;
	}

	if (s != d) {
	  pass_1_print_error("Error in SLOT's ID. ID's start from 0.\n", ERROR_DIR);
	  return FAILED;
	}

	o = d;

	q = input_number();

	if (q == FAILED)
	  return FAILED;
	if (q != SUCCEEDED || d < 0) {
	  pass_1_print_error("SLOT needs a non-negative value as an ADDRESS.\n", ERROR_DIR);
	  return FAILED;
	}

	slots[o].address = d;
	slots[o].size = slotsize;
	slots_amount++;
	s++;
      }
      else {
	ind = FAILED;
	break;
      }
    }

    if (ind != SUCCEEDED) {
      pass_1_print_error("Error in MEMORYMAP definition.\n", ERROR_DIR);
      return FAILED;
    }

    memorymap_defined = 1;

    return SUCCEEDED;
  }

  /* BACKGROUND */

  if (strcmp(cp, "BACKGROUND") == 0) {

    FILE *file_in_ptr;



    if (output_format != OUTPUT_ROM) {
      pass_1_print_error("BACKGROUND can only be used in ROM output mode.\n", ERROR_DIR);
      return FAILED;
    }
    if (background_defined == 1) {
      pass_1_print_error("Only one BACKGROUND can be specified.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != INPUT_NUMBER_STRING) {
      pass_1_print_error("BACKGROUND needs a file name string.\n", ERROR_DIR);
      return FAILED;
    }

    if (memorymap_defined == 0) {
      pass_1_print_error("No MEMORYMAP defined.\n", ERROR_DIR);
      return FAILED;
    }
    if (rombanks_defined == 0) {
      pass_1_print_error("No ROMBANKS defined.\n", ERROR_DIR);
      return FAILED;
    }

    strcpy(full_name, include_dir);
    strcat(full_name, label);

    if ((file_in_ptr = fopen(full_name, "rb")) == NULL) {
      sprintf(emsg, "Error opening BACKGROUND file \"%s\".\n", full_name);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    fseek(file_in_ptr, 0, SEEK_END);
    background_size = ftell(file_in_ptr);
    fseek(file_in_ptr, 0, SEEK_SET);

    if (max_address != background_size) {
      sprintf(emsg, "BACKGROUND file \"%s\" size (%d) and ROM size (%d) don't match.\n", full_name,
	      background_size, max_address);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    memset(rom_banks, 0, sizeof(unsigned char) * max_address);
    memset(rom_banks_usage_table, 2, sizeof(unsigned char) * max_address);
    fread(rom_banks, 1, max_address, file_in_ptr);

    background_defined = 1;
    fclose(file_in_ptr);

    return SUCCEEDED;
  }

#ifdef GB
  /* RAMSIZE */

  if (strcmp(cp, "RAMSIZE") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take RAMSIZE.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || q < 0) {
      pass_1_print_error("RAMSIZE needs a non-negative value.\n", ERROR_DIR);
      return FAILED;
    }

    if (rambanks_defined != 0) {
      if (rambanks != d) {
	pass_1_print_error("RAMSIZE was defined for the second time.\n", ERROR_DIR);
	return FAILED;
      }
      return SUCCEEDED;
    }

    if (d != 0 && d != 1 && d != 2 && d != 3 && d != 4) {
      pass_1_print_error("Unsupported RAM size.\n", ERROR_DIR);
      return FAILED;
    }

    rambanks = d;
    rambanks_defined = 1;
    return SUCCEEDED;
  }

  /* CARTRIDGETYPE */

  if (strcmp(cp, "CARTRIDGETYPE") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take CARTRIDGETYPE.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (cartridgetype_defined != 0) {
      if (cartridgetype != d) {
	pass_1_print_error("CARTRIDGETYPE was defined for the second time.\n", ERROR_DIR);
	return FAILED;
      }
      return SUCCEEDED;
    }

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error("CARTRIDGETYPE needs an immediate value.\n", ERROR_DIR);
      return FAILED;
    }

    cartridgetype = d;
    cartridgetype_defined = 1;
    return SUCCEEDED;
  }

  /* LICENSEECODENEW */

  if (strcmp(cp, "LICENSEECODENEW") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take LICENSEECODENEW.\n", ERROR_DIR);
      return FAILED;
    }
    if (licenseecodeold_defined != 0) {
      pass_1_print_error("LICENSEECODENEW and LICENSEECODEOLD cannot both be defined.\n", ERROR_DIR);
      return FAILED;
    }

    if ((ind = get_next_token()) == FAILED)
      return FAILED;

    if (ind != GET_NEXT_TOKEN_STRING) {
      pass_1_print_error("LICENSEECODENEW requires a string of two letters.\n", ERROR_DIR);
      return FAILED;
    }
    if (!(tmp[0] != 0 && tmp[1] != 0 && tmp[2] == 0)) {
      pass_1_print_error("LICENSEECODENEW requires a string of two letters.\n", ERROR_DIR);
      return FAILED;
    }

    if (licenseecodenew_defined != 0) {
      if (tmp[0] != licenseecodenew_c1 || tmp[1] != licenseecodenew_c2) {
	pass_1_print_error("LICENSEECODENEW was defined for the second time.\n", ERROR_DIR);
	return FAILED;
      }
      return SUCCEEDED;
    }

    licenseecodenew_c1 = tmp[0];
    licenseecodenew_c2 = tmp[1];
    licenseecodenew_defined = 1;
    return SUCCEEDED;
  }

  /* LICENSEECODEOLD */

  if (strcmp(cp, "LICENSEECODEOLD") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take LICENSEECODEOLD.\n", ERROR_DIR);
      return FAILED;
    }
    if (licenseecodenew_defined != 0) {
      pass_1_print_error("LICENSEECODENEW and LICENSEECODEOLD cannot both be defined.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || d < -127 || d > 255) {
      pass_1_print_error("LICENSEECODEOLD needs a 8bit value.\n", ERROR_DIR);
      return FAILED;
    }

    if (licenseecodeold_defined != 0) {
      if (licenseecodeold != d) {
	pass_1_print_error("LICENSEECODEOLD was defined for the second time.\n", ERROR_DIR);
	return FAILED;
      }
      return SUCCEEDED;
    }

    licenseecodeold = d;
    licenseecodeold_defined = 1;
    return SUCCEEDED;
  }
#endif

  /* EMPTYFILL */

  if (strcmp(cp, "EMPTYFILL") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take EMPTYFILL.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || d < -127 || d > 255) {
      pass_1_print_error("EMPTYFILL needs a 8bit value.\n", ERROR_DIR);
      return FAILED;
    }

    if (emptyfill_defined != 0) {
      if (emptyfill != d) {
	pass_1_print_error("EMPTYFILL was defined for the second time.\n", ERROR_DIR);
	return FAILED;
      }
      return SUCCEEDED;
    }

    emptyfill = d;
    emptyfill_defined = 1;
    return SUCCEEDED;
  }

  /* DEFINE */

  if (strcmp(cp, "DEFINE") == 0) {

    char k[256];
    int j, size;


    if (get_next_token() == FAILED)
      return FAILED;

    q = get_a_new_definition_data(&j, k, &size);
    if (q == FAILED)
      return FAILED;
    if (!(q == SUCCEEDED || q == INPUT_NUMBER_STRING)) {
      pass_1_print_error("DEFINE can only be used to assign an immediate value or a string to a definition label.\n", ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED)
      add_a_new_definition(tmp, j, NULL, DEFINITION_TYPE_VALUE, 0);
    else
      add_a_new_definition(tmp, 0, k, DEFINITION_TYPE_STRING, size);

    return SUCCEEDED;
  }

  /* REDEFINE */

  if (strcmp(cp, "REDEFINE") == 0) {

    char k[256];
    int j, size;


    if (get_next_token() == FAILED)
      return FAILED;

    q = get_a_new_definition_data(&j, k, &size);
    if (q == FAILED)
      return FAILED;
    if (!(q == SUCCEEDED || q == INPUT_NUMBER_STRING)) {
      pass_1_print_error("REDEFINE can only be used to reassign an immediate value or string to a definition.\n", ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED)
      redefine(tmp, j, NULL, DEFINITION_TYPE_VALUE, 0);
    else
      redefine(tmp, 0, k, DEFINITION_TYPE_STRING, size);

    return SUCCEEDED;
  }

  /* EXPORT */

  if (strcmp(cp, "EXPORT") == 0) {

    if (get_next_token() == FAILED)
      return FAILED;

    export_tmp = export_first;
    while (export_tmp != NULL) {
      if (strcmp(export_tmp->name, tmp) == 0) {
	sprintf(emsg, "\"%s\" was .EXPORT'd for the second time.\n", tmp);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }
      export_tmp = export_tmp->next;
    }

    export_tmp = malloc(sizeof(struct export_def));
    if (export_tmp == NULL) {
      sprintf(emsg, "Out of memory while allocating room for \".EXPORT %s\".\n", tmp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    strcpy(export_tmp->name, tmp);
    export_tmp->next = NULL;

    if (export_first == NULL) {
      export_first = export_tmp;
      export_last = export_tmp;
    }
    else {
      export_last->next = export_tmp;
      export_last = export_tmp;
    }

    return SUCCEEDED;
  }

  /* ENUM */

  if (strcmp(cp, "ENUM") == 0) {

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error("ENUM needs a starting address.\n", ERROR_DIR);
      return FAILED;
    }

    o = d;

    while ((t = get_next_token()) != FAILED) {
      if (strcmp(tmp, ".ENDE") == 0)
	return SUCCEEDED;
      if (add_a_new_definition(tmp, o, NULL, DEFINITION_TYPE_VALUE, 0) == FAILED)
	return FAILED;
      if (get_next_token() == FAILED)
	return FAILED;
      if ((strcmp(tmp, "DB") == 0) || (strcmp(tmp, "db") == 0))
	o++;
      else if ((strcmp(tmp, "DW") == 0) || (strcmp(tmp, "dw") == 0))
	o += 2;
      else if ((strcmp(tmp, "DS") == 0) || (strcmp(tmp, "ds") == 0)) {
	q = input_number();
	if (q == FAILED)
	  return FAILED;
	if (q != SUCCEEDED) {
	  pass_1_print_error("DS needs size.\n", ERROR_DIR);
	  return FAILED;
	}
	o += d;
      }
      else if (tmp[0] == '.')
	continue;
      else {
	sprintf(emsg, "Unknown symbol \"%s\" in .ENUM.\n", tmp);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }
    }
    return FAILED;
  }

#ifdef GB
  /* COMPUTECHECKSUM */

  if (strcmp(cp, "COMPUTECHECKSUM") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take COMPUTECHECKSUM.\n", ERROR_DIR);
      return FAILED;
    }

    computechecksum_defined = 1;

    return SUCCEEDED;
  }

  /* COMPUTECOMPLEMENTCHECK */

  if (strcmp(cp, "COMPUTECOMPLEMENTCHECK") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take COMPUTECOMPLEMENTCHECK.\n", ERROR_DIR);
      return FAILED;
    }

    computecomplementcheck_defined = 1;

    return SUCCEEDED;
  }
#endif

#ifdef Z80
  /* COMPUTESMSCHECKSUM */

  if (strcmp(cp, "COMPUTESMSCHECKSUM") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take COMPUTESMSCHECKSUM.\n", ERROR_DIR);
      return FAILED;
    }

    computesmschecksum_defined++;

    return SUCCEEDED;
  }

  if (strcmp(cp, "SMSTAG") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .SMSTAG.\n", ERROR_DIR);
      return FAILED;
    }

    smstag_defined++;
    computesmschecksum_defined++;

    return SUCCEEDED;
  }

  /* SDSCTAG */

  if (strcmp(cp, "SDSCTAG") == 0) {

    if (sdsctag_defined != 0) {
      pass_1_print_error(".SDSCTAG can be defined only once.\n", ERROR_DIR);
      return FAILED;
    }
    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .SDSCTAG.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();
    if (q != SUCCEEDED && q != INPUT_NUMBER_FLOAT) {
      pass_1_print_error(".SDSCTAG needs the version number.\n" , ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED) {
      sdsc_ma = d;
      sdsc_mi = 0;
    }
    else {
      sdsc_ma = flo_h;
      sdsc_mi = flo_l;
    }

    if (sdsc_ma > 255 || sdsc_mi > 255) {
      pass_1_print_error(".SDSCTAG major and minor version numbers must be inside the range 0-255.\n" , ERROR_DIR);
      return FAILED;
    }

    /* program name */
    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (!(q == INPUT_NUMBER_STRING || q == SUCCEEDED || q == INPUT_NUMBER_STACK || q == INPUT_NUMBER_ADDRESS_LABEL)) {
      pass_1_print_error(".SDSCTAG requires program name string (or a pointer to it).\n", ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED) {
      sdsctag_name_type = TYPE_VALUE;
      sdsctag_name_value = d;
    }
    else if (q == INPUT_NUMBER_STRING) {
      if (label[0] == 0) {
	sdsctag_name_type = TYPE_VALUE;
	sdsctag_name_value = 0xFFFF; /* null string */
      }
      else {
	sdsctag_name_type = TYPE_STRING;
	strcpy(sdsctag_name_str, label);
      }
    }
    else if (q == INPUT_NUMBER_ADDRESS_LABEL) {
      sdsctag_name_type = TYPE_LABEL;
      strcpy(sdsctag_name_str, label);
    }
    else {
      sdsctag_name_type = TYPE_STACK_CALCULATION;
      sdsctag_name_value = latest_stack;
    }

    /* notes */
    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (!(q == INPUT_NUMBER_STRING || q == SUCCEEDED || q == INPUT_NUMBER_STACK || q == INPUT_NUMBER_ADDRESS_LABEL)) {
      pass_1_print_error(".SDSCTAG requires program release notes string (or a pointer to it).\n", ERROR_DIR);
      return FAILED;
    }

    if (q == SUCCEEDED) {
      sdsctag_notes_type = TYPE_VALUE;
      sdsctag_notes_value = d;
    }
    else if (q == INPUT_NUMBER_STRING) {
      if (label[0] == 0) {
	sdsctag_notes_type = TYPE_VALUE;
	sdsctag_notes_value = 0xFFFF; /* null string */
      }
      else {
	sdsctag_notes_type = TYPE_STRING;
	strcpy(sdsctag_notes_str, label);
      }
    }
    else if (q == INPUT_NUMBER_ADDRESS_LABEL) {
      sdsctag_notes_type = TYPE_LABEL;
      strcpy(sdsctag_notes_str, label);
    }
    else {
      sdsctag_notes_type = TYPE_STACK_CALCULATION;
      sdsctag_notes_value = latest_stack;
    }

    sdsctag_defined++;
    smstag_defined++;
    computesmschecksum_defined++;

    return SUCCEEDED;
  }
#endif

  /* MACRO */

  if (strcmp(cp, "MACRO") == 0) {

    struct gba_macro *m;
    unsigned char argz[256], e;
    char emsg[256];
    int d, k, macro_start = 0;



    if (get_next_token() == FAILED)
      return FAILED;

    if (error_level == ERROR_LEVEL_MAX)
      macro_start = active_file_info_last->line_current;

    m = macros_first;
    while (m != NULL) {
      if (strcmp(tmp, m->name) == 0) {
	sprintf(emsg, "Macro \"%s\" was defined for the second time.\n", tmp);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }
      m = m->next;
    }

    m = malloc(sizeof(struct gba_macro));
    if (m == NULL) {
      pass_1_print_error("Out of memory while allocating room for a new MACRO.\n", ERROR_DIR);
      return FAILED;
    }

    if (macros_first == NULL) {
      macros_first = m;
      macros_last = m;
    }
    else {
      macros_last->next = m;
      macros_last = m;
    }

    strcpy(m->name, tmp);
    m->next = NULL;
    m->arguments = 0;
    m->calls = 0;
    m->start = i;
    if (error_level == ERROR_LEVEL_MAX) {
      m->start_line = active_file_info_last->line_current;
      m->filename_id = active_file_info_last->filename_id;
    }

    /* clear argument table */
    for (d = 0; d < 256; d++)
      argz[d] = 0;

    for (; i < size; i++) {

      if (buffer[i] == 0x0A && error_level == ERROR_LEVEL_MAX) {
	next_line();
	continue;
      }

      /* a new argument */
      if (buffer[i] == '\\') {
	e = buffer[++i];
	if (e >= '0' && e <= '9') {
	  d = (e - '0') * 10;
	  for (k = 2; k < 8; k++, d *= 10) {
	    e = buffer[++i];
	    if (e >= '0' && e <= '9')
	      d += e - '0';
	    else {
	      i--;
	      break;
	    }
	  }

	  d /= 10;
	}
	else
	  continue;

	if (d == 0 || d > 256) {
	  pass_1_print_error("MACRO argument number must be between 1 and 256.\n", ERROR_DIR);
	  return FAILED;
	}

	/* mark the argument as used */
	if (argz[d - 1] == 0) {
	  m->arguments++;
	  argz[d - 1] = 1;
	}
      }

      else if ((strncmp(&buffer[i], ".ENDM", 5) == 0) && (buffer[i + 5] == 0x0A || buffer[i + 5] == ' ')) {
	i += 5;
	break;
      }
      else if ((strncmp(&buffer[i], ".E", 2) == 0) && (buffer[i + 2] == 0x0A || buffer[i + 2] == ' ')) {
	if (error_level == ERROR_LEVEL_MAX)
	  active_file_info_last->line_current = macro_start;
	sprintf(emsg, "MACRO \"%s\" wasn't terminated with .ENDM.\n", m->name);
	pass_1_print_error(emsg, ERROR_DIR);
	return FAILED;
      }
    }

    for (k = 0, d = 0; argz[d] == 1 && d < 256; d++)
      k++;

    /* an argument is missing! */
    if (k != m->arguments) {
      if (error_level == ERROR_LEVEL_MAX)
	active_file_info_last->line_current = macro_start;
      sprintf(emsg, "MACRO \"%s\"'s argument number %d isn't applied.\n", m->name, d + 1);
      pass_1_print_error(emsg, ERROR_NUM);
      return FAILED;
    }

    return SUCCEEDED;
  }

  /* REPT */

  if (strcmp(cp, "REPT") == 0) {

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error("REPT needs a count.\n", ERROR_INP);
      return FAILED;
    }

    if (d < 0) {
      pass_1_print_error("REPT count value must be positive or zero.\n", ERROR_DIR);
      return FAILED;
    }

    if (d == 0) {

      int l = 0, r, m;


      if (error_level == ERROR_LEVEL_MAX)
	l = active_file_info_last->line_current;
      /* find the next compiling point */
      r = 1;
      m = macro_active;
      /* disable macro decoding */
      macro_active = 0;
      while (get_next_token() != FAILED) {
	if (tmp[0] == '.') {
	  if (strcmp(cp, "ENDR") == 0)
	    r--;
	  if (strcmp(cp, "E") == 0)
	    break;
	  if (strcmp(cp, "REPT") == 0)
	    r++;
	}
	if (r == 0) {
	  macro_active = m;
	  return SUCCEEDED;
	}
      }

      /* return the condition's line number */
      if (error_level == ERROR_LEVEL_MAX)
	active_file_info_last->line_current = l;
      sprintf(emsg, "REPT must end to ENDR.\n");
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (repeat_active == REPEAT_STACK_DEPTH) {
      sprintf(emsg, "REPT maximum stack depth of %d items is reached.\n", REPEAT_STACK_DEPTH);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    repeat_stack[repeat_active].start = i;
    repeat_stack[repeat_active].counter = d;

    if (error_level == ERROR_LEVEL_MAX)
      repeat_stack[repeat_active].start_line = active_file_info_last->line_current;

    repeat_active++;

    return SUCCEEDED;
  }

  /* ENDR */

  if (strcmp(cp, "ENDR") == 0) {

    if (repeat_active == 0)
      return EVALUATE_TOKEN_NOT_IDENTIFIED;

    repeat_stack[repeat_active - 1].counter--;
    if (repeat_stack[repeat_active - 1].counter == 0) {
      repeat_active--;
      return SUCCEEDED;
    }

    i = repeat_stack[repeat_active - 1].start;

    if (error_level == ERROR_LEVEL_MAX)
      active_file_info_last->line_current = repeat_stack[repeat_active - 1].start_line;

    return SUCCEEDED;
  }

  /* E */

  if (strcmp(cp, "E") == 0) {
    if (error_level == ERROR_LEVEL_MAX) {
      if (active_file_info_last != NULL) {
	active_file_info_tmp = active_file_info_last;
	active_file_info_last = active_file_info_last->prev;
	free(active_file_info_tmp);
	if (active_file_info_last == NULL)
	  active_file_info_first = NULL;
	else
	  fprintf(file_out_ptr, "f %d ", active_file_info_last->filename_id);
      }
    }
    fprintf(file_out_ptr, "E ");
    open_files--;
    if (open_files == 0)
      return EVALUATE_TOKEN_EOP;

    if (extra_definitions == ON && error_level == ERROR_LEVEL_MAX)
      redefine("WLA_FILENAME", 0, get_file_name(active_file_info_last->filename_id), DEFINITION_TYPE_STRING, strlen(get_file_name(active_file_info_last->filename_id)));

    return SUCCEEDED;
  }

  /* M */

  if (strcmp(cp, "M") == 0) {
    if (line_count_status == OFF)
      line_count_status = ON;
    else
      line_count_status = OFF;
    return SUCCEEDED;
  }

#ifdef GB
  /* ROMGBC */

  if (strcmp(cp, "ROMGBC") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take ROMGBC.\n", ERROR_DIR);
      return FAILED;
    }
    else if (romdmg != 0) {
      pass_1_print_error("ROMDMG was defined prior to ROMGBC.\n", ERROR_DIR);
      return FAILED;
    }

    romgbc++;
    return SUCCEEDED;
  }

  /* ROMDMG */

  if (strcmp(cp, "ROMDMG") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take ROMDMG.\n", ERROR_DIR);
      return FAILED;
    }
    else if (romgbc != 0) {
      pass_1_print_error("ROMGBC was defined prior to ROMDMG.\n", ERROR_DIR);
      return FAILED;
    }
    else if (romsgb != 0) {
      pass_1_print_error("ROMDMG and ROMSGB cannot be mixed.\n", ERROR_DIR);
      return FAILED;
    }

    romdmg++;
    return SUCCEEDED;
  }

  /* ROMSGB */

  if (strcmp(cp, "ROMSGB") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take ROMSGB.\n", ERROR_DIR);
      return FAILED;
    }
    else if (romdmg != 0) {
      pass_1_print_error("ROMDMG and ROMSGB cannot be mixed.\n", ERROR_DIR);
      return FAILED;
    }

    romsgb++;
    return SUCCEEDED;
  }
#endif

#ifndef GB
  /* HEADER */

  if (strcmp(cp, "HEADER") == 0) {

    int s, r;


    if (output_format != OUTPUT_ROM && output_format != OUTPUT_PROGRAM) {
      pass_1_print_error("HEADER can only be used in ROM/PROGRAM output mode.\n", ERROR_DIR);
      return FAILED;
    }
    if (header_defined != 0) {
      pass_1_print_error("HEADER is already defined.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != INPUT_NUMBER_STRING) {
      pass_1_print_error("HEADER needs a file name string.\n", ERROR_DIR);
      return FAILED;
    }

    if (incbin_file(label, &header, &inz, &s, &r) == FAILED)
      return FAILED;

    header_defined = 1;
    return SUCCEEDED;
  }

  /* FOOTER */

  if (strcmp(cp, "FOOTER") == 0) {

    int s, r;


    if (output_format != OUTPUT_ROM && output_format != OUTPUT_PROGRAM) {
      pass_1_print_error("FOOTER can only be used in ROM/PROGRAM output mode.\n", ERROR_DIR);
      return FAILED;
    }
    if (footer_defined != 0) {
      pass_1_print_error("FOOTER is already defined.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != INPUT_NUMBER_STRING) {
      pass_1_print_error("FOOTER needs a file name string.\n", ERROR_DIR);
      return FAILED;
    }

    if (incbin_file(label, &footer, &inz, &s, &r) == FAILED)
      return FAILED;

    footer_defined = 1;
    return SUCCEEDED;
  }
#endif

  /* ROMBANKSIZE */

  if (strcmp(cp, "ROMBANKSIZE") == 0 || strcmp(cp, "BANKSIZE") == 0) {

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || d < 0) {
      pass_1_print_error("ROMBANKSIZE needs a positive integer value.\n", ERROR_DIR);
      return FAILED;
    }

    if (banksize_defined != 0) {
      if (banksize != d) {
	pass_1_print_error("ROMBANKSIZE was defined for the second time.\n", ERROR_DIR);
	return FAILED;
      }
      return SUCCEEDED;
    }

    banksize = d;
    banksize_defined = 1;

    return SUCCEEDED;
  }

  /* ENDM */

  if (strcmp(cp, "ENDM") == 0) {
    if (macro_active != 0) {

      struct gba_def *d, *p;


      macro_active--;
      if (macro_active == 0) {
	/* delete NARGS */
	d = defines;
	p = NULL;
	while (d != NULL) {
	  if (strcmp("NARGS", d->alias) == 0) {
	    if (p != NULL)
	      p->next = d->next;
	    else
	      defines = d->next;
	    free(d);
	    break;
	  }
	  p = d;
	  d = d->next;
	}

	i = macro_stack[0].macro_end;

	if (error_level == ERROR_LEVEL_MAX) {
	  if ((extra_definitions == ON) && (active_file_info_last->filename_id != macro_stack[0].macro_end_filename_id))
	    redefine("WLA_FILENAME", 0, get_file_name(macro_stack[0].macro_end_filename_id), DEFINITION_TYPE_STRING,
		     strlen(get_file_name(macro_stack[0].macro_end_filename_id)));

	  active_file_info_last->filename_id = macro_stack[0].macro_end_filename_id;
	  active_file_info_last->line_current = macro_stack[0].macro_end_line;
	}

	return SUCCEEDED;
      }

      /* redefine NARGS */
      d = defines;
      while (d != NULL) {
	if (strcmp("NARGS", d->alias) == 0) {
	  d->value = macro_stack[macro_active - 1].supplied_arguments;
	  break;
	}
	d = d->next;
      }

      i = macro_stack[macro_active].macro_end;

      if (error_level == ERROR_LEVEL_MAX) {
	if ((extra_definitions == ON) && (active_file_info_last->filename_id != macro_stack[macro_active].macro_end_filename_id))
	  redefine("WLA_FILENAME", 0, get_file_name(macro_stack[macro_active].macro_end_filename_id), DEFINITION_TYPE_STRING,
		   strlen(get_file_name(macro_stack[macro_active].macro_end_filename_id)));

	active_file_info_last->filename_id = macro_stack[macro_active].macro_end_filename_id;
	active_file_info_last->line_current = macro_stack[macro_active].macro_end_line;
      }

      macros_current = macro_stack[macro_active - 1].macro;
      return SUCCEEDED;
    }

    return EVALUATE_TOKEN_NOT_IDENTIFIED;
  }

#if defined MCS6502 || MCS6510 || W65816

  /* 8BIT */

  if (strcmp(cp, "8BIT") == 0) {
    skip_xbit = SKIP_NO;
    return SUCCEEDED;
  }

  /* 16BIT */

  if (strcmp(cp, "16BIT") == 0) {
    skip_xbit = SKIP_8BIT;
    return SUCCEEDED;
  }

#endif

#ifdef W65816

  /* 24BIT */

  if (strcmp(cp, "24BIT") == 0) {
    skip_xbit = SKIP_16BIT;
    return SUCCEEDED;
  }

  /* BASE */

  if (strcmp(cp, "BASE") == 0) {
    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .BASE definitions.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();

    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED || d < 0) {
      pass_1_print_error(".BASE number must be zero or positive.\n", ERROR_DIR);
      return FAILED;
    }

    fprintf(file_out_ptr, "b %d ", d);

    return SUCCEEDED;
  }

  /* HIROM */

  if (strcmp(cp, "HIROM") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .HIROM.\n", ERROR_DIR);
      return FAILED;
    }
    else if (lorom_defined != 0) {
      pass_1_print_error(".LOROM was defined prior to .HIROM.\n", ERROR_DIR);
      return FAILED;
    }

    hirom_defined++;
    snes_mode++;
    return SUCCEEDED;
  }

  /* LOROM */

  if (strcmp(cp, "LOROM") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .LOROM.\n", ERROR_DIR);
      return FAILED;
    }
    else if (hirom_defined != 0) {
      pass_1_print_error(".HIROM was defined prior to .LOROM.\n", ERROR_DIR);
      return FAILED;
    }

    lorom_defined++;
    snes_mode++;
    return SUCCEEDED;
  }

  /* SLOWROM */

  if (strcmp(cp, "SLOWROM") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .SLOWROM.\n", ERROR_DIR);
      return FAILED;
    }
    else if (lorom_defined != 0) {
      pass_1_print_error(".FASTROM was defined prior to .SLOWROM.\n", ERROR_DIR);
      return FAILED;
    }

    slowrom_defined++;
    snes_mode++;
    return SUCCEEDED;
  }

  /* FASTROM */

  if (strcmp(cp, "FASTROM") == 0) {

    if (output_format == OUTPUT_LIBRARY) {
      pass_1_print_error("Library files don't take .FASTROM.\n", ERROR_DIR);
      return FAILED;
    }
    else if (lorom_defined != 0) {
      pass_1_print_error(".SLOWROM was defined prior to .FASTROM.\n", ERROR_DIR);
      return FAILED;
    }

    fastrom_defined++;
    snes_mode++;
    return SUCCEEDED;
  }

#endif

  /* PRINTT */

  if (strcmp(cp, "PRINTT") == 0) {

    char t[256];
    int s, u;



    inz = input_number();

    if (inz != INPUT_NUMBER_STRING) {
      pass_1_print_error(".PRINTT needs a string.\n", ERROR_DIR);
      return FAILED;
    }

    for (s = 0, u = 0; label[s] != 0;) {
      if (label[s] == '\\' && label[s + 1] == 'n') {
#ifdef MSDOS
	t[u++] = 0x0D;
	t[u++] = 0x0A;
#else
	t[u++] = 0x0A;
#endif
	s += 2;
      }
      else
	t[u++] = label[s++];
    }

    t[u] = 0;
    printf("%s", t);

    return SUCCEEDED;
  }

  /* PRINTV */

  if (strcmp(cp, "PRINTV") == 0) {

    int m = 0;



    if (get_next_token() == FAILED)
      return FAILED;

    if (strcmp(tmp, "HEX") == 0)
      m = 0;
    else if (strcmp(tmp, "DEC") == 0)
      m = 1;
    else {
      pass_1_print_error(".PRINTV needs to know the format of the output, HEX or DEC.\n", ERROR_DIR);
      return FAILED;
    }

    q = input_number();
    if (q == FAILED)
      return FAILED;
    if (q != SUCCEEDED) {
      pass_1_print_error(".PRINTV can only print currently known values.\n", ERROR_DIR);
      return FAILED;
    }

    if (m == 0)
      printf("%x", d);
    else
      printf("%d", d);

    return SUCCEEDED;
  }

  /* DWSIN/DBSIN/DWCOS/DBCOS */

  if (strcmp(cp, "DWSIN") == 0 || strcmp(cp, "DBSIN") == 0 || strcmp(cp, "DWCOS") == 0 || strcmp(cp, "DBCOS") == 0) {

    double m, a, s, n;
    int p, c, o, f;
    char emsg[256];


    if (cp[1] == 'W')
      o = 1;
    else
      o = 2;
    if (cp[2] == 'S')
      f = 1;
    else
      f = 2;

    p = input_number();
    if (p != SUCCEEDED && p != INPUT_NUMBER_FLOAT) {
      sprintf(emsg, ".%s needs a value for starting angle.\n", cp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (p == SUCCEEDED)
      a = d;
    else
      a = f;

    if (input_number() != SUCCEEDED || d < 0) {
      sprintf(emsg, ".%s needs an non-negative integer value for additional angles.\n", cp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    c = d;

    p = input_number();
    if (p != SUCCEEDED && p != INPUT_NUMBER_FLOAT) {
      sprintf(emsg, ".%s needs a value for angle step.\n", cp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (p == SUCCEEDED)
      s = d;
    else
      s = f;

    p = input_number();
    if (p != SUCCEEDED && p != INPUT_NUMBER_FLOAT) {
      sprintf(emsg, ".%s needs a value to multiply the result with.\n", cp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (p == SUCCEEDED)
      m = d;
    else
      m = f;

    p = input_number();
    if (p != SUCCEEDED && p != INPUT_NUMBER_FLOAT) {
      sprintf(emsg, ".%s needs a value to add to the result.\n", cp);
      pass_1_print_error(emsg, ERROR_DIR);
      return FAILED;
    }

    if (p == SUCCEEDED)
      n = d;
    else
      n = f;

    for (c++; c > 0; c--) {
	a += 360;
      while (a >= 360)
	a -= 360;

      if (f == 1)
	d = (int)((m * sin(a * 2 * M_PI / 360)) + n);
      else
	d = (int)((m * cos(a * 2 * M_PI / 360)) + n);

      if (o == 1) {
	if (d < -32768 || d > 65535) {
	  sprintf(emsg, ".%s: Expected a 16bit value, computed %d.\n", cp, d);
	  pass_1_print_error(emsg, ERROR_NONE);
	  return FAILED;
	}
	fprintf(file_out_ptr, "y %d ", d);
      }
      else {
	if (d > 255 || d < -127) {
	  sprintf(emsg, ".%s: Expected a 8bit value, computed %d.\n", cp, d);
	  pass_1_print_error(emsg, ERROR_NONE);
	  return FAILED;
	}
	fprintf(file_out_ptr, "d %d ", d);
      }

      a += s;
    }

    return SUCCEEDED;
  }

  /* FAIL */

  if (strcmp(cp, "FAIL") == 0) {
    pass_1_print_error("HALT: .FAIL found.\n", ERROR_NONE);
    return FAILED;
  }

  /* UNDEF */

  if (strcmp(cp, "UNDEF") == 0) {

    struct gba_def *d, *p;



    if (get_next_token() == FAILED)
      return FAILED;

    d = defines;
    p = NULL;
    while (d != NULL) {
      if (strcmp(tmp, d->alias) == 0) {
	if (p != NULL)
	  p->next = d->next;
	else
	  defines = d->next;
	free(d);
	return SUCCEEDED;
      }
      p = d;
      d = d->next;
    }

    return SUCCEEDED;
  }

  /* ASM */

  if (strcmp(cp, "ASM") == 0)
    return SUCCEEDED;

  /* ENDASM */

  if (strcmp(cp, "ENDASM") == 0) {

    int asm = 1;


    while (1) {
      q = get_next_token();
      if (q == GET_NEXT_TOKEN_STRING)
	continue;
      if (strcmp(tmp, ".E") == 0) {
	i = i_e2;
	return SUCCEEDED;
      }

      if (strcmp(tmp, ".ASM") == 0) {
	asm--;
	if (asm == 0)
	  return SUCCEEDED;
      }
      if (strcmp(tmp, ".ENDASM") == 0)
	asm++;
    }
  }

  sprintf(emsg, "Unknown directive \"%s\".\n", tmp);
  pass_1_print_error(emsg, ERROR_NONE);
  return FAILED;
}


int find_next_point(char *n) {

  int r, m, l = 0;


  if (error_level == ERROR_LEVEL_MAX)
    l = active_file_info_last->line_current;
  /* find the next compiling point */
  r = 1;
  m = macro_active;
  /* disable macro decoding */
  macro_active = 0;
  while (get_next_token() != FAILED) {
    if (tmp[0] == '.') {
      if (strcmp(cp, "ENDIF") == 0)
	r--;
      if (strcmp(cp, "ELSE") == 0 && r == 1)
	r--;
      if (strcmp(cp, "E") == 0)
	break;
      if (strcmp(cp, "IFDEF") == 0 || strcmp(cp, "IFNDEF") == 0 || strcmp(cp, "IFGR") == 0 || strcmp(cp, "IFLE") == 0 ||
	  strcmp(cp, "IFEQ") == 0 || strcmp(cp, "IFNEQ") == 0 || strcmp(cp, "IFDEFM") == 0 || strcmp(cp, "IFNDEFM") == 0 ||
	  strcmp(cp, "IF") == 0 || strcmp(cp, "IFGREQ") == 0 || strcmp(cp, "IFLEEQ") == 0 || strcmp(cp, "IFEXISTS") == 0)
	r++;
    }
    if (r == 0) {
      if (strcmp(cp, "ELSE") == 0)
	ifdef++;
      macro_active = m;
      return SUCCEEDED;
    }
  }

  /* return the condition's line number */
  if (error_level == ERROR_LEVEL_MAX)
    active_file_info_last->line_current = l;
  sprintf(emsg, "%s must end to ENDIF/ELSE.\n", n);
  pass_1_print_error(emsg, ERROR_DIR);
  return FAILED;
}


#ifndef GB
void delete_stack(struct stack *s) {

  free(s->stack);
  free(s);
}
#endif


int get_a_new_definition_data(int *b, char *c, int *size) {

  int a, x, n, s;


  x = input_number();
  a = x;
  n = 0;
  s = 0;

  while (x != INPUT_NUMBER_EOL) {
    /* the first fetch will conserve both classes */
    if (n == 0) {
      if (x == SUCCEEDED)
	*b = d;
      else if (x == INPUT_NUMBER_STRING) {
	strcpy(c, label);
	s = strlen(label);
      }
      else
	return x;

      n++;
      x = input_number();
      continue;
    }
    /* the next fetches will produce strings */
    else if (n == 1 && a == SUCCEEDED) {
      if (*b > 255) {
	c[0] = (*b % 255);
	c[1] = (*b >> 8) & 255;
	c[2] = 0;
	s = 2;
      }
      else {
	c[0] = *b;
	c[1] = 0;
	s = 1;
      }
      a = INPUT_NUMBER_STRING;
    }

    if (x == INPUT_NUMBER_STRING) {
      strcpy(&c[s], label);
      s += strlen(label);
    }
    else if (x == SUCCEEDED) {
      if (d > 255) {
	c[s] = (d % 255);
	c[s + 1] = (d >> 8) & 255;
	c[s + 2] = 0;
	s += 2;
      }
      else {
	c[s] = d;
	c[s + 1] = 0;
	s++;
      }
    }
    else
      return x;

    n++;
    x = input_number();
  }

  if (error_level == ERROR_LEVEL_MAX)
    next_line();

  if (a == INPUT_NUMBER_STRING)
    c[s] = 0;

  *size = s;
  return a;
}
