
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"

#include "include_file.h"
#include "pass_3.h"


extern struct incbin_file_data *incbin_file_data_first, *ifd_tmp;
extern struct section_def *sections_first, *sections_last, *sec_tmp, *sec_next;
extern struct file_name_info *file_name_info_first, *file_name_info_last, *file_name_info_tmp;
extern FILE *file_out_ptr;
extern unsigned char *rom_banks, *rom_banks_usage_table;
extern char gba_tmp_name[32], *gba_tmp_last_name, tmp[4096];
extern int verbose_mode, section_status, cartridgetype, output_format, error_level;


struct label_def *label_next, *label_last, *label_tmp, *labels = NULL;



int pass_3(void) {

  FILE *f_in;
  int bank = 0, slot = 0, add = 0, file_name_id = 0, ind, inz, line_number = 0, o;
#ifdef W65816
  int base = 0x00;
#endif
  char c;




  sec_tmp = NULL;

  if (verbose_mode == ON)
    printf("Internal pass 1...\n");

  if ((f_in = fopen(gba_tmp_name, "rb")) == NULL) {
    fprintf(stderr, "INTERNAL_PASS_1: Error opening file \"%s\".\n", gba_tmp_name);
    return FAILED;
  }

  /* first loop */

  if (output_format != OUTPUT_LIBRARY) {
    o = 0;
    while (o == 0 && fread(&c, 1, 1, f_in) != 0) {
      switch (c) {

      case ' ':
      case 'E':
	continue;

      case 'f':
	fscanf(f_in, "%d ", &file_name_id);
	continue;

      case 'B':
	fscanf(f_in, "%d %d ", &bank, &slot);
	continue;

      case 'O':
	fscanf(f_in, "%d ", &add);
	o++;
	break;

      default:
	fprintf(stderr, "INTERNAL_PASS_1: ORG needs to be defined before any code/data can be accepted.\n");
	return FAILED;
      }
    }
  }

  /* second (major) loop */

  while (fread(&c, 1, 1, f_in) != 0) {
    switch (c) {

    case ' ':
    case 'E':
      continue;

    case 'A':
    case 'S':
      if (c == 'A')
	fscanf(f_in, "%d %*d", &inz);
      else
	fscanf(f_in, "%d ", &inz);

      if (section_status == ON) {
	fprintf(stderr, "INTERNAL_PASS_1: Section \"%s\" is open and we are trying to define already another section called ", sec_tmp->name);
	sec_next = sections_first;
	while (sec_next != NULL) {
	  if (sec_next->id == inz) {
	    fprintf(stderr, "\"%s\".\n", sec_next->name);
	    return FAILED;
	  }
	  sec_next = sec_next->next;
	}
      }

      sec_next = sections_first;
      while (sec_next != NULL) {
	if (sec_next->id == inz) {
	  sec_tmp = sec_next;
	  if (sec_tmp->status == SECTION_STATUS_FREE)
	    add = 0;
	  sec_tmp->address = add;
	  sec_tmp->bank = bank;
	  sec_tmp->slot = slot;
	  break;
	}
	sec_next = sec_next->next;
      }
      section_status = ON;
      continue;

    case 's':
      sec_tmp->size = add - sec_tmp->address;
      section_status = OFF;
      sec_tmp = NULL;
      continue;

    case 'x':
      fscanf(f_in, "%d %*d ", &inz);
      add += inz;
      continue;

    case 'X':
      fscanf(f_in, "%d %*d ", &inz);
      add += inz << 1;
      continue;

#ifdef W65816
    case 'z':
    case 'q':
      fscanf(f_in, "%*s ");
      add += 3;
      continue;

    case 'T':
      fscanf(f_in, "%*d ");
      add += 3;
      continue;

    case 'b':
      fscanf(f_in, "%d ", &base);
      continue;
#endif

    case 'R':
    case 'Q':
    case 'd':
    case 'c':
      fscanf(f_in, "%*s ");
      add++;
      continue;

    case 'r':
      fscanf(f_in, "%*s ");
      add += 2;
      continue;

    case 'y':
    case 'C':
      fscanf(f_in, "%*d ");
      add += 2;
      continue;

    case 'a':
      fscanf(f_in, "%d ", &inz);
      ind = add / inz;
      ind = ind * inz;
      if ((add - ind) != 0)
	add = ind + inz;
      continue;

    case 'D':
      fscanf(f_in, "%*d %*d %*d %d ", &inz);
      add += inz;
      continue;

    case 'O':
      fscanf(f_in, "%d ", &add);
      continue;

    case 'B':
      fscanf(f_in, "%d %d ", &bank, &slot);
      continue;

    case 'L':
      fscanf(f_in, "%256s ", tmp);

      label_tmp = malloc(sizeof(struct label_def));
      if (label_tmp == NULL) {
	if (error_level == ERROR_LEVEL_MAX)
	  fprintf(stderr, "%s:%d INTERNAL_PASS_1: Out of memory while trying to allocate room for label \"%s\".\n", get_file_name(file_name_id), line_number, tmp);
	else
	  fprintf(stderr, "INTERNAL_PASS_1: Out of memory while trying to allocate room for label \"%s\".\n", tmp);
	return FAILED;
      }
      label_tmp->next = NULL;
      strcpy(label_tmp->label, tmp);
      label_tmp->section_status = section_status;
      if (section_status == ON) {
	label_tmp->section_id = sec_tmp->id;
	/* section labels get a relative address */
	label_tmp->address = add - sec_tmp->address;
      }
      else
	label_tmp->address = add;

      label_tmp->bank = bank;
      label_tmp->slot = slot;
#ifdef W65816
      label_tmp->base = base;
#endif

      if (labels != NULL) {
	label_next = labels;
	while (label_next != NULL) {
	  label_last = label_next;
	  if (strcmp(tmp, label_next->label) == 0) {
	    if ((tmp[0] != '_') || /* both are not local labels */
		(tmp[0] == '_' && section_status == OFF && label_next->section_status == OFF) ||
		(tmp[0] == '_' && section_status == ON && label_next->section_status == ON && label_next->section_id == label_tmp->section_id)) {
	      if (error_level == ERROR_LEVEL_MAX)
		fprintf(stderr, "%s:%d: INTERNAL_PASS_1: Label \"%s\" was defined for the second time.\n", get_file_name(file_name_id), line_number, tmp);
	      else
		fprintf(stderr, "INTERNAL_PASS_1: Label \"%s\" was defined for the second time.\n", tmp);
	      return FAILED;
	    }
	  }
	  label_next = label_next->next;
	}
	label_last->next = label_tmp;
      }
      else
	labels = label_tmp;
      continue;

    case 'f':
      fscanf(f_in, "%d ", &file_name_id);
      continue;

    case 'k':
      fscanf(f_in, "%d ", &line_number);
      continue;

    default:
      if (error_level == ERROR_LEVEL_MAX)
	fprintf(stderr, "%s: INTERNAL_PASS_1: Unknown internal symbol \"%c\".\n", get_file_name(file_name_id), c);
      else
	fprintf(stderr, "INTERNAL_PASS_1: Unknown internal symbol \"%c\".\n", c);
      return FAILED;
    }
  }

  fclose(f_in);

  return SUCCEEDED;
}
