
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"

#include "parse.h"
#include "pass_1.h"
#include "stack.h"
#include "include_file.h"


extern int input_number_error_msg, bankheader_status;
extern int i, size, d, error_level, macro_active;
extern char xyz[256], *buffer, tmp[4096];
extern struct gba_def *defines, *tmp_def, *next_def;
extern struct active_file_info *active_file_info_first, *active_file_info_last, *active_file_info_tmp;
extern struct gba_macro *macros_current;

int latest_stack = 0, stacks_inside = 0, stacks_outside = 0;
struct stack *stacks_first = NULL, *stacks_tmp = NULL, *stacks_last = NULL;
struct stack *stacks_header_first = NULL, *stacks_header_last = NULL;

#ifndef GB
extern int stack_inserted;
#endif




int stack_calculate(char *in, int *value) {

  struct stackitem si[256], ta[256];
  struct stack s;
  int q = 0, b = 0, d, k, op[256];
  unsigned char e;



  /* slice the data into infix format */
  while (*in != 0x0A) {
    if (*in == ' ') {
      in++;
      continue;
    }
    else if (*in == '-') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_MINUS;
      q++;
      in++;
    }
    else if (*in == '+') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_PLUS;
      q++;
      in++;
    }
    else if (*in == '*') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_MULTIPLY;
      q++;
      in++;
    }
    else if (*in == '/') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_DIVIDE;
      q++;
      in++;
    }
    else if (*in == '|') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_OR;
      q++;
      in++;
    }
    else if (*in == '&') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_AND;
      q++;
      in++;
    }
    else if (*in == '^') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_POWER;
      q++;
      in++;
    }
    else if (*in == '#') {
      if (q == 0) {
	if (input_number_error_msg == 0)
	  pass_1_print_error("Syntax error. Invalid use of modulo.\n", ERROR_STC);
	return FAILED;
      }
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_MODULO;
      q++;
      in++;
    }
    else if (*in == '<' && *(in + 1) == '<') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_SHIFT_LEFT;
      q++;
      in += 2;
    }
    else if (*in == '>' && *(in + 1) == '>') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_SHIFT_RIGHT;
      q++;
      in += 2;
    }
    else if (*in == '>' && *(in + 1) != '>')
      break;
    else if (*in == '<' && *(in + 1) != '<')
      break;
    else if (*in == '=' && *(in + 1) == '=')
      break;
    else if (*in == '!' && *(in + 1) == '=')
      break;
    else if (*in == '(') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_LEFT;
      q++;
      b++;
      in++;
    }
    else if (*in == ')') {
      si[q].type = STACK_ITEM_TYPE_OPERATOR;
      si[q].value = SI_OP_RIGHT;
      /* end of expression? */
      if (b == 0)
	break;
      b--;
      q++;
      in++;
    }
    else if (*in == ',')
      break;
    else if (*in == ']')
      break;
    else if (*in == '%') {
      d = 0;
      for (k = 0; k < 32; k++, d = d << 1) {
	in++;
	e = *in;
	if (e == '0' || e == '1')
	  d += e - '0';
	else if (e == ' ' || e == ')' || e == '|' || e == '&' || e == '+' || e == '-' || e == '*' || e == '/' || e == ',' || e == '^' || e == '<' || e == '>' || e == '#' || e == 0x0A)
	  break;
	else {
	  if (input_number_error_msg == 0) {
	    sprintf(xyz, "Got '%c' (%d) when expected (0-1).\n", e, e);
	    pass_1_print_error(xyz, ERROR_NUM);
	  }
	  return FAILED;
	}
      }

      d = d >> 1;

#ifndef W65816
      if (d < -32768 || d > 65535) {
	pass_1_print_error("Out of 16bit range.\n", ERROR_NUM);
	return FAILED;
      }
#else
      if (d < -8388608 || d > 16777215) {
	pass_1_print_error("Out of 24bit range.\n", ERROR_NUM);
	return FAILED;
      }
#endif

      si[q].type = STACK_ITEM_TYPE_VALUE;
      si[q].value = d;
      si[q].sign = SI_SIGN_POSITIVE;
      q++;
    }
    else if (*in == '\'') {
      in++;
      d = *in;
      in++;
      if (*in != '\'') {
        sprintf(xyz, "Got '%c' (%d) when expected \"'\".\n", *in, *in);
        pass_1_print_error(xyz, ERROR_NUM);
	return FAILED;
      }
      in++;

      si[q].type = STACK_ITEM_TYPE_VALUE;
      si[q].value = d;
      si[q].sign = SI_SIGN_POSITIVE;
      q++;
    }
    else if (*in == '$') {
      d = 0;
      for (k = 0; k < 16; k++, d = d << 4) {
	in++;
	e = *in;
	if (e >= '0' && e <= '9')
	  d += e - '0';
	else if (e >= 'A' && e <= 'F')
	  d += e - 'A' + 10;
	else if (e >= 'a' && e <= 'f')
	  d += e - 'a' + 10;
	else if (e == ' ' || e == ')' || e == '|' || e == '&' || e == '+' || e == '-' || e == '*' || e == '/' || e == ',' || e == '^' || e == '<' || e == '>' || e == '#' || e == 0x0A)
	  break;
	else {
	  if (input_number_error_msg == 0) {
	    sprintf(xyz, "Got '%c' (%d) when expected (0-F).\n", e, e);
	    pass_1_print_error(xyz, ERROR_NUM);
	  }
	  return FAILED;
	}
      }

      d = d >> 4;

#ifndef W65816
      if (d < -32768 || d > 65535) {
	pass_1_print_error("Out of 16bit range.\n", ERROR_NUM);
	return FAILED;
      }
#else
      if (d < -8388608 || d > 16777215) {
	pass_1_print_error("Out of 24bit range.\n", ERROR_NUM);
	return FAILED;
      }
#endif

      si[q].type = STACK_ITEM_TYPE_VALUE;
      si[q].value = d;
      si[q].sign = SI_SIGN_POSITIVE;
      q++;
    }
    else if (*in >= '0' && *in <= '9') {
      d = (*in - '0') * 10;
      for (k = 0; k < 16; k++, d *= 10) {
	in++;
	e = *in;
	if (e >= '0' && e <= '9')
	  d += e - '0';
	else if (e == ' ' || e == ')' || e == '|' || e == '&' || e == '+' || e == '-' || e == '*' || e == '/' || e == ',' || e == '^' || e == '<' || e == '>' || e == '#' || e == 0x0A)
	  break;
	else {
	  if (input_number_error_msg == 0) {
	    sprintf(xyz, "Got '%c' (%d) when expected (0-9).\n", e, e);
	    pass_1_print_error(xyz, ERROR_NUM);
	  }
	  return FAILED;
	}
      }

      d /= 10;

#ifndef W65816
      if (d < -32768 || d > 65535) {
	pass_1_print_error("Out of 16bit range.\n", ERROR_NUM);
	return FAILED;
      }
#else
      if (d < -8388608 || d > 16777215) {
	pass_1_print_error("Out of 24bit range.\n", ERROR_NUM);
	return FAILED;
      }
#endif

      si[q].type = STACK_ITEM_TYPE_VALUE;
      si[q].value = d;
      si[q].sign = SI_SIGN_POSITIVE;
      q++;
    }
    else {
      si[q].type = STACK_ITEM_TYPE_STRING;
      for (k = 0; k < 63; k++) {
	e = *in;
	if (e == ' ' || e == ')' || e == '|' || e == '&' || e == '+' || e == '-' || e == '*' || e == '/' || e == ',' || e == '^' || e == '<' || e == '>' || e =='#' || e == 0x0A)
	  break;
	si[q].string[k] = e;
	in++;
      }
      si[q].string[k] = 0;
      si[q].sign = SI_SIGN_POSITIVE;
      q++;
    }

    if (q == 255) {
      pass_1_print_error("Out of space.\n", ERROR_STC);
      return FAILED;
    }
  }

  if (b != 0) {
    pass_1_print_error("Unbalanced parentheses.\n", ERROR_STC);
    return FAILED;
  }

  /* check if there was data before the computation */
  if (q > 1 && (si[0].type == STACK_ITEM_TYPE_STRING || si[0].type == STACK_ITEM_TYPE_VALUE)) {
    if (si[1].type == STACK_ITEM_TYPE_STRING || si[1].type == STACK_ITEM_TYPE_VALUE)
      return STACK_CALCULATE_DELAY;
    if (si[1].type == STACK_ITEM_TYPE_OPERATOR) {
      if (si[1].value == SI_OP_LEFT)
	return STACK_CALCULATE_DELAY;
    }
  }

  i = in - buffer;

  /* fix the sign in every operand */
  for (b = 1, k = 0; k < q; k++) {
    if ((q - k) != 1 && si[k].type == STACK_ITEM_TYPE_OPERATOR && si[k + 1].type == STACK_ITEM_TYPE_OPERATOR) {
      if (si[k].value != SI_OP_LEFT && si[k].value != SI_OP_RIGHT && si[k + 1].value != SI_OP_LEFT && si[k + 1].value != SI_OP_RIGHT) {
	pass_1_print_error("Error in computation syntax.\n", ERROR_STC);
	return FAILED;
      }
    }
    if (si[k].type == STACK_ITEM_TYPE_OPERATOR && si[k].value == SI_OP_MINUS && b == 1) {
      if (si[k + 1].type == STACK_ITEM_TYPE_VALUE || si[k + 1].type == STACK_ITEM_TYPE_STRING) {
	if (si[k + 1].sign == SI_SIGN_POSITIVE)
	  si[k + 1].sign = SI_SIGN_NEGATIVE;
	else
	  si[k + 1].sign = SI_SIGN_POSITIVE;
	/* it wasn't a minus operator, it was a sign */
	si[k].type = STACK_ITEM_TYPE_DELETED;
      }
      else if (si[k + 1].type == STACK_ITEM_TYPE_OPERATOR && si[k + 1].value == SI_OP_LEFT) {
	int o, l;


	o = 1;
	l = k + 2;
	while (o > 0 && l < q) {
	  if (si[l].type == STACK_ITEM_TYPE_VALUE || si[l].type == STACK_ITEM_TYPE_STRING) {
	    if (si[l].sign == SI_SIGN_POSITIVE)
	      si[l].sign = SI_SIGN_NEGATIVE;
	    else
	      si[l].sign = SI_SIGN_POSITIVE;
	  }
	  else if (si[l].type == STACK_ITEM_TYPE_OPERATOR) {
	    if (si[l].value == SI_OP_LEFT)
	      o++;
	    else if (si[l].value == SI_OP_RIGHT)
	      o--;
	  }
	  l++;
	}

	if (o != 0) {
	  pass_1_print_error("Unbalanced parentheses.\n", ERROR_STC);
	  return FAILED;
	}

	si[k].type = STACK_ITEM_TYPE_DELETED;
      }
    }
    /* remove unnecessary + */
    if (si[k].type == STACK_ITEM_TYPE_OPERATOR && si[k].value == SI_OP_PLUS && b == 1) {
      if (si[k + 1].type == STACK_ITEM_TYPE_VALUE || si[k + 1].type == STACK_ITEM_TYPE_STRING)
	si[k].type = STACK_ITEM_TYPE_DELETED;
      else if (si[k + 1].type == STACK_ITEM_TYPE_OPERATOR && si[k + 1].value == SI_OP_LEFT)
	si[k].type = STACK_ITEM_TYPE_DELETED;
    }
    else if (si[k].type == STACK_ITEM_TYPE_VALUE || si[k].type == STACK_ITEM_TYPE_STRING)
      b = 0;
    else if (si[k].type == STACK_ITEM_TYPE_OPERATOR && si[k].value == SI_OP_LEFT)
      b = 1;
  }

  /* convert infix stack into postfix stack */
  for (b = 0, k = 0, d = 0; k < q; k++) {
    /* operands pass through */
    if (si[k].type == STACK_ITEM_TYPE_VALUE) {
      ta[d].type = si[k].type;
      ta[d].value = si[k].value;
      ta[d].sign = si[k].sign;
      d++;
    }
    else if (si[k].type == STACK_ITEM_TYPE_STRING) {
      ta[d].type = si[k].type;
      strcpy(ta[d].string, si[k].string);
      ta[d].sign = si[k].sign;
      d++;
    }
    /* operators get inspected */
    else if (si[k].type == STACK_ITEM_TYPE_OPERATOR) {
      if (b == 0) {
	op[0] = si[k].value;
	b++;
      }
      else {
	if (si[k].value == SI_OP_PLUS) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_PLUS;
	  b++;
	}
	else if (si[k].value == SI_OP_MINUS) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_MINUS;
	  b++;
	}
	else if (si[k].value == SI_OP_MULTIPLY) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_MULTIPLY;
	  b++;
	}
	else if (si[k].value == SI_OP_DIVIDE) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_DIVIDE;
	  b++;
	}
	else if (si[k].value == SI_OP_MODULO) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_MODULO;
	  b++;
	}
	else if (si[k].value == SI_OP_POWER) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_POWER;
	  b++;
	}
	else if (si[k].value == SI_OP_SHIFT_LEFT) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_SHIFT_LEFT;
	  b++;
	}
	else if (si[k].value == SI_OP_SHIFT_RIGHT) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_SHIFT_RIGHT;
	  b++;
	}
	else if (si[k].value == SI_OP_AND) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_AND;
	  b++;
	}
	else if (si[k].value == SI_OP_OR) {
	  b--;
	  while (b != -1 && op[b] != SI_OP_LEFT && op[b] != SI_OP_PLUS && op[b] != SI_OP_MINUS) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	  if (b == -1)
	    b = 0;
	  else
	    b++;
	  op[b] = SI_OP_OR;
	  b++;
	}
	else if (si[k].value == SI_OP_LEFT) {
	  op[b] = SI_OP_LEFT;
	  b++;
	}
	else if (si[k].value == SI_OP_RIGHT) {
	  b--;
	  while (op[b] != SI_OP_LEFT) {
	    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
	    ta[d].value = op[b];
	    b--;
	    d++;
	  }
	}
      }
    }
  }

  /* empty the operator stack */
  while (b > 0) {
    b--;
    ta[d].type = STACK_ITEM_TYPE_OPERATOR;
    ta[d].value = op[b];
    d++;
  }

  /* debug output */
  /*
  for (k = 0; k < d; k++) {
    char ar[] = "+-*()|&/^<>%";

    if (ta[k].type == STACK_ITEM_TYPE_OPERATOR)
      printf("%c ", ar[ta[k].value]);
    else if (ta[k].type == STACK_ITEM_TYPE_VALUE)
      printf("%d ", ta[k].value);
    else
      printf("%s ", ta[k].string);
  }
  printf("\n");
  */

  /* are all the symbols known? */
  if (resolve_stack(ta, d) == SUCCEEDED) {
    s.stack = ta;
    if (error_level == ERROR_LEVEL_MAX) {
      s.linenumber = active_file_info_last->line_current;
      s.filename_id = active_file_info_last->filename_id;
    }
    if (compute_stack(&s, d, value) == FAILED)
      return FAILED;
    return SUCCEEDED;
  }

  /* we have a stack full of computation and we save it for later */
  stacks_tmp = malloc(sizeof(struct stack));
  if (stacks_tmp == NULL) {
    pass_1_print_error("Out of memory error while allocating room for a new stack.\n", ERROR_STC);
    return FAILED;
  }
  stacks_tmp->next = NULL;
  stacks_tmp->stacksize = d;
  stacks_tmp->stack = malloc(sizeof(struct stackitem) * d);
  if (stacks_tmp->stack == NULL) {
    free(stacks_tmp);
    pass_1_print_error("Out of memory error while allocating room for a new stack.\n", ERROR_STC);
    return FAILED;
  }

  if (error_level == ERROR_LEVEL_MAX) {
    stacks_tmp->linenumber = active_file_info_last->line_current;
    stacks_tmp->filename_id = active_file_info_last->filename_id;
  }
  else
    stacks_tmp->linenumber = 0;

  for (q = 0; q < d; q++) {
    if (ta[q].type == STACK_ITEM_TYPE_OPERATOR) {
      stacks_tmp->stack[q].type = STACK_ITEM_TYPE_OPERATOR;
      stacks_tmp->stack[q].value = ta[q].value;
    }
    else if (ta[q].type == STACK_ITEM_TYPE_VALUE) {
      stacks_tmp->stack[q].type = STACK_ITEM_TYPE_VALUE;
      stacks_tmp->stack[q].value = ta[q].value;
      stacks_tmp->stack[q].sign = ta[q].sign;
    }
    else {
      stacks_tmp->stack[q].type = STACK_ITEM_TYPE_STRING;
      stacks_tmp->stack[q].sign = ta[q].sign;
      strcpy(stacks_tmp->stack[q].string, ta[q].string);
    }
  }

  /* outside bankheader sections */
  if (bankheader_status == OFF) {
    if (stacks_first == NULL) {
      stacks_first = stacks_tmp;
      stacks_last = stacks_tmp;
    }
    else {
      stacks_last->next = stacks_tmp;
      stacks_last = stacks_tmp;
    }
    latest_stack = stacks_outside;
    stacks_outside++;

#ifndef GB
    stack_inserted = STACK_OUTSIDE;
#endif
  }
  /* inside a bankheader section */
  else {
    if (stacks_header_first == NULL) {
      stacks_header_first = stacks_tmp;
      stacks_header_last = stacks_tmp;
    }
    else {
      stacks_header_last->next = stacks_tmp;
      stacks_header_last = stacks_tmp;
    }
    latest_stack = stacks_inside;
    stacks_inside++;

#ifndef GB
    stack_inserted = STACK_INSIDE;
#endif
  }

  return INPUT_NUMBER_STACK;
}


int resolve_stack(struct stackitem s[], int x) {

  while (x > 0) {
    if (s->type == STACK_ITEM_TYPE_STRING) {

      if (macro_active != 0 && s->string[0] == '\\') {
	
	char c;
	int a, b, p, k;


	if (s->string[1] == '@' && s->string[2] == 0) {
	  s->type = STACK_ITEM_TYPE_VALUE;
	  s->value = macros_current->calls - 1;
	}
	else {
	  for (a = 0, b = 0; s->string[a + 1] != 0 && a < 4; a++) {
	    c = s->string[a + 1];
	    if (c < '0' && c > '9') {
	      pass_1_print_error("Error in MACRO argument number definition.\n", ERROR_DIR);
	      return FAILED;
	    }
	    b = (b * 10) + (c - '0');
	  }

	  if (b == 0 || b > 256) {
	    pass_1_print_error("MACRO argument number must be between 1 and 256.\n", ERROR_DIR);
	    return FAILED;
	  }

	  p = i;
	  i = macros_current->arguments_start[b - 1];
	  /* read the argument */
	  k = input_number();
	  i = p;

	  if (k != SUCCEEDED)
	    return FAILED;

	  s->type = STACK_ITEM_TYPE_VALUE;
	  s->value = d;
	}
      }
      else {
	tmp_def = defines;
	while (tmp_def != NULL) {
	  if (strcmp(tmp_def->alias, s->string) == 0) {
	    if (tmp_def->type == DEFINITION_TYPE_STRING) {
	      sprintf(xyz, "Definition \"%s\" is a string definition.\n", tmp_def->alias);
	      pass_1_print_error(xyz, ERROR_STC);
	      return FAILED;
	    }
	    s->type = STACK_ITEM_TYPE_VALUE;
	    s->value = tmp_def->value;
	    break;
	  }
	  tmp_def = tmp_def->next;
	}
	/* it wasn't a define? */
	if (tmp_def == NULL)
	  return FAILED;
      }
    }
    s++;
    x--;
  }

  return SUCCEEDED;
}


int compute_stack(struct stack *sta, int x, int *result) {

  struct stackitem *s;
  double v[256];
  int r, t, q, z;



  s = sta->stack;
  for (r = 0, t = 0; r < x; r++, s++) {
    if (s->type == STACK_ITEM_TYPE_VALUE) {
      if (s->sign == SI_SIGN_NEGATIVE)
	v[t] = -s->value;
      else
	v[t] = s->value;
      t++;
    }
    else {
      switch (s->value) {
      case SI_OP_PLUS:
	v[t - 2] += v[t - 1];
	t--;
	break;
      case SI_OP_MINUS:
	v[t - 2] -= v[t - 1];
	t--;
	break;
      case SI_OP_MULTIPLY:
	v[t - 2] *= v[t - 1];
	t--;
	break;      
      case SI_OP_OR:
	v[t - 2] = (int)v[t - 1] | (int)v[t - 2];
	t--;
	break;
      case SI_OP_AND:
	v[t - 2] = (int)v[t - 1] & (int)v[t - 2];
	t--;
	break;
      case SI_OP_MODULO:
	if (((int)v[t - 1]) == 0) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: COMPUTE_STACK: Modulo by zero.\n", get_file_name(sta->filename_id), sta->linenumber);
	  else
	    fprintf(stderr, "COMPUTE_STACK: Modulo by zero.\n");
	  return FAILED;
	}
	v[t - 2] = (int)v[t - 2] % (int)v[t - 1];
	t--;
	break;
      case SI_OP_DIVIDE:
	if (((int)v[t - 1]) == 0) {
	  if (error_level == ERROR_LEVEL_MAX)
	    fprintf(stderr, "%s:%d: COMPUTE_STACK: Division by zero.\n", get_file_name(sta->filename_id), sta->linenumber);
	  else
	    fprintf(stderr, "COMPUTE_STACK: Division by zero.\n");
	  return FAILED;
	}
	v[t - 2] /= v[t - 1];
	t--;
	break;
      case SI_OP_POWER:
	q = 1;
	for (z = 0; z < v[t - 1]; z++)
	  q *= v[t - 2];
	v[t - 2] = q;
	t--;
	break;
      case SI_OP_SHIFT_LEFT:
	v[t - 2] = (int)v[t - 2] << (int)v[t - 1];
	t--;
	break;
      case SI_OP_SHIFT_RIGHT:
	v[t - 2] = (int)v[t - 2] >> (int)v[t - 1];
	t--;
	break;
      }
    }
  }

  *result = v[0];
  return SUCCEEDED;
}
