
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "analyze.h"
#include "memory.h"


#define READ_T t[3] + (t[2] << 8) + (t[1] << 16) + (t[0] << 24); t += 4;


extern struct object_file *obj_first, *obj_last, *obj_tmp;
extern struct reference *reference_first, *reference_last;
extern struct section *sec_first, *sec_last, *sec_hd_first, *sec_hd_last;
extern struct stack *stacks_first, *stacks_last;
extern struct label *label_first, *label_last;
extern struct slot slots[256];
extern unsigned char header[256];
extern int rombanks, verbose_mode, wla_version, name_status, section_overwrite, symbol_mode, mode_status;
extern int rombankmap_position, object_format, emptyfill;
extern int *banks, *bankaddress;



int add_reference(struct reference *r) {

  r->file_id = obj_tmp->id;
  r->next = NULL;

  if (reference_first == NULL) {
    reference_first = r;
    reference_last = r;
    r->prev = NULL;
  }
  else {
    r->prev = reference_last;
    reference_last->next = r;
    reference_last = r;
  }

  return SUCCEEDED;
}


int add_stack(struct stack *sta) {

  sta->file_id = obj_tmp->id;
  sta->next = NULL;

  if (stacks_first == NULL) {
    stacks_first = sta;
    stacks_last = sta;
    sta->prev = NULL;
  }
  else {
    sta->prev = stacks_last;
    stacks_last->next = sta;
    stacks_last = sta;
  }

  return SUCCEEDED;
}


int add_section(struct section *s) {

  struct section *q;



  s->file_id = obj_tmp->id;
  s->next = NULL;

  if (strcmp(s->name, "BANKHEADER") == 0) {

    q = sec_hd_first;
    while (q != NULL) {
      if (q->bank == s->bank) {
	fprintf(stderr, "%s: ADD_SECTION: BANKHEADER section for bank %d was defined for the second time.\n", obj_tmp->name, s->bank);
	return FAILED;
      }
      q = q->next;
    }

    if (sec_hd_first == NULL) {
      sec_hd_first = s;
      sec_hd_last = s;
    }
    else {
      sec_hd_last->next = s;
      sec_hd_last = s;
    }
  }
  else {
    if (sec_first == NULL) {
      sec_first = s;
      sec_last = s;
      s->prev = NULL;
    }
    else {
      s->prev = sec_last;
      sec_last->next = s;
      sec_last = s;
    }
  }

  return SUCCEEDED;
}


int add_label(struct label *l) {

  l->file_id = obj_tmp->id;
  l->next = NULL;

  if (label_first == NULL) {
    label_first = l;
    label_last = l;
    l->prev = NULL;
  }
  else {
    l->prev = label_last;
    label_last->next = l;
    label_last = l;
  }

  return SUCCEEDED;
}


int obtain_rombankmap(void) {

  int map_found = OFF, i, x, banksize, a;
  struct object_file *o;
  unsigned char *t;



  /* initialize values */
  for (i = 0; i < rombanks; i++)
    banks[i] = 0;

  o = obj_first;
  while (o != NULL) {
    if (o->format == object_format) {
      t = o->data + rombankmap_position;

      /* obtain status */
      i = *t;
      t++;

      /* general rombanksize? */
      if (i == 0) {
	/* obtain banksize */
	banksize = READ_T

	o->memorymap = t;
	map_found = ON;
	for (i = 0; i < o->rom_banks; i++) {
	  if (banks[i] == 0) {
	    banks[i] = banksize;
	    bankaddress[i] = i * banksize;
	  }
	  else if (banks[i] != banksize) {
	    fprintf(stderr, "OBTAIN_ROMBANKMAP: ROMBANKMAPs don't match.\n");
	    return FAILED;
	  }
	}
      }
      else {
	for (a = 0, x = 0; x < o->rom_banks; x++) {
	  banksize = READ_T
	  if (banks[x] == 0) {
	    banks[x] = banksize;
	    bankaddress[x] = a;
	  }
	  else if (banks[x] != banksize) {
	    fprintf(stderr, "OBTAIN_ROMBANKMAP: ROMBANKMAPs don't match.\n");
	    return FAILED;
	  }
	  a += banksize;
	}
	
	o->memorymap = t;
	map_found = ON;
      }
    }

    o = o->next;
  }

  if (map_found == OFF) {
    fprintf(stderr, "OBTAIN_ROMBANKMAP: No object files.\n");
    return FAILED;
  }

  return SUCCEEDED;
}


int obtain_memorymap(void) {

  struct object_file *o;
  int map_found = OFF, i, x;
  unsigned char *t;
  struct slot s[256];



  for (i = 0; i < 256; i++)
    slots[i].usage = OFF;

  o = obj_first;
  while (o != NULL) {
    if (o->format == object_format) {
      t = o->memorymap;

      /* obtain slots */
      i = *t;
      t++;

      for (x = 0; i > 0; i--, x++) {
	slots[x].usage = ON;
	slots[x].address =  READ_T
	slots[x].size =  READ_T
      }

      o->exported_defines = t;
      map_found = ON;
      break;
    }

    o = o->next;
  }

  if (map_found == OFF) {
    fprintf(stderr, "OBTAIN_MEMORYMAP: No object files.\n");
    return FAILED;
  }

  /* check if the following memorymaps equal to the previous one */
  o = o->next;
  while (o != NULL) {
    if (o->format == object_format) {

      for (i = 0; i < 256; i++)
	s[i].usage = OFF;
      t = o->memorymap;

      /* obtain slots */
      i = *t;
      t++;

      for (x = 0; i > 0; i--, x++) {
	s[x].usage = ON;
	s[x].address =  READ_T
	s[x].size =  READ_T
      }

      o->exported_defines = t;

      for (x = 0, i = 0; i < 256; i++) {
	if (s[i].usage == ON) {
	  if (slots[i].usage == OFF) {
	    x = 1;
	    break;
	  }
	  if (slots[i].address == s[i].address && slots[i].size == s[i].size)
	    continue;
	  x = 1;
	  break;
	}
	else {
	  if (slots[i].usage == ON) {
	    x = 1;
	    break;
	  }
	}
      }

      if (x == 1) {
	fprintf(stderr, "OBTAIN_MEMORYMAP: The object files are compiled for different memory architectures.\n");
	return FAILED;
      }
    }

    o = o->next;
  }

  return SUCCEEDED;
}


int collect_dlr(void) {

  struct reference *r;
  struct stack *s;
  struct label *l;
  int section, x, i, n, q;
  unsigned char *t;



  section = 0;
  obj_tmp = obj_first;
  while (obj_tmp != NULL) {

    /* VERSION GB/Z80 */
    if (obj_tmp->format == object_format) {
      t = obj_tmp->exported_defines;
      i = READ_T

      /* load defines */
      for (; i > 0; i--) {
	l = malloc(sizeof(struct label));
	if (l == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}
	for (x = 0; *t != 0; t++, x++)
	  l->name[x] = *t;
	l->name[x] = 0;
	t++;
	l->status = LABEL_STATUS_DEFINE;
	l->address = READ_T
	l->base = 0;

	add_label(l);
      }

      /* load labels */
      i = READ_T

      for (; i > 0; i--) {
	l = malloc(sizeof(struct label));
	if (l == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}
	for (x = 0; *t != 0; t++, x++)
	  l->name[x] = *t;
	l->name[x] = 0;
	t++;
	x = *t++;
	if (x == 0)
	  l->section_status = OFF;
	else {
	  l->section_status = ON;
	  l->section = x + section;
	}
	l->status = LABEL_STATUS_LABEL;
	l->address = READ_T
	l->base = (l->address >> 16) & 0xFF;
	l->address = l->address & 0xFFFF;
	l->bank = READ_T
	l->slot = *t++;

	add_label(l);
      }

      i = READ_T

      /* load references */
      for (; i > 0; i--) {
	r = malloc(sizeof(struct reference));
	if (r == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}
	for (x = 0; *t != 0; t++, x++)
	  r->name[x] = *t;
	r->name[x] = 0;
	t++;
	r->type = *t++;
	x = *t++;
	if (x == 0)
	  r->section_status = OFF;
	else {
	  r->section_status = ON;
	  r->section = x + section;
	}
	r->address = READ_T
	r->bank = READ_T
	r->slot = *t++;

	add_reference(r);
      }

      i = READ_T

      /* load pending calculations */
      for (; i > 0; i--) {
	s = malloc(sizeof(struct stack));
	if (s == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}

	s->type = *t++;
	s->section = section + *t++;

	if (s->section == section)
	  s->section_status = OFF;
	else
	  s->section_status = ON;

	s->address = READ_T
	s->linenumber = READ_T
	s->bank = READ_T
	s->slot = *t++;
	x = *t++;
	s->stacksize = x;

	s->stack = malloc(sizeof(struct stackitem) * x);
	if (s->stack == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  free(s);
	  return FAILED;
	}

	add_stack(s);

	for (n = 0; n != x; n++) {
	  s->stack[n].type = *t++;
	  s->stack[n].sign = *t++;
	  if (s->stack[n].type == STACK_ITEM_TYPE_STRING) {
	    for (q = 0; *t != 0; t++, q++)
	      s->stack[n].string[q] = *t;
	    s->stack[n].string[q] = 0;
	    t++;
	  }
	  else {
	    s->stack[n].value = READ_T
	  }
	}
      }

      /* save pointer to data block area */
      obj_tmp->data_blocks = t;
    }

    /* VERSION LIB */
    else if (obj_tmp->format == WLA_VERSION_LIB) {
      t = obj_tmp->data + LIB_DEFINES;
      i = READ_T

      /* load definitions */
      for (; i > 0; i--) {
	l = malloc(sizeof(struct label));
	if (l == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}
	for (x = 0; *t != 0; t++, x++)
	  l->name[x] = *t;
	l->name[x] = 0;
	t++;
	l->status = LABEL_STATUS_DEFINE;
	l->address = READ_T
	l->bank = obj_tmp->bank;
	l->slot = obj_tmp->slot;
	l->base = 0;

	add_label(l);
      }

      i = READ_T

      /* load label */
      for (; i > 0; i--) {
	l = malloc(sizeof(struct label));
	if (l == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}
	for (x = 0; *t != 0; t++, x++)
	  l->name[x] = *t;
	l->name[x] = 0;
	t++;
	x = *t++;
	l->section = x + section;
	l->section_status = ON;
	l->status = LABEL_STATUS_LABEL;
	l->address = READ_T
	l->base = (l->address >> 16) & 0xFF;
	l->address = l->address & 0xFFFF;
	l->bank = obj_tmp->bank;
	l->slot = obj_tmp->slot;

	add_label(l);
      }

      i = READ_T

      /* load references */
      for (; i > 0; i--) {
	r = malloc(sizeof(struct reference));
	if (r == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}
	for (x = 0; *t != 0; t++, x++)
	  r->name[x] = *t;
	r->name[x] = 0;
	t++;
	r->type = *t++;
	x = *t++;
	r->section = x + section;
	r->section_status = ON;
	r->address = READ_T

	r->bank = obj_tmp->bank;
	r->slot = obj_tmp->slot;

	add_reference(r);
      }

      i = READ_T

      /* load pending calculations */
      for (; i > 0; i--) {
	s = malloc(sizeof(struct stack));
	if (s == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  return FAILED;
	}

	s->bank = obj_tmp->bank;
	s->slot = obj_tmp->slot;
	s->type = *t++;
	s->section_status = ON;
	s->section = section + *t++;
	s->address = READ_T
	s->linenumber = READ_T
	x = *t++;
	s->stacksize = x;

	s->stack = malloc(sizeof(struct stackitem) * x);
	if (s->stack == NULL) {
	  fprintf(stderr, "COLLECT_DLR: Out of memory.\n");
	  free(s);
	  return FAILED;
	}

	add_stack(s);

	for (n = 0; n != x; n++) {
	  s->stack[n].type = *t++;
	  s->stack[n].sign = *t++;
	  if (s->stack[n].type == STACK_ITEM_TYPE_STRING) {
	    for (q = 0; *t != 0; t++, q++)
	      s->stack[n].string[q] = *t;
	    s->stack[n].string[q] = 0;
	    t++;
	  }
	  else {
	    s->stack[n].value = READ_T
	  }
	}
      }

      /* save pointer to data block area */
      obj_tmp->data_blocks = t;
    }

    obj_tmp = obj_tmp->next;
    section += 0x1000;
  }

  return SUCCEEDED;
}


int parse_data_blocks(void) {

  struct section *s;
  int section, i, y, x;
  unsigned char *t, *p;



  obj_tmp = obj_first;
  section = 0;
  while (obj_tmp != NULL) {

    if (obj_tmp->format == object_format) {
      t = obj_tmp->data_blocks;
      p = obj_tmp->data + obj_tmp->size;
      for (y = 1; t < p; ) {
	x = *t++;

	if (x == DATA_TYPE_BLOCK) {

	  /* address */
	  i = READ_T
	  /* amount of bytes */
	  x = READ_T

	  for (; x > 0; x--, i++)
	    if (mem_insert(i, *t++) == FAILED)
	      return FAILED;
	}

	else if (x == DATA_TYPE_SECTION) {

	  s = malloc(sizeof(struct section));
	  if (s == NULL) {
	    fprintf(stderr, "PARSE_DATA_BLOCKS: Out of memory.\n");
	    return FAILED;
	  }
	  s->id = y + section;
	  y++;

	  /* name */
	  i = 0;
	  while (*t != 0 && *t != 1 && *t != 2 && *t != 3 && *t != 4 && *t != 5)
	    s->name[i++] = *t++;
	  s->name[i] = 0;
	  s->status = *t++;
	  s->address = READ_T
	  s->bank = READ_T
	  s->slot = *t++;
	  s->size = READ_T
	  s->data = t;
	  t += s->size;

	  if (add_section(s) == FAILED)
	    return FAILED;
	}
      }
      obj_tmp = obj_tmp->next;
      section += 0x1000;
      continue;
    }

    else if (obj_tmp->format == WLA_VERSION_LIB) {

      t = obj_tmp->data_blocks;
      p = obj_tmp->data + obj_tmp->size;
      for (y = 1; t < p; ) {

	s = malloc(sizeof(struct section));
	if (s == NULL) {
	  fprintf(stderr, "PARSE_DATA_BLOCKS: Out of memory.\n");
	  return FAILED;
	}
	s->id = y + section;
	y++;

	/* name */
	i = 0;
	while (*t != 0)
	  s->name[i++] = *t++;
	s->name[i] = 0;
	t++;
	s->size = READ_T
	s->status = SECTION_STATUS_FREE;
	s->data = t;
	s->address = 0;
	s->bank = obj_tmp->bank;
	s->slot = obj_tmp->slot;
	t += s->size;

	add_section(s);
      }
      obj_tmp = obj_tmp->next;
      section += 0x1000;
      continue;
    }
  }

  return SUCCEEDED;
}


int obtain_rombanks(void) {

  unsigned char *t;
  int rb = 0, k, s;



  /* GB */
  if (mode_status == MODE_GB) {

    /* obtain the biggest rom size */
    s = 0;
    obj_tmp = obj_first;
    while (obj_tmp != NULL) {
      if (obj_tmp->format == WLA_VERSION_GB) {
	if (get_gb_rombanks(obj_tmp->data[GB_ROM_SIZE], &k) == FAILED) {
	  fprintf(stderr, "%s: OBTAIN_ROMBANKS: Unknown amount of ROM banks.\n", obj_tmp->name);
	  return FAILED;
	}
	obj_tmp->rom_banks = k;
	if (k != rb)
	  s++;
	if (k > rb) {
	  /* copy the biggest cartridgetype id as well */
	  header[0] = obj_tmp->data[GB_ROM_SIZE];
	  rb = k;
	}
      }
      obj_tmp = obj_tmp->next;
    }

    rombanks = rb;
    emptyfill = header[GB_EMPTY_FILL - GB_ROM_SIZE];

    if (s > 1)
      fprintf(stderr, "OBTAIN_ROMBANKS: Using the biggest selected amount of ROM banks (%d).\n", rombanks);
  }
  /* Z80 */
  else {

    /* obtain the biggest rom size */
    s = 0;
    obj_tmp = obj_first;
    while (obj_tmp != NULL) {
      if (obj_tmp->format == WLA_VERSION_Z80) {

	t = obj_tmp->data + Z80_ROMBANKS;
	k = t[3] + (t[2] << 8) + (t[1] << 16) + (t[0] << 24);

	obj_tmp->rom_banks = k;

	if (k != rb)
	  s++;
	if (k > rb)
	  rb = k;
      }
      obj_tmp = obj_tmp->next;
    }

    /* emptyfill has been obtained in the header checks */
    rombanks = rb;

    if (s > 1)
      fprintf(stderr, "OBTAIN_ROMBANKS: Using the biggest selected amount of ROM banks (%d).\n", rombanks);
  }

  return SUCCEEDED;
}


int get_gb_rombanks(int i, int *rombanks) {

  if (i == 0)
    *rombanks = 2;
  else if (i == 1)
    *rombanks =  4;
  else if (i == 2)
    *rombanks =  8;
  else if (i == 3)
    *rombanks =  16;
  else if (i == 4)
    *rombanks =  32;
  else if (i == 5)
    *rombanks =  64;
  else if (i == 6)
    *rombanks =  128;
  else if (i == 0x52)
    *rombanks =  72;
  else if (i == 0x53)
    *rombanks =  80;
  else if (i == 0x54)
    *rombanks =  96;
  else
    return FAILED;

  return SUCCEEDED;
}


int clean_up_dlr(void) {

  struct reference *r, *re;
  struct stack *s, *st;
  struct label *l, *la;
  struct section *se, *sec, *sect;



  se = sec_first;
  while (se != NULL) {
    /* remove duplicates of unique sections and all the related data */
    if (strlen(se->name) >= 2 && se->name[0] == '_' && se->name[1] == '_') {
      sec = se->next;
      while (sec != NULL) {
	if (strcmp(se->name, sec->name) == 0) {
	  sect = sec;

	  (sect->prev)->next = sect->next;
	  if (sect->next != NULL)
	    (sect->next)->prev = sect->prev;
	  else
	    sec_last = sect;

	  /* free references */
	  r = reference_first;
	  while (r != NULL) {
	    if (r->section_status == ON && r->section == sect->id) {
	      re = r;
	      if (re->prev == NULL)
		reference_first = re->next;
	      else
		re->prev->next = re->next;
	      if (re->next == NULL)
		reference_last = NULL;
	      else
		re->next->prev = re->prev;

	      r = r->next;
	      free(re);
	      continue;
	    }
	    r = r->next;
	  }

	  /* free stacks */
	  s = stacks_first;
	  while (s != NULL) {
	    if (s->section_status == ON && s->section == sect->id) {
	      st = s;
	      if (st->stack != NULL) {
		free(st->stack);
		st->stack = NULL;
	      }
	      if (st->prev == NULL)
		stacks_first = st->next;
	      else
		st->prev->next = st->next;
	      if (st->next == NULL)
		stacks_last = NULL;
	      else
		st->next->prev = st->prev;

	      s = s->next;
	      free(st);
	      continue;
	    }
	    s = s->next;
	  }

	  /* free labels */
	  l = label_first;
	  while (l != NULL) {
	    if (l->section_status == ON && l->section == sect->id) {
	      la = l;
	      if (la->prev == NULL)
		label_first = la->next;
	      else
		la->prev->next = la->next;
	      if (la->next == NULL)
		label_last = NULL;
	      else
		la->next->prev = la->prev;

	      l = l->next;
	      free(la);
	      continue;
	    }
	    l = l->next;
	  }

	  sec = sec->next;
	  free(sect);
	  continue;
	}
	sec = sec->next;
      }
    }
    se = se->next;
  }

  return SUCCEEDED;
}
