
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "main.h"
#include "memory.h"
#include "files.h"



extern unsigned char header[256], name[16];
extern struct object_file *obj_first;
extern int emptyfill, checksum_type, snes_rom_mode, snes_mode, snes_rom_speed;

int rombankmap_position, memorymap_position;
int object_format, emptyfill;
int mode_status = MODE_NONE;



int check_file_types(void) {

  struct object_file *o;


  o = obj_first;
  while (o != NULL) {
    if (strncmp((char *)o->data, "WLA9", 4) == 0) {
      if (mode_status == MODE_Z80) {
	fprintf(stderr, "You can't mix GB-Z80 and Z80/6502/6510/65816 files.\n");
	return FAILED;
      }
      o->format = WLA_VERSION_GB;
      mode_status = MODE_GB;
    }
    else if (strncmp((char *)o->data, "WLAB", 4) == 0) {
      if (mode_status == MODE_GB) {
	fprintf(stderr, "You can't mix GB-Z80 and Z80/6502/6510/65816 files.\n");
	return FAILED;
      }
      o->format = WLA_VERSION_Z80;
      mode_status = MODE_Z80;
    }
    else if (strncmp((char *)o->data, "WLAP", 4) == 0)
      o->format = WLA_VERSION_LIB;
    else {
      fprintf(stderr, "File \"%s\" is of unknown format (\"%c%c%c%c\").\n", o->name, o->data[0], o->data[1], o->data[2], o->data[3]);
      return FAILED;
    }
    o = o->next;
  }

  /* use the correct adder */
  if (mode_status == MODE_GB) {
    rombankmap_position = GB_ROMBANKMAP;
    object_format = WLA_VERSION_GB;
  }
  else {
    rombankmap_position = Z80_ROMBANKMAP;
    object_format = WLA_VERSION_Z80;
  }

  return SUCCEEDED;
}


int check_headers(int *name_status) {

  struct object_file *o;
  int x, misc_bits = 0;


  x = 0;
  *name_status = 0;
  o = obj_first;

  /* GB-Z80 */
  if (mode_status == MODE_GB) {
    while (o != NULL) {
      if (o->format == WLA_VERSION_GB) {
	if ((*name_status == 0) && (o->data[GB_ROM_NAME_STATUS] != 0)) {
	  memcpy(name, o->data + GB_ROM_NAME, 16);
	  *name_status = 1;
	}
	else if (o->data[GB_ROM_NAME_STATUS] != 0 && memcmp(name, o->data + GB_ROM_NAME, 16) != 0) {
	  fprintf(stderr, "CHECK_HEADERS: The object files are from different projects.\n");
	  return FAILED;
	}

	if (x == 0) {
	  memcpy(header, o->data + GB_ROM_SIZE, GB_COMPUTE_COMPLEMENT_CHECK - GB_ROM_SIZE + 1);
	  x++;
	}
	else if (memcmp(header + 1, o->data + GB_RAM_SIZE, GB_COMPUTE_COMPLEMENT_CHECK - GB_RAM_SIZE + 1) != 0) {
	  fprintf(stderr, "CHECK_HEADERS: The object files are from different projects.\n");
	  return FAILED;
	}
      }

      o = o->next;
    }
  }
  /* other systems (z80, 6502, 6510, 65816) */
  else {
    while (o != NULL) {
      if (o->format == WLA_VERSION_Z80) {
	if (x == 0) {
	  emptyfill = *(o->data + Z80_EMPTY_FILL);
	  misc_bits = *(o->data + Z80_MISC_BITS);
	  checksum_type = misc_bits & 1;
	  snes_rom_mode = misc_bits & 2;
	  snes_mode = misc_bits & 4;
	  snes_rom_speed = misc_bits & 8;
	  x++;
	}
	else if (emptyfill != *(o->data + Z80_EMPTY_FILL) || misc_bits != *(o->data + Z80_MISC_BITS)) {
	  fprintf(stderr, "CHECK_HEADERS: The object files are from different projects.\n");
	  return FAILED;
	}
      }

      o = o->next;
    }
  }

  if (x == 0) {
    fprintf(stderr, "CHECK_HEADERS: There's nothing to link.\n");
    return FAILED;
  }

  return SUCCEEDED;
}
