
/* note that Z80 here means systems with no GB-Z80 */

#define Z80_FORMAT_INFO 0
#define Z80_ROMBANKS 4
#define Z80_EMPTY_FILL 8
#define Z80_MISC_BITS 9
#define Z80_ROMBANKMAP 10

#define OUTPUT_ROM 0
#define OUTPUT_PRG 1

#define SNES_ROM_MODE_LOROM 0
#define SNES_ROM_MODE_HIROM 2

#define SNES_ROM_SPEED_SLOWROM 0
#define SNES_ROM_SPEED_FASTROM 8

#define REFERENCE_TYPE_DIRECT_16BIT 0
#define REFERENCE_TYPE_RELATIVE 1
#define REFERENCE_TYPE_DIRECT_8BIT 2
#define REFERENCE_TYPE_DIRECT_24BIT 3

#define GB_FORMAT_INFO 0
#define GB_ROM_NAME_STATUS 4
#define GB_ROM_NAME 5
#define GB_ROM_SIZE 21
#define GB_RAM_SIZE 22
#define GB_EMPTY_FILL 23
#define GB_CARTRIDGE_TYPE 24
#define GB_LICENSEE_CODE 25
#define GB_LICENSEE_CODE_OLD 26
#define GB_LICENSEE_CODE_NEW_C1 27
#define GB_LICENSEE_CODE_NEW_C2 28
#define GB_COMPUTE_CHECKSUM 29
#define GB_COMPUTE_COMPLEMENT_CHECK 30
#define GB_ROMBANKMAP 31

#define LIB_FORMAT_INFO 0
#define LIB_DEFINES 4

#define SECTION_STATUS_FREE 0
#define SECTION_STATUS_FORCE 1
#define SECTION_STATUS_OVERWRITE 2
#define SECTION_STATUS_HEADER 3
#define SECTION_STATUS_SEMIFREE 4
#define SECTION_STATUS_ABSOLUTE 5

#define WLA_VERSION_GB 0
#define WLA_VERSION_Z80 1
#define WLA_VERSION_LIB 2

#define MODE_NONE 0
#define MODE_GB 1
#define MODE_Z80 2

#define DATA_TYPE_BLOCK 0
#define DATA_TYPE_SECTION 1

#define LABEL_STATUS_LABEL 0
#define LABEL_STATUS_DEFINE 1

#define FAILED 1
#define SUCCEEDED 0

#define STATE_NONE 0
#define STATE_OBJECT 1
#define STATE_LIBRARY 2
#define STATE_HEADER 3
#define STATE_FOOTER 4

#define OFF 0
#define ON 1

#define MAX_NAME_LENGTH 63



struct object_file {
  unsigned char *data;
  unsigned char *memorymap;
  unsigned char *exported_defines;
  unsigned char *data_blocks;
  char name[MAX_NAME_LENGTH + 1];
  int rom_banks;
  int format;
  int bank;
  int slot;
  int size;
  int id;
  struct object_file *next;
};

struct label {
  char name[MAX_NAME_LENGTH + 1];
  int file_id;
  int address;
  int rom_address;
  int section;
  int section_status;
  int bank;
  int base;
  int slot;
  int status;
  struct label *next;
  struct label *prev;
};

struct reference {
  char name[MAX_NAME_LENGTH + 1];
  int  address;
  int  type;
  int  section;
  int  section_status;
  int  bank;
  int  slot;
  int  file_id;
  struct reference *next;
  struct reference *prev;
};

struct section {
  char name[MAX_NAME_LENGTH + 1];
  int  file_id;
  int  address;
  int  status;
  int  bank;
  int  slot;
  int  size;
  int  id;
  int  i;
  unsigned char *data;
  struct section *next;
  struct section *prev;
};

struct slot {
  int address;
  int size;
  int usage;
};

#define STACK_ITEM_TYPE_VALUE 0
#define STACK_ITEM_TYPE_OPERATOR 1
#define STACK_ITEM_TYPE_STRING 2

#define SI_OP_PLUS 0
#define SI_OP_MINUS 1
#define SI_OP_MULTIPLY 2
#define SI_OP_OR 5
#define SI_OP_AND 6
#define SI_OP_DIVIDE 7
#define SI_OP_POWER 8
#define SI_OP_SHIFT_LEFT 9
#define SI_OP_SHIFT_RIGHT 10
#define SI_OP_MODULO 11

#define SI_SIGN_POSITIVE 0
#define SI_SIGN_NEGATIVE 1

#define STACKS_TYPE_8BIT 0
#define STACKS_TYPE_16BIT 1
#define STACKS_TYPE_24BIT 2

struct stackitem {
  char string[MAX_NAME_LENGTH + 1];
  int type;
  int value;
  int sign;
};

struct stack {
  struct stackitem *stack;
  struct stack *next;
  struct stack *prev;
  int file_id;
  int stacksize;
  int linenumber;
  int type;
  int bank;
  int slot;
  int section_status;
  int section;
  int address;
};
