
/*
 wlalink - part of wla dx gb-z80/z80/6502/6510/65816 macro assembler package by/copyright
 ville helin <vhelin@cc.hut.fi>. this is gpl software.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "main.h"
#include "memory.h"
#include "files.h"
#include "check.h"
#include "analyze.h"
#include "write.h"



struct object_file *obj_first = NULL, *obj_last = NULL, *obj_tmp;
struct reference *reference_first = NULL, *reference_last = NULL;
struct section *sec_first = NULL, *sec_last = NULL, *sec_hd_first = NULL, *sec_hd_last = NULL;
struct stack *stacks_first = NULL, *stacks_last = NULL;
struct label *label_first = NULL, *label_last = NULL;
struct slot slots[256];
unsigned char header[256], *rom, *rom_usage, name[16], *file_header = NULL, *file_footer = NULL;
int romsize, rombanks, banksize, verbose_mode = OFF, wla_version, name_status, section_overwrite = OFF, symbol_mode = OFF;
int pc_bank, pc_full, pc_slot, pc_slot_max;
int file_header_size, file_footer_size, *banks = NULL, *bankaddress = NULL;
int output_mode = OUTPUT_ROM;
int program_start, program_end, checksum_type = 0, snes_rom_mode = SNES_ROM_MODE_LOROM, snes_mode = 0, snes_rom_speed = SNES_ROM_SPEED_SLOWROM;

extern int mode_status, emptyfill;

#ifdef AMIGA
char version_string[] = "$VER: WLALINK 4.2 (10.05.2001)";
#endif



int main(int argc, char *argv[]) {

  int i, x, y, q, inz;
  struct section *s;



  atexit(procedures_at_exit);

  i = SUCCEEDED;
  x = SUCCEEDED;

  if (argc > 1)
    x = parse_flags(argv[1]);
  else
    i = FAILED;

  if (x == FAILED && argc != 3)
    i = FAILED;
  if (x == SUCCEEDED && argc != 4)
    i = FAILED;

  if (i == FAILED) {
    printf("\nWLALINK GB-Z80/Z80/6502/6510/65816 WLA Macro Assembler Linker v4.2\n");
    printf("Copyright 2000-2001 Ville Helin\n");
    printf("USAGE: %s [-brvs] <LINK FILE> <OUTPUT FILE>\n", argv[0]);
    printf("Options:\n");
    printf("b  Program file output (not in GB-Z80 mode)\n");
    printf("r  ROM file output (default)\n");
    printf("v  Verbose messages\n");
    printf("s  NO$GMB symbol file\n\n");
    return 0;
  }

  /* load files */
  if (load_files(argv, argc) == FAILED)
    return 1;

  /* check file types */
  if (check_file_types() == FAILED)
    return 1;

  /* check object headers and get project information */
  if (check_headers(&name_status) == FAILED)
    return 1;

  /* obtain the amount of rom banks */
  if (obtain_rombanks() == FAILED)
    return 1;

  banks = malloc(sizeof(int) * rombanks);
  if (banks == NULL) {
    fprintf(stderr, "MAIN: Out of memory error.\n");
    return 1;
  }
  bankaddress = malloc(sizeof(int) * rombanks);
  if (bankaddress == NULL) {
    fprintf(stderr, "MAIN: Out of memory error.\n");
    return 1;
  }

  /* obtain rom bank map and check the project integrity */
  if (obtain_rombankmap() == FAILED)
    return 1;

  /* obtain memory map and check the project integrity */
  if (obtain_memorymap() == FAILED)
    return 1;

  /* calculate romsize */
  for (romsize = 0, x = 0; x < rombanks; x++)
    romsize += banks[x];

  /* collect all defines, labels and outside references */
  if (collect_dlr() == FAILED)
    return 1;

  /* take rom size and allocate memory */
  if (allocate_rom() == FAILED)
    return 1;

  /* parse data blocks */
  if (parse_data_blocks() == FAILED)
    return 1;

  /* clean up the structures */
  if (clean_up_dlr() == FAILED)
    return 1;

  /*
  l = label_first;
  while (l != NULL) {
    printf("name: \"%s\"\n", l->name);
    printf("sect: \"%d\"\n", l->section);
    printf("slot: \"%d\"\n", l->slot);
    printf("address: \"%d\"\n", l->address);
    l = l->next;
  }
  */

  /* insert sections */
  if (insert_sections() == FAILED)
    return 1;

  /*
  s = sec_first;
  while (s != NULL) {
    printf("file: \"%s\"\n", s->file);
    printf("name: \"%s\"\n", s->name);
    printf("addr: %d\n", s->address);
    printf("stat: %d\n", s->status);
    printf("bank: %d\n", s->bank);
    printf("slot: %d\n", s->slot);
    printf("size: %d\n", s->size);
    s = s->next;
  }
  */

  /* fix references */
  if (fix_references() == FAILED)
    return 1;

  /* compute pending calculations */
  if (compute_pending_calculations() == FAILED)
    return 1;

  /* write gb header */
  if (mode_status == MODE_GB) {
    if (write_gb_header(name_status) == FAILED)
      return 1;
  }

  /* write !gb information (checksum) */
  if (mode_status == MODE_Z80 && checksum_type != 0) {
    /* sms checksum */
    if (checksum_type == 1) {
      if (romsize < 0x8000) {
	fprintf(stderr, "MAIN: SMS/GG checksum computing requires a ROM of at least 32KB.\n");
	return 1;
      }

      inz = 0;
      for (x = 0; x < 0x7FF0; x++)
	inz += rom[x];

      if (romsize > (256 * 1024))
	y = 256 * 1024;
      else
	y = romsize;

      for (x = 0x8000; x < y; x++)
	inz += rom[x];

      mem_insert(0x7FFA, inz & 0xFF);
      mem_insert(0x7FFB, (inz & 0xFF00) >> 8);
    }
  }

  /* write snes information */
  if (snes_mode != 0) {
    if (romsize < 0x10000) {
      fprintf(stderr, "MAIN: SNES cartridge information writing requires a ROM of at least 64KB.\n");
      return 1;
    }

    inz = 32;
    if (snes_rom_mode == SNES_ROM_MODE_HIROM)
      inz += 1;
    if (snes_rom_speed == SNES_ROM_SPEED_FASTROM)
      inz += 16;
    mem_insert(0xFFD5, inz);
  }

  /* write rom file */
  if (write_rom_file(argv[argc - 1]) == FAILED)
    return 1;

  /* export symbolic information file */
  if (symbol_mode == ON)
    if (write_sym_file(argv[argc - 1]) == FAILED)
      return FAILED;

  /* show rom information */
  y = 0;
  if (verbose_mode == ON) {
    x = 0;
    for (i = 0; i < romsize; i++) {
      if (rom_usage[i] == 0 && x == 0) {
	x = 1;
	y = i;
      }
      else if (rom_usage[i] != 0 && x == 1) {
	if (y == (i - 1))
	  fprintf(stderr, "Free space at $%.4x.\n", y);
	else
	  fprintf(stderr, "Free space at $%.4x-$%.4x.\n", y, i - 1);
	x = 0;
      }
    }

    if (x == 1) {
      if (y == (i - 1))
	fprintf(stderr, "Free space at $%.4x.\n", y);
      else
	fprintf(stderr, "Free space at $%.4x-$%.4x.\n", y, i - 1);
    }

    for (y = 0, q = 0; y < romsize; q++) {
      for (x = 0, inz = 0; inz < banks[q]; inz++) {
	if (rom_usage[y++] == 0)
	  x++;
      }
      printf("Bank %.2d has %.5d bytes (%.2f%%) free.\n", q, x, (((double)x) / banks[q]) * 100);
    }

    /* ROM data */
    if (output_mode == OUTPUT_ROM) {
      for (i = 0, y = 0; i < romsize; i++) {
	if (rom_usage[i] == 0)
	  y++;
      }

      fprintf(stderr, "%d unused bytes of total %d.\n", y, romsize);
      q = romsize;
    }
    /* program file data */
    else {
      for (i = program_start, y = 0; i < program_end; i++) {
	if (rom_usage[i] == 0)
	  y++;
      }

      q = program_end - program_start + 1;
      fprintf(stderr, "%d unused bytes (%.2f%%) of total %d.\n", y, (((double)y) / q) * 100, q);
    }

    if (file_header_size != 0)
      fprintf(stderr, "File header size %d.\n", file_header_size);
    if (file_footer_size != 0)
      fprintf(stderr, "File footer size %d.\n", file_footer_size);
    i = file_header_size + file_footer_size;

    s = sec_hd_first;
    while (s != NULL) {
      fprintf(stderr, "Bank %d header section size %d.\n", s->bank, s->size);
      i += s->size;
      s = s->next;
    }

    if (i != 0) {
      fprintf(stderr, "Total %d additional bytes (from headers and footers).\n", i);
      fprintf(stderr, "Total size %d bytes.\n", i + q);
    }
  }

  return 0;
}


void procedures_at_exit(void) {

  struct object_file *o;
  struct reference *r;
  struct section *s;
  struct stack *sta;
  struct label *l;



  while (obj_first != NULL) {
    if (obj_first->data != NULL)
      free(obj_first->data);
    o = obj_first;
    obj_first = obj_first->next;
    free(o);
  }

  while (label_first != NULL) {
    l = label_first;
    label_first = label_first->next;
    free(l);
  }

  while (reference_first != NULL) {
    r = reference_first;
    reference_first = reference_first->next;
    free(r);
  }

  while (stacks_first != NULL) {
    sta = stacks_first->next;
    free(stacks_first->stack);
    free(stacks_first);
    stacks_first = sta;
  }

  while (sec_first != NULL) {
    s = sec_first->next;
    free(sec_first);
    sec_first = s;
  }

  while (sec_hd_first != NULL) {
    s = sec_hd_first->next;
    free(sec_hd_first);
    sec_hd_first = s;
  }

  if (banks != NULL)
    free(banks);
  if (bankaddress != NULL)
    free(bankaddress);
}


int parse_flags(char *f) {

  int l, output_mode_defined = 0;



  if (*f != '-') {
    return FAILED;
  }

  l = strlen(f);
  if (l == 1)
    return FAILED;

  for (f++, l--; l > 0; l--, f++) {
    switch (*f) {

    case 'v':
      verbose_mode = ON;
      continue;

    case 's':
      symbol_mode = ON;
      continue;

    case 'b':
      if (output_mode_defined == 1)
	return FAILED;
      output_mode_defined++;
      output_mode = OUTPUT_PRG;
      continue;

    case 'r':
      if (output_mode_defined == 1)
	return FAILED;
      output_mode_defined++;
      output_mode = OUTPUT_ROM;
      continue;

    default:
      return FAILED;
    }
  }

  return SUCCEEDED;
}


int allocate_rom(void) {

  rom = malloc(sizeof(unsigned char) * romsize);
  if (rom == NULL) {
    fprintf(stderr, "ALLOCATE_ROM: Out of memory.\n");
    return FAILED;
  }
  rom_usage = malloc(sizeof(unsigned char) * romsize);
  if (rom_usage == NULL) {
    fprintf(stderr, "ALLOCATE_ROM: Out of memory.\n");
    return FAILED;
  }
  memset(rom, emptyfill, romsize);
  memset(rom_usage, 0, romsize);

  return SUCCEEDED;
}
