
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "memory.h"
#include "write.h"
#include "files.h"



extern struct reference *reference_first, *reference_last;
extern struct label *label_first, *label_last;
extern struct object_file *obj_first, *obj_last, *obj_tmp;
extern struct section *sec_first, *sec_last, *sec_hd_first, sec_hd_last;
extern struct stack *stacks_first, *stacks_last;
extern struct slot slots[256];
extern unsigned char header[256], *rom, *rom_usage, name[16];
extern unsigned char *file_header, *file_footer;
extern int romsize, rombanks, banksize, verbose_mode, wla_version, name_status, section_overwrite, symbol_mode;
extern int pc_bank, pc_full, pc_slot, pc_slot_max, snes_rom_mode;
extern int file_header_size, file_footer_size, *bankaddress, *banks;
extern int memory_file_id, output_mode;
extern int program_start, program_end;



int insert_sections(void) {

  struct section *s;
  int d, f, i, x;


  /* force sections */
  s = sec_first;
  while (s != NULL) {
    if (s->status == SECTION_STATUS_FORCE) {
      memory_file_id = s->file_id;
      banksize = banks[s->bank];
      pc_bank = s->address;
      pc_slot = slots[s->slot].address + pc_bank;
      pc_full = pc_bank + bankaddress[s->bank];
      pc_slot_max = slots[s->slot].address + slots[s->slot].size;
      d = pc_full;
      i = d + s->size;
      section_overwrite = OFF;
      for (; d < i; d++) {
	if (rom_usage[d] != 0 && rom[d] != s->data[d - pc_full])
	  break;
      }
      if (d == i) {
	for (i = 0; i < s->size; i++)
	  if (mem_insert_pc(s->data[i]) == FAILED)
	    return FAILED;
      }
      else {
	fprintf(stderr, "%s: INSERT_SECTIONS: No room for a section \"%s\".\n", get_file_name(s->file_id), s->name);
	return FAILED;
      }
    }

    s = s->next;
  }

  /* absolute sections */
  s = sec_first;
  while (s != NULL) {
    if (s->status == SECTION_STATUS_ABSOLUTE) {
      d = s->address;
      section_overwrite = ON;
      for (i = 0; i < s->size; i++)
	if (mem_insert(d + i, s->data[i]) == FAILED)
	  return FAILED;
    }

    s = s->next;
  }

  /* free sections */
  s = sec_first;
  while (s != NULL) {
    if (s->status == SECTION_STATUS_FREE || s->status == SECTION_STATUS_SEMIFREE) {
      pc_bank = s->address;
      d = bankaddress[s->bank];
      i = FAILED;
      while (i == FAILED) {
	f = pc_bank;
	for (x = 0; pc_bank < banks[s->bank] && rom_usage[pc_bank + d] == 0 && x < s->size; pc_bank++, x++);
	if (x == s->size) {
	  i = SUCCEEDED;
	  break;
	}
	if (pc_bank == banks[s->bank])
	  break;
	for (; pc_bank < banks[s->bank] && rom_usage[pc_bank + d] != 0; pc_bank++);
      }

      if (i == SUCCEEDED) {
	memory_file_id = s->file_id;
	banksize = banks[s->bank];
	pc_bank = f;
	pc_slot = slots[s->slot].address + pc_bank;
	pc_full = pc_bank + bankaddress[s->bank];
	pc_slot_max = slots[s->slot].address + slots[s->slot].size;
	s->address = pc_bank;

	section_overwrite = OFF;
	for (i = 0; i < s->size; i++)
	  if (mem_insert_pc(s->data[i]) == FAILED)
	    return FAILED;
      }
      else {
	fprintf(stderr, "%s: INSERT_SECTIONS: No room for a section \"%s\".\n", get_file_name(s->file_id), s->name);
	return FAILED;
      }
    }

    s = s->next;
  }

  /* overwrite sections */
  s = sec_first;
  while (s != NULL) {
    if (s->status == SECTION_STATUS_OVERWRITE) {
      memory_file_id = s->file_id;
      banksize = banks[s->bank];
      pc_bank = s->address;
      pc_slot = slots[s->slot].address + pc_bank;
      pc_full = pc_bank + bankaddress[s->bank];
      pc_slot_max = slots[s->slot].address + slots[s->slot].size;
      section_overwrite = ON;
      for (i = 0; i < s->size; i++)
	if (mem_insert_pc(s->data[i]) == FAILED)
	  return FAILED;
    }

    s = s->next;
  }

  return SUCCEEDED;
}


int fix_references(void) {

  struct reference *r;
  struct section *s;
  struct label *l, *m;
  int i, x;


  section_overwrite = OFF;

  /* fix label addresses */
  l = label_first;
  while (l != NULL) {
    if (l->status != LABEL_STATUS_DEFINE) {
      if (l->section_status == ON) {
	s = sec_first;
	while (s != NULL) {
	  if (s->id == l->section) {
	    l->address += s->address;
	    l->rom_address = l->address;
	    if (s->status != SECTION_STATUS_ABSOLUTE)
	      l->address += slots[l->slot].address;
	    break;
	  }
	  s = s->next;
	}
      }
      else {
	l->rom_address = l->address;
	l->address += slots[l->slot].address;
      }
    }
    l = l->next;
  }

  /* check out if a label exists more than once in a different place */
  l = label_first;
  while (l != NULL) {
    m = l->next;
    while (m != NULL) {
      if (strcmp(m->name, l->name) == 0) {
	if (l->address != m->address && m->name[0] != '_') {
	  if (l->status == LABEL_STATUS_DEFINE)
	    fprintf(stderr, "%s: FIX_REFERENCES: Definition \"%s\" was defined more than once.\n", get_file_name(l->file_id), l->name);
	  else
	    fprintf(stderr, "%s: FIX_REFERENCES: Label \"%s\" was defined more than once.\n", get_file_name(l->file_id), l->name);
	  return FAILED;
	}
     }
      m = m->next;
    }
    l = l->next;
  }

  /* insert references */
  r = reference_first;
  while (r != NULL) {
    x = r->address;
    /* search for the section of the reference and fix the address */
    if (r->section_status == ON) {
      s = sec_first;
      while (s != NULL) {
	if (s->id == r->section) {
	  x += s->address;
	  r->address += s->address;
	  break;
	}
	s = s->next;
      }
      if (s == NULL) {
	if (write_bank_header_references(r) == FAILED)
	  return FAILED;
	r = r->next;
	continue;
      }
    }

    if (!(r->section_status == ON && s->status == SECTION_STATUS_ABSOLUTE)) {
      x += bankaddress[r->bank];
      r->address += slots[r->slot].address;
    }

    /* find the destination */
    l = label_first;
    while (l != NULL) {
      if (strcmp(l->name, r->name) == 0) {
	/* search for the section of the referencee */
	if (r->name[0] == '_') {
	  if (l->file_id == r->file_id) {
	    if (l->section_status != r->section_status) {
	      l = l->next;
	      continue;
	    }
	    if (l->section_status == ON && l->section != r->section) {
	      l = l->next;
	      continue;
	    }
	  }
	  else {
	    l = l->next;
	    continue;
	  }
	}

	/* direct 16bit */
	if (r->type == REFERENCE_TYPE_DIRECT_16BIT) {
	  i = l->address;
	  mem_insert_ref(x, (unsigned char)(i & 0xff));
	  mem_insert_ref(x + 1, (unsigned char)((i >> 8) & 0xff));
	  break;
	}
	/* direct / relative 8bit with a definition */
	else if (l->status == LABEL_STATUS_DEFINE && (r->type == REFERENCE_TYPE_DIRECT_8BIT || r->type == REFERENCE_TYPE_RELATIVE)) {
	  i = (l->address & 0xFFFF);
	  if (i > 255 || i < -127) {
	    fprintf(stderr, "%s: FIX_REFERENCES: Value ($%x) of \"%s\" is too much to be a 8bit value.\n", get_file_name(r->file_id), i, l->name);
	    return FAILED;
	  }
	  mem_insert_ref(x, (unsigned char)(i & 0xff));
	  break;
	}
	/* direct 24bit */
	else if (r->type == REFERENCE_TYPE_DIRECT_24BIT) {
	  i = l->address;
	  if (l->status == LABEL_STATUS_LABEL)
	    i += get_snes_pc_bank(l);
	  mem_insert_ref(x, (unsigned char)(i & 0xff));
	  mem_insert_ref(x + 1, (unsigned char)((i >> 8) & 0xff));
	  mem_insert_ref(x + 2, (unsigned char)((i >> 16) & 0xff));
	  break;
	}
	/* relative 8bit with a label */
	else if (r->type == REFERENCE_TYPE_RELATIVE) {
	  i = (l->address & 0xFFFF) - r->address - 1;
	  if (i < -127 || i > 128) {
	    fprintf(stderr, "%s: FIX_REFERENCES: Too large distance (%d bytes from $%x to $%x \"%s\") for a 8bit reference.\n", get_file_name(r->file_id), i, r->address, l->address, l->name);
	    return FAILED;
	  }
	  mem_insert_ref(x, (unsigned char)(i & 0xff));
	  break;
	}
	else {
	  fprintf(stderr, "%s: FIX_REFERENCES: A direct 8bit reference to label \"%s\"?\n", get_file_name(r->file_id), l->name);
	  return FAILED;
	}
      }
      l = l->next;
    }
    if (l == NULL) {
      fprintf(stderr, "%s: FIX_REFERENCES: Reference to an unknown label \"%s\".\n", get_file_name(r->file_id), r->name);
      return FAILED;
    }

    r = r->next;
  }

  return SUCCEEDED;
}


int write_sym_file(char *outname) {

  struct label *l;
  FILE *fop;
  char *p;
  int y;


  strcpy((char *)header, outname);
  p = (char *)header;
  for (y = 0; y < 255 && *p != '.' && *p != 0x0; y++, p++);
  *p++ = '.';
  *p++ = 's';
  *p++ = 'y';
  *p++ = 'm';
  *p = 0x0;

  fop = fopen((char *)header, "wb");
  if (fop == NULL) {
    fprintf(stderr, "MAIN: Error opening file \"%s\".\n", (char *)header);
    return FAILED;
  }

  fprintf(fop, "; this file was created with wlalink by ville helin <vhelin@cc.hut.fi>.\n");
  fprintf(fop, "; no$gmb symbolic information for \"%s\".\n", outname);

  l = label_first;
  while (l != NULL) {
    fprintf(fop, "%.4x:%.4x %s\n", l->bank, l->address, l->name);
    l = l->next;
  }

  fclose(fop);

  return SUCCEEDED;
}


int write_gb_header(int name_status) {

  int i, x;


  if (name_status != 0) {
    for (i = 0; i < 16; i++)    /* name */
      mem_insert_gb_rom_header(308 + i, name[i]);
  }
  mem_insert_gb_rom_header(327, header[GB_CARTRIDGE_TYPE - GB_ROM_SIZE]); /* cartridge type */
  mem_insert_gb_rom_header(328, header[GB_ROM_SIZE - GB_ROM_SIZE]); /* rom size */
  mem_insert_gb_rom_header(329, header[GB_RAM_SIZE - GB_ROM_SIZE]); /* ram size */
  if (header[GB_LICENSEE_CODE - GB_ROM_SIZE] == 0)   /* licensee code */
    mem_insert_gb_rom_header(331, header[GB_LICENSEE_CODE_OLD - GB_ROM_SIZE]);
  else {
    mem_insert_gb_rom_header(331, 51);
    mem_insert_gb_rom_header(324, header[GB_LICENSEE_CODE_NEW_C1 - GB_ROM_SIZE]);
    mem_insert_gb_rom_header(325, header[GB_LICENSEE_CODE_NEW_C2 - GB_ROM_SIZE]);
  }

  if (header[GB_COMPUTE_COMPLEMENT_CHECK - GB_ROM_SIZE] != 0) { /* complement check */
    i = 0;
    for (x = 0x134; x <= 0x14C; x++)
      i += rom[x];
    i += 25;
    mem_insert_gb_rom_header(0x14d, (unsigned char)(0 - (i & 0xff)));
  }

  if (header[GB_COMPUTE_CHECKSUM - GB_ROM_SIZE] != 0) { /* checksum */
    i = 0;
    for (x = 0; x < 0x14E; x++)
      i += rom[x];
    for (x = 0x150; x < romsize; x++)
      i += rom[x];

    mem_insert_gb_rom_header(0x14e, (unsigned char)((i & 0xff00) >> 8));
    mem_insert_gb_rom_header(0x14f, (unsigned char)(i & 0xff));
  }

  return SUCCEEDED;
}


int write_rom_file(char *outname) {

  struct section *s;
  FILE *fop;
  int i, b, e;


  fop = fopen(outname, "wb");
  if (fop == NULL) {
    fprintf(stderr, "WRITE_ROM_FILE: Error opening file \"%s\".\n", outname);
    return FAILED;
  }

  if (file_header != NULL)
    fwrite(file_header, 1, file_header_size, fop);

  /* ROM output mode */
  if (output_mode == OUTPUT_ROM) {
    /* write bank by bank and bank header sections */
    for (i = 0; i < rombanks; i++) {
      s = sec_hd_first;
      while (s != NULL) {
	if (s->bank == i) {
	  fwrite(s->data, 1, s->size, fop);
	  break;
	}
	s = s->next;
      }

      fwrite(rom + bankaddress[i], 1, banks[i], fop);
    }
  }
  /* program file output mode */
  else {
    for (i = 0; i < romsize; i++)
      if (rom_usage[i] != 0)
	break;
    b = i;
    for (e = b; i < romsize; i++)
      if (rom_usage[i] != 0)
	e = i;

    s = sec_hd_first;
    while (s != NULL) {
      if (s->bank == 0) {
	fwrite(s->data, 1, s->size, fop);
	break;
      }
      s = s->next;
    }

    fwrite(rom + b, 1, e - b + 1, fop);
    program_start = b;
    program_end = e;
  }

  if (file_footer != NULL)
    fwrite(file_footer, 1, file_footer_size, fop);

  fclose(fop);

  return SUCCEEDED;
}


int compute_pending_calculations(void) {

  struct section *s;
  struct stack *sta;
  int k;


  section_overwrite = ON;

  sta = stacks_first;
  while (sta != NULL) {

    /* find source address */
    if (sta->section_status == ON) {

      /* get section address */
      s = sec_first;
      while (s != NULL) {
	if (sta->section == s->id)
	  break;
	s = s->next;
      }
      /* it must be a bank header section! */
      if (s == NULL) {
	if (write_bank_header_calculations(sta) == FAILED)
	  return FAILED;
	sta = sta->next;
	continue;
      }

      if (s->status != SECTION_STATUS_ABSOLUTE) {
	pc_bank = sta->address + s->address;
	pc_full = pc_bank + bankaddress[s->bank];

	/* get slot address */
	pc_slot = slots[s->slot].address + pc_bank;
	pc_slot_max = slots[s->slot].address + slots[s->slot].size;
	banksize = banks[s->bank];
      }
      else {
	pc_full = sta->address + s->address;
	pc_bank = 0;
	pc_slot = 0;
	pc_slot_max = 0x10000;
	banksize = 0x10000;
      }
    }
    else {
      pc_full = sta->address + bankaddress[sta->bank];
      pc_bank = sta->address;

      /* get slot address */
      pc_slot = slots[sta->slot].address + pc_bank;
      pc_slot_max = slots[sta->slot].address + slots[sta->slot].size;

      banksize = banks[sta->bank];
    }

    /* parse stack items */
    if (parse_stack(sta) == FAILED)
      return FAILED;

    /* all the references have been decoded, now compute */
    if (compute_stack(sta, &k) == FAILED)
      return FAILED;

    memory_file_id = sta->file_id;

    if (sta->type == STACKS_TYPE_8BIT) {
      if (k < -127 || k > 255) {
	if (sta->linenumber == 0)
	  fprintf(stderr, "%s: COMPUTE_PENDING_CALCULATIONS: Result (%d/$%x) of a computation is out of 8bit range.\n",
		  get_file_name(sta->file_id), k, k);
	else
	  fprintf(stderr, "%s:%d COMPUTE_PENDING_CALCULATIONS: Result (%d/$%x) of a computation is out of 8bit range.\n",
		  get_file_name(sta->file_id), sta->linenumber, k, k);
	return FAILED;
      }
      if (mem_insert_pc_fix(k) == FAILED)
	return FAILED;
    }
    else if (sta->type == STACKS_TYPE_16BIT) {
      if (k < -32768 || k > 65535) {
	if (sta->linenumber == 0)
	  fprintf(stderr, "%s: COMPUTE_PENDING_CALCULATIONS: Result (%d/$%x) of a computation is out of 16bit range.\n",
		  get_file_name(sta->file_id), k, k);
	else
	  fprintf(stderr, "%s:%d COMPUTE_PENDING_CALCULATIONS: Result (%d/$%x) of a computation is out of 16bit range.\n",
		  get_file_name(sta->file_id), sta->linenumber, k, k);
	return FAILED;
      }
      if (mem_insert_pc_fix(k & 0xFF) == FAILED)
	return FAILED;
      if (mem_insert_pc_fix((k >> 8) & 0xFF) == FAILED)
	return FAILED;
    }
    else {
      if (k < -8388608 || k > 16777215) {
	if (sta->linenumber == 0)
	  fprintf(stderr, "%s: COMPUTE_PENDING_CALCULATIONS: Result (%d/$%x) of a computation is out of 24bit range.\n",
		  get_file_name(sta->file_id), k, k);
	else
	  fprintf(stderr, "%s:%d COMPUTE_PENDING_CALCULATIONS: Result (%d/$%x) of a computation is out of 24bit range.\n",
		  get_file_name(sta->file_id), sta->linenumber, k, k);
	return FAILED;
      }
      if (mem_insert_pc_fix(k & 0xFF) == FAILED)
	return FAILED;
      if (mem_insert_pc_fix((k >> 8) & 0xFF) == FAILED)
	return FAILED;
      if (mem_insert_pc_fix((k >> 16) & 0xFF) == FAILED)
	return FAILED;
    }

    /* next stack computation */
    sta = sta->next;
  }

  return SUCCEEDED;
}


int compute_stack(struct stack *sta, int *result) {

  struct stackitem *s;
  int r, t, q, z, x;
  double v[256];


  x = sta->stacksize;
  s = sta->stack;
  for (r = 0, t = 0; r < x; r++, s++) {
    if (s->type == STACK_ITEM_TYPE_VALUE) {
      if (s->sign == SI_SIGN_NEGATIVE)
	v[t] = -s->value;
      else
	v[t] = s->value;
      t++;
    }
    else {
      switch (s->value) {
      case SI_OP_PLUS:
	v[t - 2] += v[t - 1];
	t--;
	break;
      case SI_OP_MINUS:
	v[t - 2] -= v[t - 1];
	t--;
	break;
      case SI_OP_MULTIPLY:
	v[t - 2] *= v[t - 1];
	t--;
	break;      
      case SI_OP_OR:
	v[t - 2] = (int)v[t - 1] | (int)v[t - 2];
	t--;
	break;
      case SI_OP_AND:
	v[t - 2] = (int)v[t - 1] & (int)v[t - 2];
	t--;
	break;
      case SI_OP_MODULO:
	if (((int)v[t - 1]) == 0) {
	  if (sta->linenumber == 0)
	    fprintf(stderr, "%s: COMPUTE_STACK: Modulo by zero.\n", get_file_name(sta->file_id));
	  else
	    fprintf(stderr, "%s:%d COMPUTE_STACK: Modulo by zero.\n", get_file_name(sta->file_id), sta->linenumber);
	  return FAILED;
	}
	v[t - 2] = (int)v[t - 2] % (int)v[t - 1];
	t--;
	break;
      case SI_OP_DIVIDE:
	if (((int)v[t - 1]) == 0) {
	  if (sta->linenumber == 0)
	    fprintf(stderr, "%s: COMPUTE_STACK: Division by zero.\n", get_file_name(sta->file_id));
	  else
	    fprintf(stderr, "%s:%d COMPUTE_STACK: Division by zero.\n", get_file_name(sta->file_id), sta->linenumber);
	  return FAILED;
	}
	v[t - 2] /= v[t - 1];
	t--;
	break;
      case SI_OP_POWER:
	q = 1;
	for (z = 0; z < v[t - 1]; z++)
	  q *= v[t - 2];
	v[t - 2] = q;
	t--;
	break;
      case SI_OP_SHIFT_LEFT:
	v[t - 2] = (int)v[t - 2] << (int)v[t - 1];
	t--;
	break;
      case SI_OP_SHIFT_RIGHT:
	v[t - 2] = (int)v[t - 2] >> (int)v[t - 1];
	t--;
	break;
      }
    }
  }

  *result = v[0];
  return SUCCEEDED;
}


int write_bank_header_calculations(struct stack *sta) {

  struct section *s;
  unsigned char *t;
  int k;


  /* parse stack items */
  if (parse_stack(sta) == FAILED)
    return FAILED;

  /* all the references have been decoded, now compute */
  if (compute_stack(sta, &k) == FAILED)
    return FAILED;

  s = sec_hd_first;
  while (sta->section != s->id)
    s = s->next;

  t = s->data + sta->address;

  if (sta->type == STACKS_TYPE_8BIT) {
    if (k < -127 || k > 255) {
      if (sta->linenumber == 0)
	fprintf(stderr, "%s: WRITE_BANK_HEADER_CALCULATIONS: Result (%d/$%x) of a computation is out of 8bit range.\n",
		get_file_name(sta->file_id), k, k);
      else
	fprintf(stderr, "%s:%d WRITE_BANK_HEADER_CALCULATIONS: Result (%d/$%x) of a computation is out of 8bit range.\n",
		get_file_name(sta->file_id), sta->linenumber, k, k);
      return FAILED;
    }
    *t = k & 0xFF;    
  }
  else if (sta->type == STACKS_TYPE_16BIT) {
    if (k < -32768 || k > 65535) {
      if (sta->linenumber == 0)
	fprintf(stderr, "%s: WRITE_BANK_HEADER_CALCULATIONS: Result (%d/$%x) of a computation is out of 16bit range.\n",
		get_file_name(sta->file_id), k, k);
      else
	fprintf(stderr, "%s:%d WRITE_BANK_HEADER_CALCULATIONS: Result (%d/$%x) of a computation is out of 16bit range.\n",
		get_file_name(sta->file_id), sta->linenumber, k, k);
      return FAILED;
    }
    *t = k & 0xFF;
    t++;
    *t = (k >> 8) & 0xFF;
  }
  else {
    if (k < -8388608 || k > 16777215) {
      if (sta->linenumber == 0)
	fprintf(stderr, "%s: WRITE_BANK_HEADER_CALCULATIONS: Result (%d/$%x) of a computation is out of 24bit range.\n",
		get_file_name(sta->file_id), k, k);
      else
	fprintf(stderr, "%s:%d WRITE_BANK_HEADER_CALCULATIONS: Result (%d/$%x) of a computation is out of 24bit range.\n",
		get_file_name(sta->file_id), sta->linenumber, k, k);
      return FAILED;
    }
    *t = k & 0xFF;
    t++;
    *t = (k >> 8) & 0xFF;
    t++;
    *t = (k >> 16) & 0xFF;
  }

  return SUCCEEDED;
}


int write_bank_header_references(struct reference *r) {

  struct section *s;
  struct label *l;
  unsigned char *t;
  int a;


  s = sec_hd_first;
  while (r->section != s->id)
    s = s->next;

  t = s->data + r->address;

  /* find the destination */
  l = label_first;
  while (l != NULL) {
    if (strcmp(l->name, r->name) == 0) {
      a = l->address;
      /* direct 16bit */
      if (r->type == REFERENCE_TYPE_DIRECT_16BIT) {
	*t = a & 0xFF;
	t++;
	*t = (a >> 8) & 0xFF;
	break;
      }
      /* direct 8bit */
      else if (r->type == REFERENCE_TYPE_DIRECT_8BIT) {
	if (a > 255 || a < -127) {
	  fprintf(stderr, "%s: WRITE_BANK_HEADER_REFERENCES: Value (%d/$%x) of \"%s\" is too much to be a 8bit value.\n",
		  get_file_name(r->file_id), a, a, l->name);
	  return FAILED;
	}
	*t = a & 0xFF;
	break;
      }
      /* direct 24bit */
      else if (r->type == REFERENCE_TYPE_DIRECT_24BIT) {
	if (l->status == LABEL_STATUS_LABEL)
	  a += get_snes_pc_bank(l);
	*t = a & 0xFF;
	t++;
	*t = (a >> 8) & 0xFF;
	t++;
	*t = (a >> 16) & 0xFF;
	break;
      }
      else {
	fprintf(stderr, "%s: WRITE_BANK_HEADER_REFERENCES: Unknown reference (type %d) to label \"%s\".\n",
		get_file_name(r->file_id), r->type, l->name);
	return FAILED;
      }
    }
    l = l->next;
  }
  if (l == NULL) {
    fprintf(stderr, "%s: WRITE_BANK_HEADER_REFERENCES: Reference to an unknown label \"%s\".\n", get_file_name(r->file_id), r->name);
    return FAILED;
  }

  return SUCCEEDED;
}


int parse_stack(struct stack *sta) {

  struct stackitem *si;
  struct label *l;
  int g, k;


  si = sta->stack;
  g = 0;
  k = 0;
  while (g != sta->stacksize) {
    if (si->type == STACK_ITEM_TYPE_STRING) {
      l = label_first;
      while (l != NULL) {
	if (strcmp(l->name, si->string) == 0) {
	  if (si->string[0] == '_') {
	    if (sta->section == l->section) {
	      k = l->address;
	      break;
	    }
	    else {
	      l = l->next;
	      continue;
	    }
	  }
	  else {
	    k = l->address;
	    break;
	  }
	}

	l = l->next;
      }

      if (l == NULL) {
	if (sta->linenumber == 0)
	  fprintf(stderr, "%s: PARSE_STACK: Unresolved reference to \"%s\".\n", get_file_name(sta->file_id), si->string);
	else
	  fprintf(stderr, "%s:%d PARSE_STACK: Unresolved reference to \"%s\".\n", get_file_name(sta->file_id), sta->linenumber, si->string);
	return FAILED;
      }

      /* 65816 cpu bank fix */
      if (sta->type == STACKS_TYPE_24BIT && l->status == LABEL_STATUS_LABEL)
	k += get_snes_pc_bank(l);

      si->value = k;
      si->type = STACK_ITEM_TYPE_VALUE;
    }
    si++;
    g++;
  }

  return SUCCEEDED;
}


int get_snes_pc_bank(struct label *l) {

  int x, k;


  /* use rom_address instead of address, as address points to
     the position in destination machine's memory, not in rom */
  k = bankaddress[l->bank] + l->rom_address;

  if (snes_rom_mode == SNES_ROM_MODE_HIROM)
    x = (k / 0x10000);
  else
    x = k / 0x8000;

  x = (x + l->base) << 16;

  return x;
}
